////////////////////////////////////////////////////////////////////////
/*! \file   BasicGLViewer.cpp
 *  \brief  Yet another basic GL Viewer.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 */
////////////////////////////////////////////////////////////////////////
// $Id: BasicGLViewer.cpp,v 1.121 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2006/11/29] : Creation.

#include "BasicGLViewer.h"

#include <QApplication>
#include <QFileDialog>
#include <QAction>
#include <QMouseEvent>

#include "Utilities.h"
#include "Scene.h"
#include "Shader.h"
#include "Feedback.h"
#include "FrameBufferObject.h"
#include "ShadowMap.h"
#include "EnvironmentMap.h"
#include "GlossyMap.h"
#include "OcclusionQuery.h"

static double smoothedDelay = 0.0;
static int    smoothedCount = 0;
static unsigned int smoothedFragmentsGenerated = 0;
static unsigned int smoothedFragmentsDrawn = 0;

static QString delayString  = "";
static void resetDelay()
{
    smoothedDelay = 0.0;
    smoothedCount = 0;
    smoothedFragmentsGenerated = 0;
    smoothedFragmentsDrawn = 0;
    delayString = "";
}

////////////////////////////////////////////////////////////////////////

BasicGLViewer::BasicGLViewer(QWidget* parent)
: QGLViewer(parent),
  _mode(RENDER_SHADERS),
  _subMode(SUBMODE_COLOR),
  _shader(0),
  _shader2(0),
  _statMode(false),
  _debug(0),
  _width(512),
  _height(512)
{
    _scene = new Scene;
    Q_CHECK_PTR(_scene);

    // Use max FPS possible from QGLViewer.
    //setAnimationPeriod(0);

    _redrawTimer = new QTimer(this);
    connect(_redrawTimer, SIGNAL(timeout()), this, SLOT(updateGL()));
    _redrawTimer->setSingleShot(false);
}

BasicGLViewer::~BasicGLViewer()
{
    saveStateToFile();

    if( _shader  ) delete _shader;
    if( _shader2 ) delete _shader2;

    _redrawTimer->stop();
    delete _redrawTimer;

    delete _scene;
}

//////////////////////////////////////////////////////////////////////////////

bool
BasicGLViewer::reloadShader()
{
    if( ! isVisible() )     // Not setup yet ?
        return false;

    // --- Cleanup any tracing or already loaded shader ----------------------
    if( _debug )        
    {
        delete _debug;
        _debug = 0;
    }
    if( _shader )
    {
        delete _shader;
        _shader = 0;
    }
    if( _shader2 )
    {
        delete _shader2;
        _shader2 = 0;
    }

    // --- Basic Phong shader, to mimic std OpenGL rendering -----------------

    Q_ASSERT( _mode < RENDER_MAX );
    static QString shaders[][3] = {
        "",         "",             "",
        "phong",    "phong",        "phong",
        "common",   "sphereTris",   "sphere",
        "common",   "parabolTris",  "parabol",
        "common",   "sphereBox",    "sphere",
        "common",   "parabolBox",   "parabol",
        "common",   "lensBox",      "lens",
    };

    QString env = _intersectChoice + "\n"
		        + _lensChoice      + "\n"
#if PARABOLIC_ENV_MAP_ARRAY
                + "#define PARABOLIC_ENV_MAP_ARRAY 1\n"
#endif
				+ _scene->env.envString()
                ;


    switch( _subMode )
    {
    case SUBMODE_COVERAGE:
        env += "#define SUBMODE_COVERAGE 1\n";
        break;
    case SUBMODE_DEPTH:
        env += "#define SUBMODE_DEPTH 1\n";
        break;
    case SUBMODE_COLOR:
        env += "#define SUBMODE_COLOR 1\n";
        break;
    case SUBMODE_DEBUG:
        env += "#define SUBMODE_DEBUG 1\n";
        env += "#define PRIMITIVE_FILTERING 1\n";
        break;
    default:
        Q_ASSERT( ! "Unknown sub mode" );
    }

	Shader *newShader = 0, *newShader2 = 0;
    if( _mode != RENDER_OPENGL )
    {
        QString dir = qApp->applicationDirPath() + "/Shaders/";
        if( GLEW_EXT_geometry_shader4 && ! _scene->env.testNoGeomShader )
        {
            int nbOut = (_mode == RENDER_SPHERE_QUAD || _mode == RENDER_PARABOL_QUAD) ? 4 : 12;

            if( _scene->env.twoPass && _mode != RENDER_SPHERE_QUAD && _mode != RENDER_PARABOL_QUAD)
            {
                newShader2 = new GLShader( (dir+shaders[_mode][0]+"_vertex.glsl").toLatin1().constData(),
                                          GL_TRIANGLES, (dir+shaders[_mode][1] + "_geometry.glsl").toLatin1().constData(), GL_TRIANGLE_STRIP,
                                          (dir+shaders[_mode][2] + "_fragment.glsl").toLatin1().constData(),
                                          (env + "#define PASS_2 1\n").toLatin1().constData());
                newShader2->setMaxOutput(9);
                nbOut = 3;
                env += "#define PASS_1 1\n";
            }

            newShader = new GLShader( (dir+shaders[_mode][0]+"_vertex.glsl").toLatin1().constData(),
                                      GL_TRIANGLES, (dir+shaders[_mode][1] + "_geometry.glsl").toLatin1().constData(), GL_TRIANGLE_STRIP,
                                      (dir+shaders[_mode][2] + "_fragment.glsl").toLatin1().constData(),
                                      env.toLatin1().constData());
            newShader->setMaxOutput(nbOut);
        }
        else
        {
            if( ! GLEW_EXT_geometry_shader4 )
                qWarning("Skipping unsupported geometry shader loading");
            newShader = new GLShader( (dir+shaders[_mode][0] + "_vertex.glsl").toLatin1().constData(),
                                      (dir+shaders[_mode][2] + "_fragment.glsl").toLatin1().constData(),
                                  env.toLatin1().constData());
        }

        if( ! newShader || ! newShader->compile()
         || _scene->env.twoPass && (! newShader2 || ! newShader2->compile()))
        {
            displayMessage( ("**** ERROR LOADING SHADER " + shaders[_mode][1]).toLatin1().constData(), 30000 );
			delete newShader;
			delete newShader2;
            return false;
        }
    }

	_shader  = newShader;
	_shader2 = newShader2;

    // Also reload all env-map shader. That recompute them on next drawing...
    if( ! EnvironmentMap::reloadShaders( _scene->env, _subMode) )
        return false;
    _scene->env.envMapsDirty = true;    // Tell Scene to indeed re-render maps.

    if( ! GlossyMap::reloadShaders( _scene->env ) )
        return false;
    _scene->env.glossyMapsDirty = true; // Tell Scene to indeed re-render maps.

    displayMessage( "All shaders reloaded.", 30000 );
    printf("All shaders reloaded...\n");
    return true;
}

void BasicGLViewer::init()
{
    setSnapshotFormat("PNG");
    restoreStateFromFile();     // Retrieve saved camera positions...

    if( qApp->argc() < 2 )
	{
        // _scene->makeBox( Vec3f(0,0,0), Vec3f(1,1,1) );
        ScnObject *m = _scene->makeBox( Vec3f(-2, -1.2f, -4), Vec3f(4, -1.1f, 2));
        m->setColorShininess( 999.0f );

        m = _scene->makeSphere(1.0f, 3);	// Radius 1, cente 0.
        m->setColorShininess( 1000.0f );

        m = _scene->makeSphere(0.45f, 3);
        m->transform( Matrix4::translate(Vec3f(1.50f,1.00f,0.75f)) );
        m->setColorShininess( 1000.0f );
	}
    else
        loadScene( qApp->argv()[1] );

    // Have the default shader/mode loaded.
    reloadShader();

    // ---- Back to QGLViewer's init() stuff :
    resetView();
    QGLViewer::init();

	// Make sure we can view transparent triangles outside bended triangles.
    setBackgroundColor( Qt::black );

	makeCurrent();				// Make sure to have be in the OpenGL context.
    glDepthFunc(GL_LEQUAL);     // Make sure to blend in DEBUG mode, when Z==0.9999 everywhere !
    glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::loadScene(QString fname)
{
    if( fname.isEmpty() )
        fname = QFileDialog::getOpenFileName(this, "Scene to load ?",
        QDir::currentPath(),
        "Any scene (*.obj *.WRL *.VRML);;"
        "Wavefront (*.obj);;"
        "VRML (*.WRL *.VRML);;"
        "Any file (*.*)");
    if( fname.isEmpty() ) return;
    QDir::setCurrent( fname.section('/', 0, -2) );

    _scene->clear();

    mergeScene(fname);
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::mergeScene(QString fname)
{
    if( fname.isEmpty() )
        fname = QFileDialog::getOpenFileName(this, "Scene to load ?",
        QDir::currentPath(),
        "Any scene (*.obj *.WRL *.VRML);;"
        "Wavefront (*.obj);;"
        "VRML (*.WRL *.VRML);;"
        "Any file (*.*)");
    if( fname.isEmpty() ) return;
    QDir::setCurrent( fname.section('/', 0, -2) );

    resetInterface();

    if( fname.endsWith(".obj", Qt::CaseInsensitive) )
        loadOBJ(_scene, fname);
    else
    if( fname.endsWith(".wrl",  Qt::CaseInsensitive)
     || fname.endsWith(".vrml", Qt::CaseInsensitive) )
        loadVRML(_scene, fname);
    else
        qWarning("Unknown scene format %s", 
                 fname.section('.', -2).toLatin1().constData());

	_scene->env.shadowsDirty = true;    // recompute shadows at each load.
	_scene->env.envMapsDirty = true;    // recompute anv maps at each load.
	_scene->env.glossyMapsDirty = true;	// recompute glossy maps at each load.
    resetView();
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::resetView()
{
    Vec3f min, max;
    _scene->getBBox(min, max);
    camera()->setSceneBoundingBox(qglviewer::Vec(min), qglviewer::Vec(max));
    camera()->setOrientation( qglviewer::Quaternion() );
    camera()->setPosition(sceneCenter());
    camera()->showEntireScene();
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::setShader(QAction* a)
{
    RenderingMode newMode = RENDER_SHADERS;

         if( a->text() == "OpenGL 1"      ) newMode = RENDER_OPENGL;
    else if( a->text() == "Shaders"       ) newMode = RENDER_SHADERS;
    else if( a->text() == "Sphere  / Tris") newMode = RENDER_SPHERE_TRIS;
    else if( a->text() == "Parabol / Tris") newMode = RENDER_PARABOL_TRIS;
    else if( a->text() == "Sphere  / Quad") newMode = RENDER_SPHERE_QUAD;
    else if( a->text() == "Parabol / Quad") newMode = RENDER_PARABOL_QUAD;
    else if( a->text() == "Lens    / Quad") newMode = RENDER_LENS_QUAD;
    else
        qWarning("Unknown action '%s'", a->text().toLatin1().constData() );

    // Mode changed ?
    if( _mode != newMode )
    {
        _mode = newMode;
        if( reloadShader() )
            displayMode();
    }
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::setSubmode(QAction* a)
{
         if( a->text() == "Coverage" ) _subMode = SUBMODE_COVERAGE;
    else if( a->text() == "Depth"    ) _subMode = SUBMODE_DEPTH;
    else if( a->text() == "Color"    ) _subMode = SUBMODE_COLOR;
    else if( a->text() == "Shadeless") _subMode = SUBMODE_COLOR;
    else if( a->text() == "DEBUG"    ) _subMode = SUBMODE_DEBUG;
    else
        qWarning("Unknown action '%s'", a->text().toLatin1().constData() );

    // Reset drawing options to a meaningfull state:
    switch( _subMode )
    {
    case SUBMODE_COLOR:     _scene->env.testShading = (a->text() == "Shadeless");
    case SUBMODE_DEPTH:     _scene->env.culling = Environment::BACK_FACE;  _scene->env.depthTest = true;  _scene->env.blending = false; break;
    case SUBMODE_DEBUG:     _scene->env.culling = Environment::NO_CULLING; _scene->env.depthTest = false; _scene->env.blending = true;  break;
    case SUBMODE_COVERAGE:  _scene->env.culling = Environment::BACK_FACE;  _scene->env.depthTest = false; _scene->env.blending = false; break;
    default:
        Q_ASSERT( ! "Unknown submode");
    };

    if( reloadShader() )
        displayMode();
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::setTwoPass(bool on)
{
	_scene->env.twoPass = on ? 1 : 0;
    printf("Two pass %s\n", _scene->env.twoPass ? "ON" : "off");

    if( reloadShader() )     // Lighting is #ifdef'ed into shader.
        displayMode();
    resetDelay();
    update();
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::setIntersect(QAction* a)
{
	_intersectChoice = QString("#define ISECT_%1 1").arg(a->text());
	if( reloadShader() )
		displayMessage(_intersectChoice, 30000);
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::setLens(QAction* a)
{
	_lensChoice = QString("#define LENS_%1 1")
        .arg(a->text().section(' ', 0, 0));     // Cut at first space.
	if( reloadShader() )
		displayMessage(_lensChoice, 30000);
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::setEnvmapMode(QAction* a)
{
         if( a->text() == "    Faces"    ) _scene->env.envMapMode = EnvironmentMap::FACES_MAP;
    else if( a->text() == "    Parabola" ) _scene->env.envMapMode = EnvironmentMap::PARABOLA_MAP;
    else if( a->text() == "    Sphere"   ) _scene->env.envMapMode = EnvironmentMap::SPHERE_MAP;
    else if( a->text() == "    Lambert"  ) _scene->env.envMapMode = EnvironmentMap::LAMBERT_MAP;
    else
        qFatal( "Unrecognized envmap mode action" );

    if( reloadShader() )
		displayMessage(_lensChoice, 30000);
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::setLights()
{
    int n = (_scene->env.nbLights >= 5) ? 0
          : (_scene->env.nbLights == 0) ? 1
          : _scene->env.nbLights<<1
          ;
    setLights(n);
}

void BasicGLViewer::setLights(int n)
{
    Q_ASSERT( 0 <= n && n <= 8);
    _scene->env.nbLights = n;
    if( _scene->env.nbLights )
        printf("Lighting per pixel, with %d lights\n", _scene->env.nbLights);
    else
        printf("Lighting per vertex\n");

    if( reloadShader() )     // Lighting is #ifdef'ed into shader.
        displayMode();
    resetDelay();
    update();
}

////////////////////////////////////////////////////////////////////////

void BasicGLViewer::setShadows(bool on)
{
	_scene->env.nbShadows = on ? 1 : 0;
    printf("Shadows %s\n", _scene->env.nbShadows ? "ON" : "off");

    if( reloadShader() )     // Lighting is #ifdef'ed into shader.
        displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setDynamicShadows(bool on)
{
	_scene->env.dynamicShadows = on ? 1 : 0;
    printf("Shadows %s\n", _scene->env.dynamicShadows ? "DYNAMIC" : "static");

    if( reloadShader() )     // Lighting is #ifdef'ed into shader.
        displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setTextures(bool on)
{
	_scene->env.colorMaps = on;
    printf("Textures %s\n", _scene->env.colorMaps ? "ON" : "off");

    if( reloadShader() )     // Lighting is #ifdef'ed into shader.
        displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setBumps(bool on)
{
	_scene->env.normalMaps = on;
    printf("Normal maps %s\n", _scene->env.normalMaps ? "ON" : "off");

    if( reloadShader() )     // Lighting is #ifdef'ed into shader.
        displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setEnvMaps(bool on)
{
	_scene->env.envMaps = on;
    printf("Environment maps %s\n", _scene->env.envMaps ? "ON" : "off");

    if( reloadShader() )     // Lighting is #ifdef'ed into shader.
        displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setDynamicEnvMaps(bool on)
{
	_scene->env.dynamicEnvMap = on;
    printf("Environment maps %s\n", _scene->env.dynamicEnvMap ? "DYNAMIC" : "static");

    displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setGlossyMaps(bool on)
{
	_scene->env.glossyMaps = on;
    printf("Glossy maps %s\n", _scene->env.glossyMaps ? "ON" : "off");

    if( reloadShader() )     // Lighting is #ifdef'ed into shader.
        displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setDynamicGlossyMaps(bool on)
{
	_scene->env.dynamicGlossyMaps = on;
    printf("Glossy maps %s\n", _scene->env.dynamicGlossyMaps ? "DYNAMIC" : "static");

    displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setWireframe(bool on)
{
    _scene->env.wireframe = on ? Environment::RENDER_EDGES : Environment::RENDER_FACES;
    resetDelay();
    update();
    displayMode();
}

void BasicGLViewer::setScribe(bool on)
{
    _scene->env.wireframe = on ? Environment::RENDER_BOTH : Environment::RENDER_FACES;
    resetDelay();
    update();
    displayMode();
}

void BasicGLViewer::setCulling(bool on)
{
    _scene->env.culling = on ? Environment::BACK_FACE : Environment::NO_CULLING;
    printf("Face culling %s\n", _scene->env.culling != Environment::NO_CULLING ? "ON" : "off");

    if( reloadShader() )
        displayMode();
    resetDelay();
    update();
}

void BasicGLViewer::setDepthTest(bool on)
{
    _scene->env.depthTest = on;
    printf("Depth test %s\n", _scene->env.depthTest ? "ON" : "off");

    resetDelay();
    displayMode();
    update();
}

void BasicGLViewer::setBlending(bool on)
{
    _scene->env.blending = on;
    printf("Blending %s\n", _scene->env.blending ? "ON" : "off");

    resetDelay();
    displayMode();
    update();
}

void BasicGLViewer::setStatistics(bool on)
{
    _statMode = on;
    if( on )
    {
        resetDelay();
        _redrawTimer->start(1); // 1KHz max redraw speed.
    }
    else
    {
        _redrawTimer->stop();
        delayString.clear();
    }

    displayMode();
    update();
}

void BasicGLViewer::setNoFragments(bool on)
{
    _scene->env.testNoFrags = on;
    if( _scene->env.testNoFrags )
        printf("KILLING ALL FRAGMENT PROCESSING\n");
    else
        printf("Back to normal fragment processing\n");

    resetDelay();
    if( reloadShader() )
        displayMode();
    update();
}


void BasicGLViewer::setNoGeometryShader(bool on)
{
    _scene->env.testNoGeomShader = on;
    if( _scene->env.testNoGeomShader )
        printf("SKIPPING GEOMETRY PROCESSING\n");
    else
        printf("Back to normal geometry mode\n");

    resetDelay();
    if( reloadShader() )
        displayMode();
    update();
}

void BasicGLViewer::displayMode()
{    
    QString message = _mode == RENDER_OPENGL        ? "OpenGL      "
                    : _mode == RENDER_SHADERS       ? "Shaders     "
                    : _mode == RENDER_SPHERE_TRIS   ? "Sphere  Tris"
                    : _mode == RENDER_PARABOL_TRIS  ? "Parabol Tris"
                    : _mode == RENDER_LENS_QUAD     ? "Lens    Quad"
                    : _mode == RENDER_SPHERE_QUAD   ? "Sphere  Quad"
                    : _mode == RENDER_PARABOL_QUAD  ? "Parabol Quad"
                    :                                 "????        ";

    message += _subMode == SUBMODE_COVERAGE  ? " / coverag"
             : _subMode == SUBMODE_DEPTH     ? " / depth  "
             : _subMode == SUBMODE_COLOR     ? " / color  "
             : _subMode == SUBMODE_DEBUG     ? " / DEBUG  "
             :                                 " / ?????  ";
    
    message += _scene->env.nbLights ? QString(" /  %1 lights").arg(_scene->env.nbLights) : QString(" / no lights");
    
    message += (_scene->env.culling == Environment::BACK_FACE)  ? " / Back  cull"
		     : (_scene->env.culling == Environment::FRONT_FACE) ? " / Front cull" 
		     :                                                    " / Two sides ";
    
    message += (_scene->env.wireframe == Environment::RENDER_EDGES) ? " / Wire  " 
             : (_scene->env.wireframe == Environment::RENDER_BOTH ) ? " / Scribe" 
             :                                                        "";

    message += _scene->env.testNoGeomShader ? " / SKIP GEOMETRY " : "";
    message += _scene->env.testNoFrags      ? " / SKIP FRAGMENTS" : "";
    message += _scene->env.testShading      ? " / SKIP SHADING" : "";

    message += _scene->env.depthTest ? "" : " / no Z-test";
    message += _scene->env.blending  ? " / Blending" : "";
    message += _scene->env.nbShadows ? 
               _scene->env.dynamicShadows ? " / DYN shadows" : " / shadows" : "";
    if( _scene->env.envMaps )
    {
        QString name;
        if( _scene->env.dynamicEnvMap )    name += "DYN ";
        switch( _scene->env.envMapMode )
        {
        case EnvironmentMap::FACES_MAP:    name += "flat "; break;
        case EnvironmentMap::PARABOLA_MAP: name += "prabola "; break;
        case EnvironmentMap::SPHERE_MAP:   name += "sphere ";  break;  // NYI
        case EnvironmentMap::LAMBERT_MAP:  name += "lambert "; break;  // NYI
        }
        message +=  " / " + name + "envmaps";
    }

    message += _scene->env.normalMaps? " / normal map" : "";

    displayMessage(message, 30000);
}

////////////////////////////////////////////////////////////////////////

// redraw without bells and wistles, for FBO or statistics.
void
BasicGLViewer::fastDraw()
{
    // Maxdistance from a 3D off-centered sphere ?
    float zFar  = camera()->distanceToSceneCenter() + 1.1f * camera()->sceneRadius();
    float zNear = camera()->distanceToSceneCenter() - 1.1f * camera()->sceneRadius();
    if( zNear < 0.01f ) zNear = 0.01f;

    // Non linear modes all have a 0 near plane :
    if( _mode != RENDER_OPENGL && _mode != RENDER_SHADERS )
    {
        camera()->setZClippingCoefficient(2.0f);
        camera()->setZNearCoefficient( 0.0f );
    }
    else
    {
        camera()->setZClippingCoefficient(1.1f);
        camera()->setZNearCoefficient( 0.01f );
    }

    // ---- Pre-compute static shadow/environment/... maps -------------------
    _scene->env.prepare(_shader, _scene, _subMode);

	// ---- Scene render -----------------------------------------------------
    _scene->draw(_shader, zNear, zFar);
    if( _scene->env.twoPass )
        _scene->draw(_shader2, zNear, zFar);
}

//////////////////////////////////////////////////////////////////////////////
void BasicGLViewer::resizeGL(int width, int height)
{
  QGLWidget::resizeGL(width, height);
  glViewport(0, 0, width, height);

  // Select the bigger square, when in 360deg shader mode.
  _width  = width;
  _height = height;
  int dw = 0, dh = 0;
  if( _mode != RENDER_OPENGL && _mode != RENDER_SHADERS )
  {
      _width = _height = (width < height) ? width : height;
      dw = (width - _width)/2;
      dh = (height - _height)/2;
  }
  glViewport( dw, dh, _width, _height );

  camera()->setScreenWidthAndHeight(_width, _height);

  printf("Viewport size = %d x %d\n", _width, _height);
  displayMessage( QString("Viewport size = %1 x %2").arg(_width).arg(_height), 30000 );
}

//////////////////////////////////////////////////////////////////////////////
void BasicGLViewer::draw()
{
    if( _statMode )
    {
        double renderTime = 0.0f;
        int    renderPass = 0;
        OcclusionQuery with, without;

        //---- Measure overdraw statistics -----------------------------------
        with.start();
            _scene->env.testOverdraw = true;
            _scene->env.depthTest       = false;
            fastDraw();
        with.stop();
        without.start();
            _scene->env.testOverdraw = false;
            fastDraw();
        without.stop();
        _scene->env.depthTest = true;

        //---- Measure averaged draw statistics ------------------------------
        glFinish();                             // Make sure pipeline is free.
        double renderStart = TBender::time();
        do {                                    // Average at least 1/4s.
            renderPass++;
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
            fastDraw();
            glFinish();                         // Make sure OpenGL finished drawing.
            renderTime = TBender::time() - renderStart;
        } while( renderTime < 0.250f && renderPass < 1000 );

        smoothedDelay += renderTime;
        smoothedCount += renderPass;
        smoothedFragmentsGenerated += renderPass * with.count();
        smoothedFragmentsDrawn     += renderPass * without.count();
    }
    else
        fastDraw();
    CHECK_GL_ERROR("draw scene");

    // ---- Add nice horizon circle if in non-linear mode --------------------
    if( _mode != RENDER_OPENGL && _mode != RENDER_SHADERS )
    {
        // Draw horizon (ellipse)
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glColor3f(0,0,0);   // black
        glLineWidth(3.5);
        glBegin(GL_LINES);
            glVertex3f(-0.05f, 0.0f, -1.0f);
            glVertex3f( 0.05f, 0.0f, -1.0f);
            glVertex3f(0.0f, -0.05f, -1.0f);
            glVertex3f(0.0f,  0.05f, -1.0f);
        glEnd();

        glColor3f(1,1,0);   // Yellow
        glLineWidth(1.5);
        glBegin(GL_LINES);
            for(float a=0.0f; a<=6.28; a+= 0.04f)
                glVertex3f(sinf(a), cosf(a), -1.0f);
            glVertex3f(-0.05f, 0.0f, -1.0f);
            glVertex3f( 0.05f, 0.0f, -1.0f);
            glVertex3f(0.0f, -0.05f, -1.0f);
            glVertex3f(0.0f,  0.05f, -1.0f);
        glEnd();
        CHECK_GL_ERROR("draw horizon");
    }

    // ---- use basic OpenGL text paint to insert render speed ---------------
    if( smoothedDelay > 1.0 || smoothedCount > 4000 )
    {
        float delay = smoothedDelay / smoothedCount;

        float hertz = 1.0f / delay;
        QString hz = "Hz ";
        if( hertz > 1000 ) hertz /= 1000, hz = "KHz";
        if( hertz > 1000 ) hertz /= 1000, hz = "MHz";

        float vertices = _scene->vertices() / delay;
        QString vps = "V/s ";
        if( vertices > 1000 ) vertices /= 1024.0f, vps = "KV/s";
        if( vertices > 1000 ) vertices /= 1024.0f, vps = "MV/s";

        float triangles = _scene->triangles() / delay;
        QString tps = "T/s ";
        if( triangles > 1000 ) triangles /= 1024.0f, tps = "KT/s";
        if( triangles > 1000 ) triangles /= 1024.0f, tps = "MT/s";

        float overdraw = 0.0f;
        if( smoothedFragmentsDrawn > 0 )
            overdraw = 100.0f * float(smoothedFragmentsGenerated - smoothedFragmentsDrawn) 
                              / float(smoothedFragmentsDrawn);

        delayString = QString("%1 = %2 %3 = %4 %5 %6 %7")
                      .arg(TBender::timeCString(delay), 9)
                      .arg(hertz,     6, 'f', 2).arg(hz)
                      .arg(vertices,  6, 'f', 2).arg(vps)
                      .arg(triangles, 6, 'f', 2).arg(tps);

        //---- Printout ------------------------------------------------------

        printf("\n");
        printf("*** %s (avg on %d)\n",
            delayString.toLatin1().constData(), smoothedCount);
        printf("Fragment generated %ld K, drawn %ld K, overdraw %5.1f %%\n",
            int(0.5f + 0.001f * smoothedFragmentsGenerated / smoothedCount),
            int(0.5f + 0.001f * smoothedFragmentsDrawn / smoothedCount),
            overdraw);

        resetDelay();
    }

    if( ! delayString.isEmpty() )
    {
        // Draw horizon (ellipse)
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();

        glColor3f(1.0f, 1.0f, 1.0f);
        drawText(15, 15, delayString);
        CHECK_GL_ERROR("draw fps");
    }
}

void BasicGLViewer::keyPressEvent(QKeyEvent *e)
{
    if( e->key() == Qt::Key_V && e->modifiers() == Qt::ControlModifier )
    {
        if( _debug )
        {
            delete _debug; 
            _debug = 0; 
        } else {
            int prog = _shader->prog();
            _debug = new Feedback(prog, GL_TRIANGLES);
            Q_CHECK_PTR(_debug);
        }
    }
    else if( e->key() == Qt::Key_PageUp )
    {
        _scene->env.primitiveFilter = (_scene->env.primitiveFilter+1)%_scene->triangles();
        qWarning("Displaying only triangle %d", _scene->env.primitiveFilter);
        update();
    }
    else if( e->key() == Qt::Key_PageDown )
    {
        int tris = _scene->triangles();
        _scene->env.primitiveFilter = (_scene->env.primitiveFilter+tris-1)%tris;
        qWarning("Displaying only triangle %d", _scene->env.primitiveFilter);
        update();
    }
    else if( e->key() == Qt::Key_Home )
    {
        _scene->env.primitiveFilter = -1;
        qWarning("Displaying all triangles");
        update();
    }
    else if( e->key() == Qt::Key_R )
    {
        reloadShader();
        update();
    }
    else if( e->key() == Qt::Key_O )
    {
        makeCurrent();  // make sure to have an openGL context.

        {
            QImage img(512, 512, QImage::Format_ARGB32);
            FrameBufferObject fbo;
            fbo.bindColorImage(img.width(), img.height(), _scene->env.depthTest);
            preDraw();  // clear depth *once* fbo is bound...

            // Average timming for very light renderings (< 100ms) :
            double t = TBender::time();
            int n = 0;

            do {
                ++n;
                fastDraw();
                glFinish();
            } while( _statMode && (TBender::time() - t < 0.100) );

            if( _statMode )
                qWarning("%dx%d image render in %g ms [mean on %d rendering]",
                    img.width(), img.height(),
                    float(1000.0*(TBender::time()-t)/n), n);

            // Get image back, save, and print.
            if( fbo.finish(img) )
            {
                img.save("snapshot.png");
                qWarning("snapshot.png saved...");
            }
        }

        {
            QImage img(512, 512, QImage::Format_Indexed8);
            ShadowMap map;
            map.bindDepthTexture(img.width(), img.height());
            map.setupLightMatrix( 10.0f * Vec3f(_scene->env.lights[0].position), Vec3f(0,0,0), 10.0f);
            fastDraw();
            map.cleanMatrix();
            if( map.finish(img) )
            {
                img.save("snadepth.png");
                qWarning("snadepth.png saved...");
            }
        }
    }

    else
        QGLViewer::keyPressEvent(e);
}

void BasicGLViewer::mouseDoubleClickEvent(QMouseEvent *e)
{
    if( e->buttons() == Qt::LeftButton )
        resetView();
    else if( e->buttons() == Qt::RightButton )
    {
        Vec3f n = Vec3f( camera()->viewDirection() );
        Vec3f u = Vec3f( camera()->upVector() );

        // Keep the dominant value, and normalize :
        int d = dominant(n);
        n[ (d+1) % 3 ]  = n[ (d+2)% 3] = 0.0f;
        n = normalize(n);
        d = dominant(u);
        u[ (d+1) % 3 ]  = u[ (d+2)% 3] = 0.0f;
        u = normalize(u);

        // Find an Up vector :
        camera()->setViewDirection( (qglviewer::Vec)n);
        camera()->setUpVector     ( (qglviewer::Vec)u);
        camera()->showEntireScene();
    }
    else
        QGLViewer::mouseDoubleClickEvent(e);
}
