////////////////////////////////////////////////////////////////////////
/*! \file   BasicGLViewer.h
 *  \brief  Yet another basic GL Viewer.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 */
////////////////////////////////////////////////////////////////////////
// $Id: BasicGLViewer.h,v 1.50 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2006/11/29] : Creation.

#include <GL/glew.h>
#include "QGLViewer/qglviewer.h"

class Scene;
class Light;
class QAction;
class Shader;
class Feedback;

class BasicGLViewer : public QGLViewer
{
public:
    enum RenderingMode {
        RENDER_OPENGL = 0,
        RENDER_SHADERS,
        RENDER_SPHERE_TRIS,
        RENDER_PARABOL_TRIS,

        RENDER_SPHERE_QUAD,
        RENDER_PARABOL_QUAD,
        RENDER_LENS_QUAD,

        RENDER_MAX
    };
    enum SubMode {
        SUBMODE_COVERAGE=0,
        SUBMODE_DEPTH,
        SUBMODE_COLOR,
        SUBMODE_DEBUG,
        SUBMODE_MAX
    };

private:
    Q_OBJECT
    Scene*          _scene;         //!< Private scene to draw.
    QTimer*         _redrawTimer;   //!< Statistic 1ms auto redraw event.

    RenderingMode   _mode;          //!< OpenGL, Phong, spherical, etc.
    SubMode         _subMode;       //!< What does the current shader draw ?
	QString			_intersectChoice; //!< #define to choose intersection code.
	QString			_lensChoice;	//!< #define to choose between various lens formulas.
    Shader*         _shader;        //!< Currently loaded shader.
    Shader*         _shader2;       //!< Currently loaded second pass shader, if any.
    Feedback*       _debug;         //!< Varying tracer for debug purpose.
    int             _width;         //!< Viewport's width and height. This is 
    int             _height;        //!< different from width(), height() when we enforce square viewport.

    bool            _statMode;      //!< Render several time to accumulate statistics.

    void fastDraw();                //!< Redraw scene, without statistics or additional display.

public:
    BasicGLViewer(QWidget* parent = 0);
    virtual ~BasicGLViewer();

protected :
    virtual void init();                //!< Called once on first draw.
    virtual void draw();                //!< Scene plus debug data.
    virtual void resizeGL(int, int);    //!< Addapt viewport to window size.

    //! Map double-click events to resetView().
    virtual void mouseDoubleClickEvent(QMouseEvent *);

	//! Map keys to debug functions.
    virtual void keyPressEvent(QKeyEvent *e);

public slots:
    void loadScene(QString fname = QString::null);
    void mergeScene(QString fname = QString::null);

    void resetView();

    bool reloadShader();                //!< Reload currently used shader. Returns true if successfull.

    void setShader(QAction*);           //!< Fired from QActionGroup::trigered().
    void setSubmode(QAction*);          //!< Fired from QActionGroup::trigered().
    void setTwoPass(bool on);           //!< Toggle second shading pass.

    void setLights();                   //!< Switch light number (0,1,2,4,8 sequence).
    void setLights(int);                //!< Set light number (0..8).
	void setShadows(bool on);		    //!< Switch number of shadow map to use (0,1).
	void setDynamicShadows(bool on);    //!< Recompute shadows at each frame ?
	void setTextures(bool on);		    //!< Switch color maps on/off.
	void setBumps(bool on);			    //!< Switch normal maps on/off.
    void setEnvMaps(bool on);           //!< Switch env maps on/off.
	void setDynamicEnvMaps(bool on);    //!< Recompute envmaps at each frame ?
	void setGlossyMaps(bool on);        //!< Enable glossy maps
	void setDynamicGlossyMaps(bool on); //!< Recompute glossy maps at each frame ?
    void setWireframe(bool on);
    void setScribe(bool on);
	void setCulling(bool on);
	void setDepthTest(bool on);
	void setBlending(bool on);
    void setStatistics(bool on);
    void setNoFragments(bool on);       //!< Use glScissor() to avoid all fragments.
    void setNoGeometryShader(bool on);  //!< Don't load geometry shader.

	void setIntersect(QAction *);	    //!< Choose #define for intersection code.
    void setLens(QAction *);            //!< Choose #define for lens formula.
    void setEnvmapMode(QAction *);      //!< Change envmap mode (cubemap, parabola, sphere, ...)

    void displayMode();
};

extern void resetInterface();