////////////////////////////////////////////////////////////////////////
/*! \file   CgGLShader.cpp
 *  \brief  Simple class to run Cg shaders on OpenGL API.
 *  \author J.D. Gascuel
 *
 */
////////////////////////////////////////////////////////////////////////
// $Id: CgShader.cpp,v 1.4 2006/12/06 12:21:19 gascuel Exp $
// File History:
// [jDG 2006/12/03] : Creation.

#include "Shader.h"

#include <GL/glew.h>

#include <Cg/cg.h>
#include <Cg/cgGL.h>
#include <QtCore>       // For Q_CHECK_PTR / Q_ASSERT

#ifdef OPENGL
#   ifdef ARB
#       define   VERTEX_SHADER_PROFILE CG_PROFILE_ARBVP1
#       define FRAGMENT_SHADER_PROFILE CG_PROFILE_ARBFP1
#       define            PROFILE_NAME "ARB1"
#   elif defined(NV2X)
#       define   VERTEX_SHADER_PROFILE CG_PROFILE_VP20
#       define FRAGMENT_SHADER_PROFILE CG_PROFILE_FP20
#       define            PROFILE_NAME "NV2X"
#   elif defined(NV3X)
#       define   VERTEX_SHADER_PROFILE CG_PROFILE_VP30
#       define FRAGMENT_SHADER_PROFILE CG_PROFILE_FP30
#       define            PROFILE_NAME "NV3X"
#   elif defined(NV4X)
#       define   VERTEX_SHADER_PROFILE CG_PROFILE_VP40
#       define FRAGMENT_SHADER_PROFILE CG_PROFILE_FP40
#       define            PROFILE_NAME "NV4X"
#   elif defined(GLSL)
#       define   VERTEX_SHADER_PROFILE CG_PROFILE_GLSLV
#       define FRAGMENT_SHADER_PROFILE CG_PROFILE_GLSLF
#       define            PROFILE_NAME "GLSL"
#   endif
#endif

////////////////////////////////////////////////////////////////////////

class CgShaderPrivate
{
public:
    const char *vertexFile;
    const char *fragmentFile;
    CGcontext context;
    CGprogram vs;
    CGprogram fs;
};

static bool
printLog(const char *op, CGcontext context)
{
    if( CGerror error = cgGetError() )
    {
        const char* errorString = cgGetErrorString(error);
        qWarning("*** Cg %s error (%s): %s",
                 op, 
                 PROFILE_NAME,
                 errorString);

        if( const char* lst = cgGetLastListing(context) )
            qWarning("*** \t%s\n", lst);

        return false;
    }
    return true;
}

////////////////////////////////////////////////////////////////////////

CgShader::CgShader(const char *vertexFile,
                   const char *fragmentFile)
{
    if( GLEW_EXT_Cg_shader )
        qDebug("OpenGL GL_EXT_Cg_shader ok.");
    else
        qWarning("*** Missing Cg support extension : GL_EXT_Cg_shader.");

    p = new CgShaderPrivate;
    Q_CHECK_PTR(p);
    p->context = 0;
    p->vs = p->fs = 0;
    p->vertexFile   = vertexFile;
    p->fragmentFile = fragmentFile;
}

//////////////////////////////////////////////////////////////////////////////

CgShader::~CgShader()
{
    if( p->context )
    {
        disable();

        if( p->vs ) cgDestroyProgram(p->vs);
        if( p->fs ) cgDestroyProgram(p->fs);
        cgDestroyContext(p->context);

        qDebug("Cg Shader deleted");
    }

    delete p;
    p = 0;
}

//////////////////////////////////////////////////////////////////////////////

bool CgShader::compile()
{
    Q_ASSERT(p);

    p->context = cgCreateContext();

    const char *args[] = {0};

    if( p->vertexFile )
    {
        p->vs = cgCreateProgramFromFile(p->context,
                                        CG_SOURCE, p->vertexFile,
                                        VERTEX_SHADER_PROFILE, "main", args);
        if( ! printLog("created vertex shader", p->context) )
            return false;

        cgGLLoadProgram(p->vs);
        if( ! printLog("load vertex shader", p->context) )
            return false;

        qDebug("Cg vertex shader (%s) %s compiled.", PROFILE_NAME, p->vertexFile);
    }

    if( p->vertexFile )
    {
        p->fs = cgCreateProgramFromFile(p->context,
                                        CG_SOURCE, p->fragmentFile,
                                        FRAGMENT_SHADER_PROFILE, "main", args);
        if( ! printLog("create fragment shader", p->context) )
            return false;

        cgGLLoadProgram(p->fs);
        if( ! printLog("load fragment shader", p->context) )
            return false;

        qDebug("Cg fragment shader (%s) %s compiled.", PROFILE_NAME, p->fragmentFile);
    }
    
    return true;
}

//////////////////////////////////////////////////////////////////////////////

void CgShader::enable()
{
    if( ! p->context )
    {
        if( !compile() )
            return;
        Q_ASSERT( p->context );
    }

#ifdef OPENGL
    if( p->vs )
    {
        cgGLEnableProfile(VERTEX_SHADER_PROFILE);
        cgGLBindProgram(p->vs);
        (void) printLog("Enable Cg vertex shader", p->context);
    }
    if( p->fs )
    {
        cgGLEnableProfile(FRAGMENT_SHADER_PROFILE);
        cgGLBindProgram(p->fs);
        (void) printLog("Enable Cg fragment shader", p->context);
    }
#else
#error DirectX Not defined yet...
#endif
}

//////////////////////////////////////////////////////////////////////////////

void CgShader::disable()
{
#ifdef OPENGL
    if( p->vs )
        cgGLDisableProfile(VERTEX_SHADER_PROFILE);

    if( p->fs )
        cgGLDisableProfile(FRAGMENT_SHADER_PROFILE);
#else
#error DirectX Not defined yet...
#endif

    (void) printLog("Disable Cg vertex/fragment shader", p->context);
}
