////////////////////////////////////////////////////////////////////////
/*! \file   EnvironmentMap.cpp
 *  \brief  EnvironmentMap for an OpenGL rendering.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Scene, Shader, FrameBufferObject.
 */
////////////////////////////////////////////////////////////////////////
// $Id: EnvironmentMap.cpp,v 1.41 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2007/04/07] : Creation.

#include "EnvironmentMap.h"

#include <GL/glew.h>

#include <QVector>
#include <QApplication>

#include "Math/Matrix4.h"

#include "Utilities.h"
#include "Shader.h"
#include "Scene.h"
#include "BasicGLViewer.h"      // Just for SubMode definitions

typedef QVector<EnvironmentMap*> Maps;
static Maps allMaps;

static float smoothedRenderTime = 0.0f;
static int   smoothedCount = 0;

#if PARABOLIC_ENV_MAP_ARRAY
#   define  TEXTURE_2D_PARABOLIC    GL_TEXTURE_2D_ARRAY_EXT
#   define  SAMPLER_2D_PARABOLIC    GL_SAMPLER_2D_ARRAY_EXT
#else
#   define  TEXTURE_2D_PARABOLIC    GL_TEXTURE_2D
#   define  SAMPLER_2D_PARABOLIC    GL_SAMPLER_2D
#endif

////////////////////////////////////////////////////////////////////////

Shader* EnvironmentMap::phongShader = 0;
Shader* EnvironmentMap::sphereShader = 0;
Shader* EnvironmentMap::parabolicShader = 0;
Shader* EnvironmentMap::lambertShader = 0;

EnvironmentMap::EnvironmentMap(Mode mode, const Vec3f& eye, int size, float radius)
: _mode(mode),
  _eye(eye),
  _size(size),
  _zNear(radius),
  _zFar(-1.0f),
  _texId(0),
  _currentRender(false)
{
    // Make sure the image is correctly initialized to hold our depth
    // data.
    static int maxRBSize = 0;
    if( ! maxRBSize )
    {
        glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &maxRBSize);
        qDebug("Max renderbuffer size is %d", maxRBSize);
    }
    Q_ASSERT( 0 < size  && size  < maxRBSize );

    allMaps.push_back(this);
}

EnvironmentMap::~EnvironmentMap()
{
    int idx = allMaps.indexOf(this);
    if( idx >= 0 )
        allMaps.remove(idx);

    if( _texId ) glDeleteTextures(1, &_texId);
}

////////////////////////////////////////////////////////////////////////

bool
EnvironmentMap::reloadShaders(const Environment& env, int subMode)
{
    if( ! GLEW_EXT_geometry_shader4
     || ! GLEW_EXT_texture_array )
    {
        static bool once = true;
        if( once )
            qWarning("**** Cannot do parabolic envmap without missing \n"
                     "     GL_EXT_geometry_shader4 or \n"
                     "     GL_EXT_texture_array\n"
                     "     OpenGL extensions");
        once = false;
    }

    QString envString = "#define LENS_2 1\n";

    switch( subMode )
    {
    case BasicGLViewer::SUBMODE_COVERAGE:
        envString += "#define SUBMODE_COVERAGE 1\n";
        break;
    case BasicGLViewer::SUBMODE_DEPTH:
        envString += "#define SUBMODE_DEPTH 1\n";
        break;
    case BasicGLViewer::SUBMODE_COLOR:
        envString += "#define SUBMODE_COLOR 1\n";
        break;
    case BasicGLViewer::SUBMODE_DEBUG:
        envString += "#define SUBMODE_DEBUG 1\n";
        envString += "#define PRIMITIVE_FILTERING 1\n";
        break;
    default:
        Q_ASSERT( ! "Unknown sub mode" );
    }

#if PARABOLIC_ENV_MAP_ARRAY
    envString += "#define PARABOLIC_ENV_MAP_ARRAY 1\n";
#endif
    envString += env.envString();

    QString dir = qApp->applicationDirPath() + "/Shaders/";
    //------------------------------------------------------------------------
    // Use a full featured phong shader to :
    //  - Render scene to construct flat cube-map faces.
    //  - Apply parabolic env map to objects w/o special shaders...
    if( phongShader ) delete phongShader;
    phongShader = new GLShader(
        (dir + "phong_vertex.glsl").toLatin1().constData(),
        GL_TRIANGLES, (dir + "phong_geometry.glsl").toLatin1().constData(), GL_TRIANGLE_STRIP,
        (dir + "phong_fragment.glsl").toLatin1().constData(),
        envString.toLatin1().constData());

    if( ! phongShader->compile() )
    {
        delete phongShader;
        phongShader = 0;
        return false;
    }

    //------------------------------------------------------------------------
    if( sphereShader ) delete sphereShader;
    sphereShader = new GLShader(
        (dir + "common_vertex.glsl").toLatin1().constData(),
        GL_TRIANGLES, (dir + "sphereBox_geometry.glsl").toLatin1().constData(), GL_TRIANGLE_STRIP,
        (dir + "sphere_fragment.glsl").toLatin1().constData(),
        envString.toLatin1().constData());
    sphereShader->setMaxOutput(4);

    if( ! sphereShader->compile() )
    {
        delete sphereShader;
        sphereShader = 0;
        return false;
    }

    //------------------------------------------------------------------------
    if( parabolicShader ) delete parabolicShader;
    parabolicShader = new GLShader(
        (dir + "common_vertex.glsl").toLatin1().constData(),
        GL_TRIANGLES, (dir + "parabolBox_geometry.glsl").toLatin1().constData(), GL_TRIANGLE_STRIP,
        (dir + "parabol_fragment.glsl").toLatin1().constData(),
        envString.toLatin1().constData());
    parabolicShader->setMaxOutput(4);

    if( ! parabolicShader->compile() )
    {
        delete parabolicShader;
        parabolicShader = 0;
        return false;
    }

    //------------------------------------------------------------------------
    if( lambertShader ) delete lambertShader;
    lambertShader = new GLShader(
        (dir + "common_vertex.glsl").toLatin1().constData(),
        GL_TRIANGLES, (dir + "lensBox_geometry.glsl").toLatin1().constData(), GL_TRIANGLE_STRIP,
        (dir + "lens_fragment.glsl").toLatin1().constData(),
        envString.toLatin1().constData());
    lambertShader->setMaxOutput(4);

    if( ! lambertShader->compile() )
    {
        delete lambertShader;
        lambertShader = 0;
        return false;
    }

    //------------------------------------------------------------------------
    // Mark all env maps to be re-rendered later on.
    for(Maps::const_iterator it(allMaps.begin()); it != allMaps.end(); ++it)
        (*it)->renderLater( *(Vec3f*)0, EnvironmentMap::Mode(env.envMapMode));

    return true;
}

//////////////////////////////////////////////////////////////////////////////

bool 
EnvironmentMap::render(const Shader* shader, const Scene* scn, int subMode)
{
    // Sanity check, before crashing on null function pointers...
    if( ! GLEW_EXT_framebuffer_object )
    {
        qWarning("*** Missing GL_EXT_framebuffer_object");
        return false;
    }

    if( ! phongShader )
        reloadShaders( scn->env, subMode);

    // ---- Lookup for near/far on first rendering ---------------------------
    if( _zFar <= 0.0f )
    {
        Vec3f bmin, bmax;
        scn->getBBox(bmin, bmax);
        float r1 = length(_eye-bmin);
        float r2 = length(_eye-bmax);
        _zFar = (r1 > r2) ? r1 : r2;
    }

    createFBO();
    createTexture();

    // ---- Common setup -----------------------------------------------------

    glPushAttrib(GL_VIEWPORT_BIT);
    glViewport(0, 0, _size, _size);   // Origin & w*h

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    CHECK_GL_ERROR("push matrices");

    bool ok = true;
    if( _mode == FACES_MAP )
        ok &= renderToCubeMap(shader, scn, subMode);
    else
        ok &= renderToParaboloid(shader, scn, subMode);

    // ---- Restore previous environment ---------------------------------
    glClearColor( 0.0, 0.0, 0.0, 1.0);
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glPopAttrib();                                  // pop viewport.

    return ok;
}

////////////////////////////////////////////////////////////////////////

int
EnvironmentMap::faces()
{
    if( _mode == FACES_MAP )
        return 6;
    else if(_mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP)
        return 2;
    else
        qFatal("Wrong faces in EnvironmentMap");
    return 0;
}

////////////////////////////////////////////////////////////////////////
void
EnvironmentMap::createFBO()
{
#if ! USE_RENDER_TO_ENV_MAPS
    if( ! _rbId )
    {
        glGenRenderbuffersEXT(1, &_rbId);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, _rbId);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                                 GL_RGBA32F_ARB,
                                 _size, _size);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
        CHECK_GL_ERROR("alloc color render");
    }
#endif

    if( ! _zId )
    {
        glGenRenderbuffersEXT(1, &_zId);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, _zId);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                                 GL_DEPTH_COMPONENT24,
                                 _size, _size);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
        CHECK_GL_ERROR("alloc depth render");
    }

    if( ! _id )
    {
	    glGenFramebuffersEXT(1, &_id);
        CHECK_GL_ERROR("alloc framebuffer");

        // Attach color+depth renderbuffers to our FBO :
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
#if ! USE_RENDER_TO_ENV_MAPS
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
                                     GL_COLOR_ATTACHMENT0_EXT,
                                     GL_RENDERBUFFER_EXT,
                                     _rbId);
#endif
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
                                     GL_DEPTH_ATTACHMENT_EXT,
                                     GL_RENDERBUFFER_EXT,
                                     _zId);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
	    CHECK_GL_ERROR("attach renderbuffers");
    }

}

////////////////////////////////////////////////////////////////////////

void 
EnvironmentMap::createTexture()
{
    // ---- Texture creation -------------------------------------------------
    // Because texture mode might have changed between FACES_MAP and 2D_ARRAY,
    //         we have to destroy and recreate a new one at each render.
    // Note: reloading shaders destroy all texture, just to make sure it will
    //       be redrawn. Hence generate it *AFTER* loading shaders !
    if( _texId ) glDeleteTextures(1, &_texId);
    glGenTextures(1, &_texId);
    CHECK_GL_ERROR("creat tex");

    // ==== FACES_MAP case ====================================================

    if( _mode == FACES_MAP )
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, _texId);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
        glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, GL_FALSE); // XXX ???
        CHECK_GL_ERROR("nearest cube map");

        // Allocate texture space for each face :
        for(int face=0; face<6; ++face)
        {
            Q_ASSERT( face <= 6 );

            glTexImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face, 
                    0,                          // mip-map level.          
                    GL_RGBA32F_ARB,             // Internal format: impose 4x32 bits floats.
                    _size, _size,               // texture size. Should be a square.
                    0,                          // no border.
                    GL_RGBA,                    // What pixel data should be loaded there ?
                    GL_FLOAT,                   // As floats.
                    0);                         // And no initialisation data (yet).
            CHECK_GL_ERROR("Define cube map face");
        }
        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    }
    // ==== PARABOLOID_MAP case ==============================================
    else if( _mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP )
    {
        // Allocate texture space for both faces :
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(TEXTURE_2D_PARABOLIC, _texId);
#if PARABOLIC_ENV_MAP_ARRAY
        glTexImage3D(TEXTURE_2D_PARABOLIC,  // target.
                0,                          // mip-map level.         
                GL_RGBA32F_ARB,             // Internal texture format.
                _size, _size,               // texture size. Should be a square.
                2,                          // 2 layers.
                0,                          // no border.
                GL_RGBA,                    // What pixel data should be loaded there ?
                GL_FLOAT,                   // Load floats.
                0);                         // And no initialisation data (yet).
        CHECK_GL_ERROR("create env tex array");
#else
        glTexImage2D(TEXTURE_2D_PARABOLIC,  // target.
                0,                          // mip-map level.         
                GL_RGBA,                    // Internal texture format.
                _size, 2*_size,             // texture size. Should be a square.
                0,                          // no border.
                GL_RGBA,                    // What pixel data should be loaded there ?
                GL_FLOAT,                   // Load floats.
                0);                         // And no initialisation data (yet).
        CHECK_GL_ERROR("create env tex 2D");
#endif

        glTexParameteri(TEXTURE_2D_PARABOLIC, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(TEXTURE_2D_PARABOLIC, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glBindTexture(TEXTURE_2D_PARABOLIC, 0);
        CHECK_GL_ERROR("mipmap env tex");
    }
    else
        qFatal("Wrong mode in EnvironmentMap texture creation");
}

////////////////////////////////////////////////////////////////////////

static Vec3f lookEnvironment[6] = {
    Vec3f(1,0,0), Vec3f(-1,0,0),
    Vec3f(0,1,0), Vec3f(0,-1,0),
    Vec3f(0,0,1), Vec3f(0,0,-1)
};
static Vec3f upEnvironment[6] = {
    Vec3f(0,-1,0), Vec3f(0,-1,0),
    Vec3f(0,0,1),  Vec3f(0,0,-1),
    Vec3f(0,-1,0), Vec3f(0,-1,0),
};

static Vec3f lookHemicube[5] = {
    Vec3f(1,0,0), Vec3f(-1,0,0),
    Vec3f(0,1,0), Vec3f(0,-1,0),
    Vec3f(0,0,1)
};
static Vec3f upHemicube[5] = {
    Vec3f(0,0,-1), Vec3f(0,0,-1),
    Vec3f(0,0,-1), Vec3f(0,0,-1),
    Vec3f(0,-1,0)
};

////////////////////////////////////////////////////////////////////////

bool 
EnvironmentMap::renderToCubeMap(const Shader* shader, const Scene* scn, int subMode)
{
    Vec3f *look = (faces()==5) ? lookHemicube : lookEnvironment;
    Vec3f *up   = (faces()==5) ?   upHemicube :   upEnvironment;

    // ---- Prepare a 45deg perspective ----------------------------------

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(90.0f,           // y-angle (degrees).
                   1.0f,            // fovy/fovx
                   _zNear,          // Near clip plane
                   _zFar);          // Far clip plane.
    CHECK_GL_ERROR("cube map perspective");

    // ---- For each face of the cube ------------------------------------
    for(int face=0; face<faces(); ++face)
    {
        // Viewing direction for that face :
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        Vec3f center = look[face] + _eye;
        gluLookAt(_eye._x,     _eye._y,     _eye._z,
                  center._x,   center._y,   center._z,
                  up[face]._x, up[face]._y, up[face]._z);
        CHECK_GL_ERROR("cube map LookAt");

        // Bind FBO as destination of rendering:
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
        CHECK_GL_ERROR("bind framebuffer");

#if USE_RENDER_TO_ENV_MAPS
        // Bind cube map texture face as destination of COLOR0 :
        glFramebufferTextureFaceEXT(GL_FRAMEBUFFER_EXT,         // target
                                    GL_COLOR_ATTACHMENT0_EXT,   // attachment
                                    _texId, 0,                  // textureId and mip-map level.
                                    GL_TEXTURE_CUBE_MAP_POSITIVE_X + face);
#endif
        // Draw to color 0 attachment (render to FBO or render to Texture):
        glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
        CHECK_GL_ERROR("draw to color 0");

        // Check everything went well :
        uint complete = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
        if( complete != GL_FRAMEBUFFER_COMPLETE_EXT )
        {
            qWarning("*** Incomplete frame buffer. Code 0x%x", complete);
            return false;
        }

        // Then Go !
        if( subMode == BasicGLViewer::SUBMODE_DEBUG )
            glClearColor( look[face]._x != 0.0f, look[face]._y != 0.0f, look[face]._z != 0.0f, 1.0);
        else
            glClearColor( 0.0f, 0.0f, 0.0f, 1.0f);    
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Make sure we start from a clear texture.

        _currentRender = true;
            scn->draw(shader, _zNear, _zFar);
        _currentRender = false;

        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
        glDrawBuffer(GL_NONE);                      // Don't draw to attachment anymore.

        // ---- And copy pixels back to the right texture face.
        // Direct render to cube-map texture do not work... So make a copy later on.
#if ! USE_RENDER_TO_ENV_MAPS
        glReadBuffer(GL_COLOR_ATTACHMENT0_EXT); // Just draw to color buffer 0, which is our cube face.
        glBindTexture(GL_TEXTURE_CUBE_MAP, _texId);
        glCopyTexSubImage2D(GL_TEXTURE_CUBE_MAP_POSITIVE_X + face,
                         0,                     // mip-map level
                         0, 0,                  // destination position
                         0, 0,                  // source position
                         _size, _size);         // width * height
        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
        glReadBuffer(GL_NONE);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        CHECK_GL_ERROR("copy cube map face");
#endif
    }
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    CHECK_GL_ERROR("unbind fbo");

    qDebug("Rendered %d faces", faces());

    return true;
}

////////////////////////////////////////////////////////////////////////

bool 
EnvironmentMap::renderToParaboloid(const Shader* shader, const Scene* scn, int subMode)
{
    static Vec3f look[2] = {
        Vec3f(0,0,1), Vec3f(0,0,-1)
    };
    static Vec3f up[2] = {
        Vec3f(0,1,0), Vec3f(0,-1,0),
    };

    // ---- For the two parabolas ----------------------------------------
    for(int face=0; face<faces(); ++face)
    {
        Q_ASSERT( face <= 2 );

        // Viewing direction for that face :
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        gluLookAt(_eye._x, _eye._y, _eye._z,
                  look[face]._x + _eye._x, look[face]._y + _eye._y, look[face]._z + _eye._z,
                  up[face]._x,   up[face]._y,   up[face]._z);
        CHECK_GL_ERROR("parabolic LookAt");

        // Make FBO to directly write results there :
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);

#if USE_RENDER_TO_ENV_MAPS
        glActiveTexture(GL_TEXTURE0);

#if PARABOLIC_ENV_MAP_ARRAY
        // Driver 100.95 --> INVALID_ENUMERANT.
        //glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,       // target
        //                          GL_COLOR_ATTACHMENT0_EXT, // attachment
        //                          GL_TEXTURE_2D_ARRAY_EXT,  // texture target
        //                          _texId, 0,                // dest texture & mipmap level
        //                          face);                    // dest layer (kka zoffset)
        //
        // Driver 100.95 --> INVALID_OPERATION.
        //glFramebufferTexture3DEXT(GL_FRAMEBUFFER_EXT,       // target
        //                          GL_COLOR_ATTACHMENT0_EXT, // attachment
        //                          GL_TEXTURE_3D,            // texture target
        //                          _texId, 0,                // dest texture & mipmap level
        //                          face);                    // dest layer (kka zoffset)
        //
        // Driver 97.94 : Ok , but *S L O W *
        // Driver >= 100.95 : Ok  and fast.
        glFramebufferTextureLayerEXT(GL_FRAMEBUFFER_EXT,    // target
                                  GL_COLOR_ATTACHMENT0_EXT, // attachment
                                  _texId, 0,                // dest texture & mipmap level
                                  face);                    // dest layer (aka zoffset)
        CHECK_GL_ERROR("attach parab envmap face");
#else
#error      Should bind a part of a 2D texture here. --> Maybe with a viewport ?
#endif
#endif

        glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);             // Do draw to texture.
        CHECK_GL_ERROR("draw to color 0");

        // Check everything went well :
        uint complete = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
        if( complete != GL_FRAMEBUFFER_COMPLETE_EXT )
        {
            qWarning("*** Incomplete frame buffer. Code 0x%x", complete);
            return false;
        }

        // Then Go !
        if( subMode == int(BasicGLViewer::SUBMODE_DEBUG) )
            glClearColor( face==0, face==1, 0.0f, 1.0f);
        else
            glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Make sure we start from a clear texture.

        _currentRender = true;
            if( _mode == SPHERE_MAP )
                scn->draw(sphereShader, _zNear, _zFar);
            else if( _mode == PARABOLA_MAP )
                scn->draw(parabolicShader, _zNear, _zFar);
            else
                scn->draw(lambertShader, _zNear, _zFar);
        _currentRender = false;

        glDrawBuffer(GL_NONE);                      // Don't draw to attachment anymore.

        // ---- And copy pixels back to the right texture face.
        // Direct render to cube-map texture do not work... So make a copy later on.
#if ! USE_RENDER_TO_ENV_MAPS
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(TEXTURE_2D_PARABOLIC, _texId);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
        glReadBuffer(GL_COLOR_ATTACHMENT0_EXT); // Just draw to color buffer 0, which is our cube face.
#if PARABOLIC_ENV_MAP_ARRAY
        glCopyTexSubImage3D(TEXTURE_2D_PARABOLIC,
                         0,                     // level
                         0, 0, face,            // destination position
                         0, 0,                  // source position
                         _size, _size);         // width * height
#else
        glCopyTexSubImage2D(TEXTURE_2D_PARABOLIC,
                         0,                     // level
                         0, face*_size,         // destination position
                         0, 0,                  // source position
                         _size, _size);         // width * height
#endif
        glReadBuffer(GL_NONE);
        CHECK_GL_ERROR("copy parab envmap face");
#endif
    }
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);    // Back to screen.

    qDebug("Rendered %d faces", faces());

    return true;
}

////////////////////////////////////////////////////////////////////////

void 
EnvironmentMap::renderLater(const Vec3f& newEye, Mode newMode)
{
    if( & newEye )
        _eye = newEye;
    if( newMode != NO_CHANGES )
        _mode = newMode;

    if( _texId )
        glDeleteTextures(1, &_texId);
    _texId = 0;

    // Reset statistics...
    smoothedRenderTime = 0.0;
    smoothedCount      = 0;
}

void
EnvironmentMap::renderAllMaps(const Shader* shader, const Scene* scene, int subMode)
{
    int n=0;                        // env map number (for PNG dumps).
    bool firstRender = false;       // Don't include first rnder in statistics.

    glFinish();                     // Make sure OpenGL pipeline is clean.
    double time = TBender::time();  // Start date.

    for(Maps::const_iterator it(allMaps.begin()); it != allMaps.end(); ++it, ++n)
    {
        EnvironmentMap *e = *it;
        if( ! e->texId() ) firstRender = true;

        if( e->render(shader, scene, subMode) )
        {
            if( scene->env.dumpImages )
            {
                glFinish();                 // Finish render before stoping timer.
                time += TBender::time();    // STOP timer.

                QImage img;
                e->getImage(img);
                img.save( QString("env%1Map.png").arg(n));

                time -= TBender::time();    // RESTART timer.
            }
        }
    }
    glFinish();                     // Make sure rendering is done.
    time = TBender::time() - time;

    if( n > 0 )
    {
        smoothedRenderTime += time;
        smoothedCount      += n;

        qWarning("Env map rendering time = %s , mean = %s (%d maps)",
            TBender::timeCString(time / n).toLatin1().constData(),
            TBender::timeCString(smoothedRenderTime / smoothedCount).toLatin1().constData(),
            smoothedCount);
    }
}

////////////////////////////////////////////////////////////////////////

bool
EnvironmentMap::enable(int texUnit, const Shader* shader) const
{
    // Not rendered yet ?
    if( ! _texId )
        return false;

    glActiveTexture(GL_TEXTURE0 + texUnit);
    if( _mode == FACES_MAP )
    {
        glBindTexture(GL_TEXTURE_CUBE_MAP, _texId);
	    CHECK_GL_ERROR("bind cube map");
    }
    else if( _mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP )
    {
        glBindTexture(TEXTURE_2D_PARABOLIC, _texId);
        CHECK_GL_ERROR("bind parab envmap");
    }

    if( shader )
    {
        if( _mode == FACES_MAP )
        {
            shader->setUniform("uHasEnvMap", true, GL_BOOL);
            shader->setUniform("uEnvMapCube", texUnit, GL_SAMPLER_CUBE);
        }
        else if( _mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP )
        {
            shader->setUniform("uHasParabolicEnvMap", true, GL_BOOL);
            shader->setUniform("uEnvMapParabolic", texUnit, SAMPLER_2D_PARABOLIC);
        }
    }
    else
    {
        if( _mode == FACES_MAP )
        {
            glEnable(GL_TEXTURE_CUBE_MAP);
            // XXX : Perfect miror: env color fully replace object's color:
            // Would be better to make some mix, but then we should change 
            // coloring in ScnObject::draw()...
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
            CHECK_GL_ERROR("enable GL cube map");

            glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
            glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
            glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GL_REFLECTION_MAP);
            glEnable(GL_TEXTURE_GEN_S);
            glEnable(GL_TEXTURE_GEN_T);
            glEnable(GL_TEXTURE_GEN_R);
            CHECK_GL_ERROR("cube map coord gen");

            // Do the same re-orientation of reflected ray as in
            // phong_fragment shader...
            // BUG : texture matrix trick don't do the right thing. Why ?
            Matrix4 modelView;
            glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat*)modelView._m);
            modelView = modelView.transpose();
            glMatrixMode(GL_TEXTURE);
            glLoadMatrixf( (GLfloat*)modelView._m);
            CHECK_GL_ERROR("cube map coord rotate");
        }
        else if( _mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP )
        {
            // Not in shader mode, but parabolic map wanted : add our own shader
            // there...
            Q_ASSERT( phongShader );
            phongShader->enable();
            phongShader->setUniform("uHasParabolicEnvMap", true, GL_BOOL);
            phongShader->setUniform("uEnvMapParabolic", texUnit, SAMPLER_2D_PARABOLIC);
        }
    }

    return true;
}

////////////////////////////////////////////////////////////////////////

void
EnvironmentMap::disable(int texUnit, const Shader* shader) const
{
    // Not rendered yet ?
    if( ! _texId )
        return;

    glActiveTexture(GL_TEXTURE0 + texUnit);
    if( _mode == FACES_MAP )
    {
        glDisable(GL_TEXTURE_CUBE_MAP);
        glBindTexture(GL_TEXTURE_CUBE_MAP, _texId);
        CHECK_GL_ERROR("disable cube map");
    }
    else if( _mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP )
    {
        glBindTexture(TEXTURE_2D_PARABOLIC, _texId);
        CHECK_GL_ERROR("disable parab map");
    }

    if( shader )
    {
        shader->setUniform("uHasEnvMap", false, GL_BOOL);
        shader->setUniform("uHasParabolicEnvMap", false, GL_BOOL);
    }
    else
    {
        if( _mode == FACES_MAP )
        {
            glDisable(GL_TEXTURE_GEN_S);
            glDisable(GL_TEXTURE_GEN_T);
            glDisable(GL_TEXTURE_GEN_R);
            glMatrixMode(GL_TEXTURE);
            glLoadIdentity();
            CHECK_GL_ERROR("disable GL env map");
        }
        else if( _mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP )
        {
            Q_ASSERT( phongShader );
            phongShader->setUniform("uHasParabolicEnvMap", false, GL_BOOL);
            phongShader->disable();
        }
    }
}

//////////////////////////////////////////////////////////////////////////////

bool
EnvironmentMap::getImage(QImage& image) const
{
    if( _mode == FACES_MAP )
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, _texId);
        CHECK_GL_ERROR("bind cube map");

        image = QImage(3*_size, 2*_size, QImage::Format_RGB32);
        image.fill( qRgba(128,0,128,255) );     // Violet where no write occurs.

        glPixelStorei(GL_PACK_ROW_LENGTH, image.width());   // in pixels, not bytes.
        for(int face=0; face<6; ++face)
        {
            int x = _size*(face % 3);
            int y = _size*(face / 3);
            GLuint order[6] = {
                GL_TEXTURE_CUBE_MAP_POSITIVE_X,
                GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,
                GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
                GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
                GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
                GL_TEXTURE_CUBE_MAP_POSITIVE_Y
            };

            uchar *ptr = image.scanLine(y) + 4 * x;         // in bytes, not pixels !
            glGetTexImage(order[face],
                          0,                    // level
                          GL_BGRA,              // Qt's byte order (on win32 at least)
                          GL_UNSIGNED_BYTE,
                          ptr);
            CHECK_GL_ERROR("get cube image");
        }

        glPixelStorei(GL_PACK_ROW_LENGTH, 0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
        CHECK_GL_ERROR("unbind cube map");
    }
    else // _mode==PARABOLOID_MAP
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(TEXTURE_2D_PARABOLIC, _texId);
        CHECK_GL_ERROR("bind parab map");

        image = QImage(_size, 2*_size, QImage::Format_RGB32);
        image.fill( qRgba(128,0,128,255) );     // Violet where no write occurs.

        // A uniq GetTexImage do returns all layers at once :
        glGetTexImage(TEXTURE_2D_PARABOLIC,
                      0,                 	    // level
                      GL_BGRA,                  // Qt's byte order (on win32 at least)
                      GL_UNSIGNED_BYTE,
                      image.scanLine(0));
        CHECK_GL_ERROR("get parab image");

        glBindTexture(TEXTURE_2D_PARABOLIC, 0);
        CHECK_GL_ERROR("unbind parab map");
    }

    return true;
}
