////////////////////////////////////////////////////////////////////////
/*! \file   EnvironmentMap.h
 *  \brief  EnvironmentMap for an OpenGL rendering.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Scene, Shader, FrameBufferObject.
 */
////////////////////////////////////////////////////////////////////////
// $Id: EnvironmentMap.h,v 1.13 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2007/04/07] : Creation.

#ifndef ENVIRONMENT_MAP_H
#define ENVIRONMENT_MAP_H

#include "FrameBufferObject.h"
#include "Math/Vec3f.h"

class Shader;
class Scene;
class Environment;

// Render to texture, when in CUBE_MAP or 2D_ARRAY mode, does not
// work... Even with 158.22 driver, but it does with other implementations !
#define USE_RENDER_TO_ENV_MAPS  0
// For parabolic envmaps, either use a texture2DArray, 
// or a single texture, twice as height. Wich don't work with render to texture
// (no sub-texture rendering, or viewports, yet).
#define PARABOLIC_ENV_MAP_ARRAY 0

//
// Note: timmings on the vostro laptop :
//                                cube map  /  parabolic env map
//   Render to FBO + viewport :    1.40 s   /  0.50 s
//   Render to FBO + ARRAY    :    1.37 s   /  0.57 s
//   Render to Tex + viewport : (cannot compile yet)
//   Render to Tex + ARRAY    :    1,80 s   /  0,80 s [Lemmings+Plan+3 spheres]


/// \brief Fake relection of the environment into an OpenGL object.
class EnvironmentMap : private FrameBufferObject
{
public:
    enum Mode {
        NO_CHANGES = 0,             //!< Keep last mode...
        FACES_MAP,                  //!< Use 5/6 plane textures.
        SPHERE_MAP,                 //!< Use 1/2 cosinus-sphere textures.
        PARABOLA_MAP,               //!< Use 1/2 paraboloic textures.
        LAMBERT_MAP                 //!< Only 1/2 HemiCube derived class.
    };

    EnvironmentMap(Mode, const Vec3f& eye, int size, float radius);

    //! Cleanup any remainign shaders.
    virtual ~EnvironmentMap();

    //! Reload and relink any used shaders.
    /// And mark all env maps dirty.
    static bool reloadShaders(const Environment& env, int subMode);

    //! Recompute this envmap from \p position, for a scene.
    virtual bool render(const Shader* shader, const Scene*, int subMode);

    //! Mark the env map to be computed later on. Possibly from a new position.
    virtual void renderLater(const Vec3f& newEye = *(Vec3f*)0, Mode newMode = NO_CHANGES);

    //! Recompute all dirty envmaps.
    static void renderAllMaps(const Shader* shader, const Scene*, int subMode);

    virtual bool enable(int texUnit, const Shader* shader) const;     //!< Start using the envmap texture.
    virtual void disable(int texUnit, const Shader* shader) const;    //!< Stop using the envmap texture.

    //! Recover the environment map, and copy it to the given image.
    virtual bool getImage(QImage& image) const;

    //!< True during rendering of this envmap. Use to cull object rendering.
    inline bool currentRender() const { return _currentRender; }

    inline unsigned int texId() const { return _texId; }

protected:
    Mode        _mode;              //!< cube map or paraboloid ?
    Vec3f       _eye;               //!< Center to compute the envmap from.
    int         _size;              //!< size of each square element of the texture.
    float       _zNear;             //!< Scene min radius from the env map.
    float       _zFar;              //!< Scene max radius from the env map.
    unsigned int _texId;            //!< CUBE_MAP texture id.

    bool        _currentRender;     //!< Self object's culling.

    static Shader* phongShader;     //!< Used to compute cube maps, and apply parabolic maps.
    static Shader* sphereShader;    //!< Used to compute hemicube/sphere env maps.
    static Shader* parabolicShader; //!< Used to compute environment/parabola env maps.
    static Shader* lambertShader;   //!< Used to compute hemicube/sphere env maps.

    virtual int faces();            //!< Number of faces to render (cube, hemicube, parabola, etc.)
    virtual void createFBO();       //!< Allocate and define FBO (z only, or z+colors)
    virtual void createTexture();   //!< Allocate texture (cube map, parabola map, layers)
    virtual bool renderToCubeMap(const Shader* shader, const Scene*, int subMode); //!< Direct or using to FBO to texture copy.
    virtual bool renderToParaboloid(const Shader* shader, const Scene*, int subMode); //!< Direct or using to FBO to texture copy.
};

#endif // ENVIRONMENT_MAP_H

