////////////////////////////////////////////////////////////////////////
/*! \file   Feedback.h
 *  \brief  Using OpenGL FeedBack buffers to debug geometry shader...
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Vec3f.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Feedback.cpp,v 1.15 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2006/12/13] : Creation.

#include "GL/glew.h"

#include "Utilities.h"
#include "Feedback.h"

#include <QtOpenGL>

// Extension GLEW_NV_transform_feedback
//
// glBeginTransformFeedbackNV(GLenum primitiveMode)
// glEndTransformFeedbackNV(void)
// glGetTransformFeedbackVaryingNV(GLuint program, GLuint index, GLint *location)
// glTransformFeedbackAttribsNV(GLuint program, GLsizei count, const GLint *locations, GLenum bufferMode)
// glTransformFeedbackVaryingsNV(GLuint count, const GLint *attribs, GLenum bufferMode)
//
// glGetActiveVaryingNV(GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLsizei *size, GLenum *type, GLchar *name)
// glGetVaryingLocationNV(GLuint program, const GLchar *name)

#define MAX_VARYING 1024     // glGetIntegerv(GL_MAX_VARYING_FLOATS, &max);

//////////////////////////////////////////////////////////////////////////////

struct Variable
{
    QString name;
    int     type;       //!< For varying: GL_FLOAT, GL_FLOAT_VEC2, etc. For attribs: num of components.
    int     size;       //!< Array size.
    int     location;   //!< Location of varying, or GL_POSITION, etc. for attribs.

    inline Variable(QString name = QString::null, int type=0, int size=0, int location=0)
    : name(name), type(type), size(size), location(location)
    {}
};
static QVector<Variable> varyings;
static QVector<Variable> uniforms;

//////////////////////////////////////////////////////////////////////////////

Feedback::Feedback(uint prog, uint mode)
: _query(0), _vbo(0),
  _prog(prog),
  _mode(mode)
{
    // ---- OpenGL setup -----------------------------------------------------
    if( ! GLEW_NV_transform_feedback )
    {
        qWarning("*** Missing needed extension GL_NV_transform_feedback");
    }
    else
    {
        qDebug("OpenGL extension GL_NV_transform_feedback ok.");

        glGenQueries(1, &_query);
        if( _query == 0 )
            qWarning("*** Unable to generate query");

        glGenBuffers(1, &_vbo);
        glBindBuffer(GL_ARRAY_BUFFER, _vbo);
        glBufferData(GL_ARRAY_BUFFER, MAX_VARYING * sizeof(float[4]), 0, GL_DYNAMIC_READ);
        glBindBuffer(GL_ARRAY_BUFFER, 0);
    }
    CHECK_GL_ERROR("alloc feedback");

    // ---- Get uniforms variables used --------------------------------------
    uniforms.clear();
    if( _prog )
    {
        int nbUniform = 0;
        glGetProgramiv(_prog, GL_ACTIVE_UNIFORMS, &nbUniform);
        for(int u=0; u<nbUniform; ++u)
        {
            char buffer[120]; int nameLen; int uniformSize; GLenum uniformType;
            glGetActiveUniform(_prog, u, sizeof buffer, &nameLen, &uniformSize, &uniformType, buffer);
            if( nameLen )
            {
                int uniformLoc = (uniformSize==1) ? glGetUniformLocation(_prog, buffer)
                                                  : glGetUniformLocation(_prog, QString("%1[0]").arg(buffer).toAscii().constData() );
                // Beware all gl_XXX uniforms do return a -1 location, because they are not uniforms...
                if( uniformLoc >= 0 )
                    uniforms.push_back( Variable(QString(buffer),
                                                uniformType,
                                                uniformSize,
                                                uniformLoc) );
            }
        }

    }
    // ---- Get shader varying variables -------------------------------------
    varyings.clear();   // Oups. We should not share that !
    
    if( _prog )
    {
        int nbVarying = 0;
        glGetProgramiv(_prog, GL_ACTIVE_VARYINGS_NV, &nbVarying);
        for(int v=0; v<nbVarying; ++v)
        {
            char buffer[120]; int nameLen; int varSize; GLenum varType;
            glGetActiveVaryingNV(_prog, v, sizeof buffer, &nameLen, &varSize, &varType, buffer);

            if( nameLen )
            {
                int varLoc = ( varSize==1 ) ? glGetVaryingLocationNV(_prog, buffer)
                                            : glGetVaryingLocationNV(_prog, (QString(buffer)+"[0]").toAscii().constData() );

                varyings.push_back( Variable(QString(buffer),
                                             varType, 
                                             varSize,
                                             varLoc) );
            }
        }
    }
    else
    {
        //int nbAttribs = 0;
        //glGetIntegerv(GL_TRANSFORM_FEEDBACK_ATTRIBS_NV, &nbAttribs);
        //for(int a=0; a<nbAttribs; ++a)
        //{
        //    static struct Names {GLenum type; const char* name; } names [12] = {
        //        GL_POSITION,                "gl_Position",
        //        GL_PRIMARY_COLOR,           "gl_FrontColor",
        //        GL_SECONDARY_COLOR_NV,      "gl_SecondaryFrontColor",
        //        GL_BACK_PRIMARY_COLOR_NV,   "gl_BackColor",
        //        GL_BACK_SECONDARY_COLOR_NV, "gl_SecondaryBackColor",
        //        GL_FOG_COORDINATE,          "gl_FogCoord",
        //        GL_POINT_SIZE,              "gl_PointSize",
        //        GL_TEXTURE_COORD_NV,        "gl_TexCoord[%1]",
        //        GL_CLIP_DISTANCE_NV,        "gl_ClipDistance[%1]",
        //        GL_VERTEX_ID_NV,            "gl_VertexID",
        //        GL_PRIMITIVE_ID_NV,         "gl_PrimitiveID",
        //        GL_GENERIC_ATTRIB_NV,       "gl_Attrib[%1]"
        //    };
        //    GLint args[3]; int loc;
        //    glGetIntegerIndexedvEXT(GL_TRANSFORM_FEEDBACK_RECORD_NV, a, args);
        //    glGetTransformFeedbackVaryingNV(_prog, a, &loc);

        //    for(int n=0; n<12; ++n)
        //    {

        //        if( names[n].type == args[0] )
        //        {
        //            GLenum type[4] = {GL_FLOAT, GL_FLOAT_VEC2, GL_FLOAT_VEC3, GL_FLOAT_VEC4};
        //            if( strchr(names[n].name, '%') )
        //                varyings.push_back( Variable( QString(names[n].name).arg(args[2]),
        //                                          type[args[1]],
        //                                          1,
        //                                          loc) );
        //            else
        //                varyings.push_back( Variable(QString(names[n].name), 
        //                                          type[args[1]],
        //                                          1,
        //                                          loc) );
        //            break;
        //        }
        //    }
        //}

        varyings.push_back( Variable("gl_Position",   4, 1, GL_POSITION) );
        varyings.push_back( Variable("gl_FrontColor", 4, 1, GL_PRIMARY_COLOR) );
        //varyings.push_back( Variable("gl_PrimitiveID",   1, 1, GL_PRIMITIVE_ID_NV) );
    }
}

//////////////////////////////////////////////////////////////////////////////

Feedback::~Feedback()
{
    if( _query )
        glDeleteQueries(1, &_query);
}

//////////////////////////////////////////////////////////////////////////////

void Feedback::start(bool disableDraw)
{
    if( _query )        // Extension presetm, and init succeded.
    {
        _rasterizerDiscard = disableDraw;
        if(_rasterizerDiscard)
            glEnable(GL_RASTERIZER_DISCARD_NV);


        if( _prog )      // A shader is active. Let's suppose there is a vertex/ geometry shader.
        {
            int locations[MAX_VARYING], count=0;
            for(int i=0; i<varyings.count(); ++i)
            {
                for(int a=0; a<varyings[i].size; ++a)
                    locations[count++] = varyings[i].location + a;
            }
            glTransformFeedbackVaryingsNV(_prog, count, locations, GL_INTERLEAVED_ATTRIBS_NV);
        }
        else
        {
            int attribs[3*MAX_VARYING], count=0;
            for(int i=0; i<varyings.count(); ++i)
            {
                for(int a=0; a<varyings[i].size; ++a)
                {
                    attribs[count++] = varyings[i].location;
                    attribs[count++] = varyings[i].type;
                    attribs[count++] = a;
                }
            }
            glTransformFeedbackAttribsNV(count/3, attribs, GL_INTERLEAVED_ATTRIBS_NV);
        }

        
        // Bind the full vbo, for interleaved attributes (hence index = 0).
        glBindBufferBaseNV(GL_TRANSFORM_FEEDBACK_BUFFER_NV, 0, _vbo);

        glBeginQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_NV, _query);
        glBeginTransformFeedbackNV(_mode);
        CHECK_GL_ERROR("begin query");
    }
}

//////////////////////////////////////////////////////////////////////////////

void Feedback::stop()
{
    if( _query )        // Extension presetm, and init succeded.
    {
        if(_rasterizerDiscard)
            glDisable(GL_RASTERIZER_DISCARD_NV);
        glEndTransformFeedbackNV();
        glEndQuery(GL_TRANSFORM_FEEDBACK_PRIMITIVES_WRITTEN_NV);
        CHECK_GL_ERROR("end query");
    }
}

//////////////////////////////////////////////////////////////////////////////

void Feedback::printAll()
{
    if( _query )        // Extension presetm, and init succeded.
    {
        // ---- Print all uniform variables ----------------------------------
        qWarning("---- Uniforms");

        for(int u=0; u<uniforms.count(); ++u)
        {
            union {
                GLfloat floats[4];
                GLint   ints[4];
                GLuint  uints[4];
            } params;

            QString value;
            for(int i=0; i<uniforms[u].size; ++i)
            {
                switch( uniforms[u].type )
                {
                case GL_FLOAT:
                    glGetUniformfv(_prog, uniforms[u].location, params.floats);
                    value = QString("%1").arg(params.floats[0]);
                    break;
                case GL_FLOAT_VEC2:
                    glGetUniformfv(_prog, uniforms[u].location, params.floats);
                    value = QString("%1 %2").arg(params.floats[0]).arg(params.floats[1]);
                    break;
                case GL_FLOAT_VEC3:
                    glGetUniformfv(_prog, uniforms[u].location, params.floats);
                    value = QString("%1 %2 %3").arg(params.floats[0]).arg(params.floats[1]).arg(params.floats[2]);
                    break;
                case GL_FLOAT_VEC4:
                    glGetUniformfv(_prog, uniforms[u].location, params.floats);
                    value = QString("%1 %2 %3 %4").arg(params.floats[0]).arg(params.floats[1]).arg(params.floats[2]).arg(params.floats[3]);
                    break;

                case GL_INT:
                    glGetUniformiv(_prog, uniforms[u].location, params.ints);
                    value = QString("%1").arg(params.ints[0]);
                    break;
                case GL_INT_VEC2:
                    glGetUniformiv(_prog, uniforms[u].location, params.ints);
                    value = QString("%1 %2").arg(params.ints[0]).arg(params.ints[1]);
                    break;
                case GL_INT_VEC3:
                    glGetUniformiv(_prog, uniforms[u].location, params.ints);
                    value = QString("%1 %2 %3").arg(params.ints[0]).arg(params.ints[1]).arg(params.ints[2]);
                    break;
                case GL_INT_VEC4:
                    glGetUniformiv(_prog, uniforms[u].location, params.ints);
                    value = QString("%1 %2 %3 %4").arg(params.ints[0]).arg(params.ints[1]).arg(params.ints[2]).arg(params.ints[3]);
                    break;

                case GL_BOOL:
                    glGetUniformiv(_prog, uniforms[u].location, params.ints);
                    value = QString("%1").arg( params.ints[0] ? "true " : "false" );
                    break;
                case GL_BOOL_VEC2:
                    glGetUniformiv(_prog, uniforms[u].location, params.ints);
                    value = QString("%1 %2").arg( params.ints[0] ? "true " : "false" )
                                            .arg( params.ints[1] ? "true " : "false" );
                    break;
                case GL_BOOL_VEC3:
                    glGetUniformiv(_prog, uniforms[u].location, params.ints);
                    value = QString("%1 %2 %3").arg( params.ints[0] ? "true " : "false" )
                                               .arg( params.ints[1] ? "true " : "false" )
                                               .arg( params.ints[2] ? "true " : "false" );
                    break;
                case GL_BOOL_VEC4:
                    glGetUniformiv(_prog, uniforms[u].location, params.ints);
                    value = QString("%1 %2 %3 %4").arg( params.ints[0] ? "true " : "false" )
                                                  .arg( params.ints[1] ? "true " : "false" )
                                                  .arg( params.ints[2] ? "true " : "false" )
                                                  .arg( params.ints[3] ? "true " : "false" );

#ifdef GL_EXT_gpu_shader4
                case GL_UNSIGNED_INT:
                    glGetUniformuivEXT(_prog, uniforms[u].location, params.uints);
                    value = QString("%1").arg(params.uints[0]);
                    break;
                case GL_UNSIGNED_INT_VEC2_EXT:
                    glGetUniformuivEXT(_prog, uniforms[u].location, params.uints);
                    value = QString("%1 %2").arg(params.uints[0]).arg(params.uints[1]);
                    break;
                case GL_UNSIGNED_INT_VEC3_EXT:
                    glGetUniformuivEXT(_prog, uniforms[u].location, params.uints);
                    value = QString("%1 %2 %3").arg(params.uints[0]).arg(params.uints[1]).arg(params.uints[2]);
                    break;
                case GL_UNSIGNED_INT_VEC4_EXT:
                    glGetUniformuivEXT(_prog, uniforms[u].location, params.uints);
                    value = QString("%1 %2 %3 %4").arg(params.uints[0]).arg(params.uints[1]).arg(params.uints[2]).arg(params.uints[3]);
                    break;
#endif

                default:
                    value = "<unknown type>";
                    break;
                }
                if( uniforms[u].size == 1 )
                    qWarning( QString("\t%1\t= %2").arg(uniforms[u].name).arg(value).toAscii().constData() );
                else
                    qWarning( QString("\t%1[%2]\t= %3").arg(uniforms[u].name).arg(i).arg(value).toAscii().constData() );
            }
        }

        // ---- Print varying, using the VBO filled durring draw -------------

        int nbPrims = 0;
        glGetQueryObjectiv(_query, GL_QUERY_RESULT, &nbPrims);
        int nbVertices = _mode == GL_POINTS ? 1
                       : _mode == GL_LINES  ? 2
                       : 3
                       ;

        glBindBuffer(GL_ARRAY_BUFFER, _vbo);
        if( float *floats = (float *)glMapBuffer(GL_ARRAY_BUFFER, GL_READ_ONLY) )
        {
            int  * ints = ( int*)floats;
            uint *uints = (uint*)floats;

            // ---- For each primitive ---------------------------------------
            int location = 0;
            for(int i=0; i<nbPrims; ++i)
            {
                for(int j=0; j<nbVertices; ++j)
                {
                    qWarning( "Primitive %d Vertex %d:", i, j );

                    // ---- For each vertex --------------------------------------
                    // int location = 4 * varyings[v].location;
                    for(int v=0; v<varyings.count(); ++v)
                    {
                        for(int a=0; a<varyings[v].size; ++a)
                        {
                            QString value;
                            int nextLocation = location;
                            switch( varyings[v].type )
                            {
                            case 4:
                            case GL_FLOAT_VEC4: value = QString::number(floats[location + 3], 'f', 6) + " " + value; ++nextLocation;
                            case 3:
                            case GL_FLOAT_VEC3: value = QString::number(floats[location + 2], 'f', 6) + " " + value; ++nextLocation;
                            case 2:
                            case GL_FLOAT_VEC2: value = QString::number(floats[location + 1], 'f', 6) + " " + value; ++nextLocation;
                            case 1:
                            case GL_FLOAT:      value = QString::number(floats[location + 0], 'f', 6) + " " + value; ++nextLocation;
                                break;
                            case GL_INT_VEC4: value = QString::number(ints[location + 3]) + " " + value; ++nextLocation;
                            case GL_INT_VEC3: value = QString::number(ints[location + 2]) + " " + value; ++nextLocation;
                            case GL_INT_VEC2: value = QString::number(ints[location + 1]) + " " + value; ++nextLocation;
                            case GL_INT:      value = QString::number(ints[location + 0]) + " " + value; ++nextLocation;
                                break;
                            case GL_UNSIGNED_INT_VEC4_EXT: value = QString::number(uints[location + 3]) + " " + value; ++nextLocation;
                            case GL_UNSIGNED_INT_VEC3_EXT: value = QString::number(uints[location + 2]) + " " + value; ++nextLocation;
                            case GL_UNSIGNED_INT_VEC2_EXT: value = QString::number(uints[location + 1]) + " " + value; ++nextLocation;
                            case GL_UNSIGNED_INT:          value = QString::number(uints[location + 0]) + " " + value; ++nextLocation;
                                break;
                            default:
                                value = "<unknown type>";
                                break;
                            }
                            location = nextLocation;

                            if( varyings[v].size == 1 )
                                qWarning( ("\t" + varyings[v].name + "\t= " + value).toAscii().constData() );
                            else
                                qWarning( ("\t" + varyings[v].name + QString("[%1]\t= ").arg(a) + value).toAscii().constData() );
                        }
                    }
                }
                qWarning(" "); // Skip empty line.
            }
            glUnmapBuffer(GL_ARRAY_BUFFER);
        }
        glBindBuffer(GL_ARRAY_BUFFER, 0);
        CHECK_GL_ERROR("dump feedback");
    }
}
