////////////////////////////////////////////////////////////////////////
/*! \file   Feedback.h
 *  \brief  Using OpenGL FeedBack buffers to debug geometry shader...
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Shader.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Feedback.h,v 1.3 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2006/12/13] : Creation.

#ifndef FEEDBACK_H
#define FEEDBACK_H

/// \brief A class to debug vertex/geometry shaders using OpenGL's Feedback buffers.
///
/// This class enable to trap data flow before the rasterisation and the fragment
/// shader, hence after a possible vertex and geometry shader.
///
/// Typical usage :
/// \code
///    shader->enable();            // Should be called *BEFORE* starting to record.
///    if( debug ) feedback->start();
///      ... standard scene redraw ...
///    if( debug ) {
///      feedback->stop();          // Back to normal mode.
///      feedback->printAll();      // Print all varying variables entering fragment shader.
///    }
///    shader->disable();
/// \endcode
///
class Feedback
{
    bool _rasterizerDiscard;        //!< draw was disabled by last enable().
    unsigned int _query;            //!< Asynchronous query to get values back.
    unsigned int _vbo;              //!< Buffer to dump variables...
    unsigned int _prog;             //!< Program to trace.
    unsigned int _mode;             //!< Feedback mode (either GL_POINTS, GL_LINES, GL_TRIANGLES)

public:
    /// \brief Setup a new feedback tracer object.
    ///
    /// \p prog is the Id of the shader program to trace.
    ///
    /// \p mode is either GL_POINTS, GL_LINES, GL_TRIANGLES, the macro type of
    ///         primitive allowed to be traced.
    Feedback(unsigned int prog, unsigned int mode);
    virtual ~Feedback();

    /// \brief Start recording.
    /// If \p disabledraw is true, the rasterization stage is disabled completly
    /// until the feedback is stop()'ed.
    void start(bool disableDraw = false);

    /// \brief Stop recording.
    /// And put back rendering pipeline in a standard state.
    void stop();

    /// \brief Dump all data recorded.
    /// Print all varying data recorded in the last start()/stop().
    void printAll();
};

#endif // FEEDBACK_H
