////////////////////////////////////////////////////////////////////////
/*! \file   FrameBufferObject.cpp
 *  \brief  OpenGL's Frame Buffer Object.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Scene, Shader.
 */
////////////////////////////////////////////////////////////////////////
// $Id: FrameBufferObject.cpp,v 1.12 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2007/03/15] : Creation.

#include "FrameBufferObject.h"

#include <GL/glew.h>

#include <qglobal.h>
#include <qimage.h>

#include "Utilities.h"

////////////////////////////////////////////////////////////////////////
FrameBufferObject::FrameBufferObject()
: _id(0),
  _rbId(0),
  _zId(0)
{}

FrameBufferObject::~FrameBufferObject()
{
    if( _rbId )
        glDeleteRenderbuffersEXT(1, &_rbId);
    if( _zId )
        glDeleteRenderbuffersEXT(1, &_zId);
    if( _id )
        glDeleteFramebuffersEXT(1, &_id);

    _id = _zId = _rbId = 0;
}

//////////////////////////////////////////////////////////////////////////////

bool
FrameBufferObject::bindColorImage(int width, int height, bool depth)
{
    Q_ASSERT( ! _id );                  // Should be bound just once.

    // Make sure the image is correctly initialized to hold our depth
    // data.
    static int maxRBSize = 0;
    if( ! maxRBSize )
    {
        glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &maxRBSize);
        qDebug("Max renderbuffer size is %d", maxRBSize);
    }
    Q_ASSERT( 0 < width  && width  <= maxRBSize );
    Q_ASSERT( 0 < height && height <= maxRBSize );

    // Sanity check, before crashing on null function pointers...
    if( ! GLEW_EXT_framebuffer_object )
    {
        qWarning("*** Missing GL_EXT_framebuffer_object");
        return false;
    }

    //------------------------------------------------------------------------

    // Creates the FBO and render buffer ids, if not already done.
    if( ! _id )
    {
    	glGenFramebuffersEXT(1, &_id);
        CHECK_GL_ERROR("alloc FBO");
    }

    if( ! _rbId )
    {
        glGenRenderbuffersEXT(1, &_rbId);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, _rbId);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                                 GL_RGBA8,
                                 width, height);
        CHECK_GL_ERROR("alloc color renderbuffer");
    }

    if( depth && ! _zId )
    {
        glGenRenderbuffersEXT(1, &_zId);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, _zId);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                                 GL_DEPTH_COMPONENT24,
                                 width, height);
        CHECK_GL_ERROR("alloc depth renderbuffer");
    }

    // Tells OpenGL the FBO is to be mapped to that texture
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
                                 GL_COLOR_ATTACHMENT0_EXT,
                                 GL_RENDERBUFFER_EXT,
                                 _rbId);
    if( depth )
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
                                     GL_DEPTH_ATTACHMENT_EXT,
                                     GL_RENDERBUFFER_EXT,
                                     _zId);
    CHECK_GL_ERROR("attach renderbuffers");

    // Use a single rendering target for that RGBA image :
	glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
    glReadBuffer(GL_NONE);          // Reccursivity is a bad idea, anyway.
    CHECK_GL_ERROR("draw to FBO");

    // Make sure everything went well :
    uint complete = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    if( complete != GL_FRAMEBUFFER_COMPLETE_EXT)
    {
        qWarning("*** Incomplete frame buffer. Code %d", complete);
        return false;
    }

	return true;
}

//////////////////////////////////////////////////////////////////////////////

bool
FrameBufferObject::finish(QImage& img) const
{
	// Make sure we use our own framebuffer :
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);

	// ---- Some sanity checking ---------------------------------------------
    int w=0, h=0;
    glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_WIDTH_EXT,  &w);
    glGetRenderbufferParameterivEXT(GL_RENDERBUFFER_EXT, GL_RENDERBUFFER_HEIGHT_EXT, &h);
    if( img.width() != w || img.height() != h )
    {
        qWarning("Wrong image size.");
        return false;
    }
    CHECK_GL_ERROR("check FBO size");

	// ---- Reading data -----------------------------------------------------
	glDrawBuffer(GL_NONE);                      // Stop render redirection.
    glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);     // Enable read-pixels source.
    CHECK_GL_ERROR("read from FBO");

    glReadPixels(0, 0,                          // From top-left corner
                 img.width(), img.height(),     // Up to bottom-right.
                 GL_BGRA,                       // Qt's byte order (on win32 at least)
                 GL_UNSIGNED_BYTE,
                 img.bits());					// Directly into Qt's image buffer.
    CHECK_GL_ERROR("get fbo image");

    // Qt's is Y up, OpenGL is Y down, so flip image...
    img = img.mirrored(false, true);

    // Back to normal rendering state.
	glReadBuffer(GL_NONE);
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    CHECK_GL_ERROR("unbind FBO");

    return true;
}
