//////////////////////////////////////////////////////////////////////////////
/*! \file   GlossyMap.cpp
 *  \brief  Glossy plan, has a 2D array of spheremaps.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa FrameBufferObject, Scene, Shader.
 */
//////////////////////////////////////////////////////////////////////////////
// $Id: GlossyMap.cpp,v 1.16 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2007/04/19] : Creation.

#include "GlossyMap.h"

#include <GL/glew.h>

#include <QVector>
#include <QApplication>

#include "Math/Matrix4.h"
#include "Utilities.h"
#include "Shader.h"
#include "Scene.h"

static bool unsupported = false;

//////////////////////////////////////////////////////////////////////////////

typedef QVector<GlossyMap*> GlossyMaps;
static GlossyMaps allGlossyMaps;

Shader* GlossyMap:: phongShader = 0;
Shader* GlossyMap::sphereShader = 0;

//////////////////////////////////////////////////////////////////////////////

GlossyMap::GlossyMap()
: _nu(6),
  _nv(6),
  _center(0,0,0),
  _zFar(-1.0f),
  _u(0.5,0,0),
  _v(0.0, 0.0, 0.5),
  _size(100),
  _texId(0)
{
    allGlossyMaps.push_back(this);
}

GlossyMap::~GlossyMap()
{
    int idx = allGlossyMaps.indexOf(this);
    if( idx >= 0 )
        allGlossyMaps.remove(idx);

    clean();
}

//////////////////////////////////////////////////////////////////////////////

void
GlossyMap::clean()
{
    if( _id )
        glDeleteFramebuffersEXT(1, &_id);
    _id = 0;

    if( _zId )
        glDeleteRenderbuffersEXT(1, &_zId);
    _zId = 0;

    if( _texId )
        glDeleteTextures(1, &_texId);
    _texId = 0;
}


//////////////////////////////////////////////////////////////////////////////

bool
GlossyMap::reloadShaders(const Environment& env)
{
    if( ! GLEW_EXT_geometry_shader4
     || ! GLEW_EXT_texture_array
     || ! GLEW_EXT_framebuffer_object )
    {
        if( !unsupported )
            qWarning("**** Cannot do glossy map without missing \n"
                     "     GL_EXT_geometry_shader4 or \n"
                     "     GL_EXT_texture_array\n"
                     "     GL_EXT_framebuffer_object\n"
                     "OpenGL extensions");
        unsupported = true;
        return false;
    }

    //------------------------------------------------------------------------

    QString envString = "#define SUBMODE_COLOR 1\n";
    envString += env.envString();

    QString dir = qApp->applicationDirPath() + "/Shaders/";
    //------------------------------------------------------------------------
    // Use a full featured phong shader to :
    //  - Apply glossy map to objects w/o special shaders...
    if( phongShader ) delete phongShader;
    phongShader = new GLShader(
        (dir + "phong_vertex.glsl").toLatin1().constData(),
        GL_TRIANGLES, (dir + "phong_geometry.glsl").toLatin1().constData(), GL_TRIANGLE_STRIP,
        (dir + "phong_fragment.glsl").toLatin1().constData(),
        envString.toLatin1().constData());

    if( ! phongShader->compile() )
    {
        delete phongShader;
        phongShader = 0;
        return false;
    }

    //------------------------------------------------------------------------
    // Use a full featured CosinusMap shader to :
    //  - Render scene to construct the glossy tex array.
    if( sphereShader ) delete sphereShader;
    sphereShader = new GLShader(
        (dir + "common_vertex.glsl").toLatin1().constData(),
        GL_TRIANGLES, (dir + "sphereBox_geometry.glsl").toLatin1().constData(), GL_TRIANGLE_STRIP,
        (dir + "sphere_fragment.glsl").toLatin1().constData(),
        envString.toLatin1().constData());
    sphereShader->setMaxOutput(4);

    if( ! sphereShader->compile() )
    {
        delete sphereShader;
        sphereShader = 0;
        return false;
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

bool GlossyMap::render(const Scene* scene, bool debug)
{
    if( unsupported ) return false;
    if( ! phongShader )
        if( ! reloadShaders( scene->env ) )
            return false;                   // Unsupported ?

    if( _zFar < 0.0f )
    {
        Vec3f bmin, bmax;
        scene->getBBox(bmin, bmax);
        float r1 = length(bmin - _center);
        float r2 = length(bmax - _center);
        _zFar = length(_u+_v) + ((r1 > r2) ? r1 : r2);
    }

    //------------------------------------------------------------------------
    // Creates the FBO and render buffer ids, if not already done.
    if( ! _id )
    {
    	glGenFramebuffersEXT(1, &_id);
        CHECK_GL_ERROR("alloc FBO");
    }

    if( ! _zId )
    {
        glGenRenderbuffersEXT(1, &_zId);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, _zId);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT,
                                 GL_DEPTH_COMPONENT24,
                                 _size, _size);
        CHECK_GL_ERROR("create depth renderbuffer");
    }

    //------------------------------------------------------------------------
    // Creates texture array.
    if( ! _texId )
    {
        glGenTextures(1, &_texId);
        glBindTexture(GL_TEXTURE_2D_ARRAY_EXT, _texId);
        glTexImage3D(GL_TEXTURE_2D_ARRAY_EXT,// target.
                0,                          // mip-map level.         
                GL_RGBA32F_ARB,             // Internal texture format.
                _size, _size,               // texture size. Should be a square.
                _nu * _nv,                  // 2 layers.
                0,                          // no border.
                GL_RGBA,                    // What pixel data should be loaded there ?
                GL_FLOAT,                   // Load floats.
                0);                         // And no initialisation data (yet).
        CHECK_GL_ERROR("create glossy tex array");

        glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D_ARRAY_EXT, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glBindTexture(GL_TEXTURE_2D_ARRAY_EXT, 0);
        CHECK_GL_ERROR("glossy env tex");
    }

    //------------------------------------------------------------------------
    // Just attach shared depth buffer to FBO
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
    glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT,
                                 GL_DEPTH_ATTACHMENT_EXT,
                                 GL_RENDERBUFFER_EXT,
                                 _zId);
    CHECK_GL_ERROR("attach depth renderbuffer");

    // ---- setup ------------------------------------------------------------

    glPushAttrib(GL_VIEWPORT_BIT);
    glViewport(0, 0, _size, _size);   // Origin & w*h

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    CHECK_GL_ERROR("push matrices");

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
    CHECK_GL_ERROR("Attach texture array");

    // ---- For each layer of the array --------------------------------------
    Vec3f n = _zFar * normalize( cross(_u, _v) );
    bool ok = true;
    for(int u=0; ok && u<_nu; ++u)
    {
        for(int v=0; ok && v<_nv; ++v)
        {
            int idx = v + u * _nv;
            Vec3f eye = _center
                      + _u * ( float(u)/float(_nu-1) - 0.5f)
                      + _v * ( float(v)/float(_nv-1) - 0.5f);

            // Viewing direction for that layer :
	        glMatrixMode(GL_MODELVIEW);
	        glLoadIdentity();
            gluLookAt(eye._x, eye._y, eye._z,
                      n._x + eye._x, n._y + eye._y, n._z + eye._z,
                      _v._x,   _v._y,   _v._z);
	        CHECK_GL_ERROR("glossy LookAt");

            // Make FBO to directly write results there :
            glFramebufferTextureLayerEXT(GL_FRAMEBUFFER_EXT,    // target
                                      GL_COLOR_ATTACHMENT0_EXT, // attachment
                                      _texId, 0,                // dest texture & mipmap level
                                      idx);                     // dest layer (aka zoffset)
            CHECK_GL_ERROR("attach glossy map layer");

            // Use a single rendering target for that RGBA image :
            glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
            glReadBuffer(GL_NONE);
            CHECK_GL_ERROR("draw to FBO");

            // Check everything went well :
            uint complete = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
            if( complete != GL_FRAMEBUFFER_COMPLETE_EXT )
            {
                qWarning("*** Uncomplete frame buffer. Code 0x%x", complete);
                ok = false;
                break;
            }

            // ---- Then Go --------------------------------------------------
            if( debug )
                glClearColor( !(u&1), !(v&2), !(idx&4), 1.0);
            else
                glClearColor( 0.0f, 0.0f, 0.0f, 1.0f);
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Make sure we start from a clear texture.

            _currentRender = true;
                scene->draw(sphereShader, 0.0001f, _zFar);
            _currentRender = false;
        }
    }
    if(ok)
        qDebug("Rendered glossy map");

    // ---- Dettach render-to-texture ----------------------------------------
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
    glDrawBuffer(GL_NONE);
    CHECK_GL_ERROR("detach glossy map layer");

    // ---- Restore previous environment -------------------------------------
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f);
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glPopAttrib();                                  // pop viewport.
    CHECK_GL_ERROR("unbind fbo");

    return ok;
}

//////////////////////////////////////////////////////////////////////////////

void GlossyMap::renderAll(const Scene* scene, bool debug)
{
    if( unsupported ) return;

    int n=0, f=0;
    double time = TBender::time();

    for(GlossyMaps::iterator it = allGlossyMaps.begin(); it != allGlossyMaps.end(); ++it, ++n)
    {
        if( (*it)->render(scene, debug) )
        {
            f += (*it)->_nu * (*it)->_nv;
            if( scene->env.dumpImages )
            {
                time += TBender::time();    // stop timer.

                QImage img;
                (*it)->getImage(img);
                img.save( QString("glossy%1Map.png").arg(n));

                time -= TBender::time();    // restart timer.
            }
        }
        else
            break;
    }

    time = TBender::time() - time;
    if( n > 0 )
        qWarning("Mean glossy map rendering time = %s (mean on %d maps, with %d elements each)",
            TBender::timeCString(time / n).toLatin1().constData(),
            n, f/n);
}

//////////////////////////////////////////////////////////////////////////////

void GlossyMap::setSize(int nu, int nv)
{
    clean();
    _nu = nu; _nv = nv;
}

//////////////////////////////////////////////////////////////////////////////

void GlossyMap::setResolution(int r)
{
    // Make sure the image is correctly initialized to hold our depth
    // data.
    static int maxRBSize = 0;
    if( ! maxRBSize )
        glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &maxRBSize);
    Q_ASSERT( 0 < r && r <= maxRBSize );

    //------------------------------------------------------------------------

    clean();
    _size = r;
}

//////////////////////////////////////////////////////////////////////////////

void GlossyMap::setPosition(const Vec3f& center, const Vec3f& u, const Vec3f& v)
{
    _center = center;
    _u      = u;
    _v      = v;
}

//////////////////////////////////////////////////////////////////////////////

bool GlossyMap::enable(int texUnit, const Shader* shader) const
{
    if( unsupported ) return false;

    if( _currentRender )    // Object self-culling during render()
        return true;

    if( ! _texId )
    {
        qDebug("Skiping glossy map: not yet rendered");
        return false;
    }

    glActiveTexture( GL_TEXTURE0 + texUnit );
    glBindTexture(GL_TEXTURE_2D_ARRAY_EXT, _texId);
    CHECK_GL_ERROR("bind glossy tex array");

    if( ! shader )
    {
        shader = phongShader;
        shader->enable();   // If a scene shader is used, it is already enabled.
    }

    // For any point P = C + x U + y V, we do have :
    // glossyMatrix * P = (a .(C + x U + y V) - a.(C-U), b.(C + x U + y U) -b.(C-U))
    //                  = (a . (x+1)U, b . (y+1) V)
    //                  = ( (x+1)/2 nu, (y+1)/2 nv )
    // Hence it maps (x,y) in [-1..1]x[-1..1] to [0..nu, 0..nv]
    Matrix4 glossyMatrix;
    Vec3f a = _u * ( 0.5f * _nu / length2(_u));
    Vec3f b = _v * ( 0.5f * _nv / length2(_v));
    glossyMatrix[0][0] = a._x;
    glossyMatrix[0][1] = a._y;
    glossyMatrix[0][2] = a._z;
    glossyMatrix[0][3] = - dot(a, _center-_u);
    glossyMatrix[1][0] = b._x;
    glossyMatrix[1][1] = b._y;
    glossyMatrix[1][2] = b._z;
    glossyMatrix[1][3] = - dot(b, _center-_v);

    // TODO glossyMatrix *= camera()->modelViewMatrix().inverse();

    shader->setUniform("uHasGlossyMap", true, GL_BOOL);
    shader->setUniform("glossyMap", _nu);
    shader->setUniform("uGlossyNV", _nv);
    shader->setUniform("uGlossyMatrix", (float*)glossyMatrix, 0x8B69); // GL_FLOAT_MAT4x2_EXT ???

    return true;
}

//////////////////////////////////////////////////////////////////////////////

void GlossyMap::disable(int texUnit, const Shader* shader) const
{
    if( unsupported ) return;

    if( _currentRender )    // Object self-culling during render()
        return;

    if( ! _texId ) return;

    glActiveTexture( GL_TEXTURE0 + texUnit );
    glBindTexture(GL_TEXTURE_2D_ARRAY_EXT, 0);
    CHECK_GL_ERROR("unbind glossy tex array");

    if( shader )
        shader->setUniform("uHasGlossyMap", false, GL_BOOL);
    else
    {
        phongShader->setUniform("uHasGlossyMap", false, GL_BOOL);
        phongShader->disable();
    }
}

//////////////////////////////////////////////////////////////////////////////



bool
GlossyMap::getImage(QImage& image) const
{
    if( unsupported ) return false;

    image = QImage(_nu*_size, _nv*_size, QImage::Format_RGB32);
    image.fill( qRgba(128,0,128,255) );     // Violet where no write occurs.

    // Note : There is no GetTexSubImage() neither GetTexImageLayer OpenGL call.
    //        Using GetTexImage() returns all layers at once, creating a 
    //        size x (nu*nv*size) very tall image.
    //        To be able to have a (nu*size) x (nv*size) image, we have to
    //        read layers one at a time. This is done by attaching
    //        each layer to the FBO, then reading pixels from there...

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
    glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
    CHECK_GL_ERROR("bind glossy map");

    glPixelStorei(GL_PACK_ROW_LENGTH, image.width());
    CHECK_GL_ERROR("pixelstore");

    for(int u=0; u<_nu; ++u)
        for(int v=0; v<_nv; ++v)
        {
            // Attach the corresponding layer of the texture array :
            int idx = v + u * _nv;
            glFramebufferTextureLayerEXT(GL_FRAMEBUFFER_EXT,    // target
                                      GL_COLOR_ATTACHMENT0_EXT, // attachment
                                      _texId, 0,                // dest texture & mipmap level
                                      idx);                     // dest layer (aka zoffset)
            CHECK_GL_ERROR("attach glossy map layer");

            void* p = image.scanLine(v*_size) + 4*u*_size;
            glReadPixels(0, 0,              // Raster position within this layer.
                         _size, _size,      // Raster size
                         GL_BGRA,           // Format
                         GL_UNSIGNED_BYTE,  // Type
                         p);                // Buffer ptr
            CHECK_GL_ERROR("get glossy mosaic");
        }

    glPixelStorei(GL_PACK_ROW_LENGTH, 0);
    glReadBuffer(GL_NONE);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    CHECK_GL_ERROR("unbind glossy map");

    return true;
}
