////////////////////////////////////////////////////////////////////////
/*! \file   GlossyMap.h
 *  \brief  Glossy plan, has a 2D array of spheremaps.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa FrameBufferObject, Scene, Shader.
 */
////////////////////////////////////////////////////////////////////////
// $Id: GlossyMap.h,v 1.7 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2007/04/19] : Creation.

#include "FrameBufferObject.h"
#include "Math/Vec3f.h"

class Shader;
class Scene;
class Environment;

class GlossyMap : private FrameBufferObject
{
    int _nu, _nv;               //!< Number of sphere maps, in both directions.
    Vec3f _center;              //!< Grid's position. In Scene coordinates.
    float _zFar;                //!< Scene radius, as seen from glossy map.
    Vec3f _u, _v;               //!< grid half size. In Scene coordinates.

    int _size;                  //!< Size of the individual textures.
    uint _texId;

    static Shader* phongShader; //!< Apply glossy maps on shaderless objects.
    static Shader* sphereShader;//!< Render glossy maps.

    bool _currentRender;        //!< This glossy map is currently rendering. Cull object.

    void clean();               //!< Free internal texture array.
    bool render(const Scene*, bool debug);  //!< Allocate & compute internal texture array.

public:
    GlossyMap();
    ~GlossyMap();

    //! Reload internal shaders (Scene environment have changed).
    static bool reloadShaders(const Environment& env);

    //! Re-render all GlossyMaps.
    static void renderAll(const Scene* scene, bool debug);

    //! Change number of textures in the grid. Reset all data.
    void setSize(int, int);

    //! Change resolution of individual textures. Reset all data.
    void setResolution(int);

    //! Change map's center, orientation and size.
    void setPosition(const Vec3f& center, const Vec3f& u, const Vec3f& v);

    //! Setup a shader to start texturing with the glossy map.
    bool enable(int texUnit, const Shader*) const;

    //! Back to normal rendering mode.
    void disable(int texUnit, const Shader*) const;

    //! Returns a mosaic image of nu * nv individual 360 images.
    bool getImage(QImage&) const;
};
