////////////////////////////////////////////////////////////////////////
/*! \file   HemiCube.cpp
 *  \brief  Hemicube for an lighting rendering.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Scene, Shader, FrameBufferObject, EnvironmentMap.
 */
////////////////////////////////////////////////////////////////////////
// $Id: HemiCube.cpp,v 1.5 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2007/12/07] : Creation.

#include "HemiCube.h"

#include <GL/glew.h>
#include <QImage>
#include "Utilities.h"

////////////////////////////////////////////////////////////////////////

HemiCube::HemiCube(Mode mode, const Vec3f& eye, int size, float radius)
: EnvironmentMap( EnvironmentMap::Mode(mode), eye, size, radius)
{}

HemiCube::~HemiCube()
{}

bool HemiCube::enable(int, const Shader*) const
{
    return false;
}

void HemiCube::disable(int, const Shader*) const
{}

////////////////////////////////////////////////////////////////////////

int
HemiCube::faces()
{
    if( _mode == FACES_MAP )
        return 5;
    else if(_mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP)
        return 1;
    else
        qFatal("Wrong faces in HemiCube");
    return 0;
}

//////////////////////////////////////////////////////////////////////////////

void 
HemiCube::createTexture()
{
    // ---- Texture creation -------------------------------------------------
    // Because texture mode might have changed between CUBE_MAP and 2D_ARRAY,
    //         we have to destroy and recreate a new one at each render.
    // Note: reloading shaders destroy all texture, just to make sure it will
    //       be redrawn. Hence generate it *AFTER* loading shaders !
    if( _texId ) glDeleteTextures(1, &_texId);
    glGenTextures(1, &_texId);
    CHECK_GL_ERROR("creat tex");

    // ==== FACES_MAP case ====================================================

    if( _mode == FACES_MAP )
    {
        // Current vube map limitations:
        //   - should define the 6 faces to be complete.
        //   - each face should be square.
        // Hence no way to optimize this vs. a full cube map....
        EnvironmentMap::createTexture();
    }
    // ==== PARABOLOID_MAP case ==============================================
    else if(_mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP)
    {
        // Allocate texture space for both faces :
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, _texId);
        glTexImage2D(GL_TEXTURE_2D,         // target.
                0,                          // mip-map level.         
                GL_RGBA,                    // Internal texture format.
                _size, _size,               // texture size. Should be a square.
                0,                          // no border.
                GL_RGBA,                    // What pixel data should be loaded there ?
                GL_FLOAT,                   // Load floats.
                0);                         // And no initialisation data (yet).
        CHECK_GL_ERROR("create spheremap tex");

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glBindTexture(GL_TEXTURE_2D, 0);
        CHECK_GL_ERROR("mipmap spheremap tex");
    }
    else
        qFatal("Wrong mode in HemiCube texture creation");
}

//////////////////////////////////////////////////////////////////////////////

bool
HemiCube::getImage(QImage& image) const
{
    if( _mode == FACES_MAP )
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, _texId);
        CHECK_GL_ERROR("bind cube map");

        image = QImage(2*_size, 2*_size, QImage::Format_RGB32);
        image.fill( qRgba(128,0,128,255) );     // Violet where no write occurs.

        glPixelStorei(GL_PACK_ROW_LENGTH, image.width());   // in pixels, not bytes.
        for(int face=0; face<5; ++face)
        {
            int x = _size*(face / 3);
            int y = ((_size+1)/2) * (face % 3);
            GLuint order[5] = {
                GL_TEXTURE_CUBE_MAP_POSITIVE_X,
                GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
                GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
                GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
                GL_TEXTURE_CUBE_MAP_POSITIVE_Z
            };

            uchar *ptr = image.scanLine(y) + 4 * x;         // in bytes, not pixels !
            glGetTexImage(order[face],
                          0,                    // level
                          GL_BGRA,              // Qt's byte order (on win32 at least)
                          GL_UNSIGNED_BYTE,
                          ptr);
            CHECK_GL_ERROR("get cube face");
        }
        glPixelStorei(GL_PACK_ROW_LENGTH, 0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
        CHECK_GL_ERROR("unbind cube map");

        // Crop bottom part of the too big image:
        image = image.copy(0, 0, 2*_size, 3*((_size+1)/2));
    }
    else if(_mode == SPHERE_MAP || _mode == PARABOLA_MAP || _mode == LAMBERT_MAP)
    {
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_2D, _texId);
        CHECK_GL_ERROR("bind sphere map");

        image = QImage(_size, _size, QImage::Format_RGB32);
        image.fill( qRgba(128,0,128,255) );     // Violet where no write occurs.

        // A uniq GetTexImage do returns all layers at once :
        glGetTexImage(GL_TEXTURE_2D,
                      0,                 	    // level
                      GL_BGRA,                  // Qt's byte order (on win32 at least)
                      GL_UNSIGNED_BYTE,
                      image.scanLine(0));
        CHECK_GL_ERROR("get parab image");

        glBindTexture(GL_TEXTURE_2D, 0);
        CHECK_GL_ERROR("unbind parab map");
    }
    else
        qFatal("Wrong hemicube get image");

    return true;
}
