////////////////////////////////////////////////////////////////////////
/*! \file   Math/Math.cpp
 *  \brief  Implementation of non inline parts of Math files
 *  \author Eric Ferley, Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * $Id: Math.cpp,v 1.6 2008/02/20 11:49:46 gascuel Exp $
 */
////////////////////////////////////////////////////////////////////////
// File History:

#include "Math.h"

#include "Vec2f.h"
#include "Vec3f.h"
#include "Vec4f.h"
#include "Quaternion.h"
#include "Matrix4.h"

///////////////////////////////////////////////////////////////
//                                                           //
//           Quaternion implementation                       //
//                                                           //
///////////////////////////////////////////////////////////////

void
Quaternion::fromAxis(const Vec3f& v, float sina, float cosa)
{
    float l2 = length2(v);
    if (l2>0)
    {
        float div = sina / sqrtf(l2);
        _q0 = v._x * div;
        _q1 = v._y * div;
        _q2 = v._z * div;
        _q3 = cosa;
    }
    else
    {
        reset();
    }
} 

void
Quaternion::fromAxis(const Vec3f& v, float angle)
{
    fromAxis(v, sinf(angle/2.0f), cosf(angle/2.0f));
}

float 
Quaternion::angle() const
{
    // float sina2 = length(axis());
    // return 2.0f * atan2( length(axis()),  _q3 );
    return 2.0f * acosf( _q3 );
}

Quaternion::Quaternion(const Vec3f& axis, float sina, float cosa)
    : _q0(0.0f), _q1(0.0f), _q2(0.0f), _q3(0.0f)
{
    fromAxis(axis, sina, cosa);
}

Quaternion::Quaternion(const Vec3f& axis, float angle)
    :  _q0(0.0f), _q1(0.0f), _q2(0.0f), _q3(0.0f)
{
    fromAxis(axis, angle);
}

void
Quaternion::set(const Vec3f& rfrom, const Vec3f& rto)
{
    // snippet from SbRotation.c++ (Inventor)
    Vec3f from(::normalize(rfrom));
    Vec3f to(::normalize(rto));

    float cost = dot(from, to);
    if (cost >= 1.0f)
    {
        _q0=_q1=_q2=0.0; _q3=1; // vector are // ..  Id rotation
        return;
    }
    if (cost <= -1.0f)
    {
        // vectors are opposite .. build any vector perpendicular to the
        // 1st to rotate around.
        Vec3f axis( cross(from, Vec3f(1,0,0)) );
        if (length(axis)<= FLT_EPSILON )
        {
            axis = cross(from, Vec3f(0,1,0)); 
            if (length(axis)<= FLT_EPSILON )
            {
                axis = cross(from, Vec3f(0,0,1));
            }
        }
        axis.normalize();
        _q0 = axis._x; _q1=axis._y; _q2=axis._z; _q3=0.0f;
        return;
    }
    Vec3f axis( ::normalize(cross( from, to )) );
	 
   // half-angle formula: sin^2(t) = (1 - cos(2t))/2
    axis *= sqrtf(0.5f*(1.0f-cost));
    _q0 = axis._x; _q1=axis._y; _q2=axis._z;
    // et aussi: cos^2(t) = (1 + cos(2t))/2
    _q3 = sqrtf(0.5f*(1.0f + cost));
}

Quaternion::Quaternion(const Vec3f& from, const Vec3f& to)
{
    set(from,to);
}

void
Quaternion::fromMatrix(const Matrix4& m)
{
//// EF: 26 fvrier 2004
// Le calcul diffrent des quaternions selon les config de matrice
    // produit des "saut" quand on effectue les rotation
    // passage brutal de (x,y,z,w)  (-x,-y,-z,-w) quaternion quivalent
    // ... mais qui ne supporte pas le filtrage composante par
    // composante:
    // i.e. retarded_q[i] += dq[i] * dt
    // => qui entraine un tours complet ...

   // First, find largest diagonal in matrix:
    int i = 2;
    if (m._m[0][0] > m._m[1][1])
    {
        if (m._m[0][0] > m._m[2][2])
        {
            i = 0;
        } 
    }
    else
    {
        if (m._m[1][1] > m._m[2][2])
        {
            i = 1;
        }
    }
    if (m._m[0][0]+m._m[1][1]+m._m[2][2] > m._m[i][i])
    {
        // Compute w first:
        _q3 =  sqrtf(m._m[0][0]+m._m[1][1]+m._m[2][2]+m._m[3][3])/2.0f;
        // And compute other values:
        _q0 = (m._m[2][1]-m._m[1][2])/(4.0f*_q3);
        _q1 = (m._m[0][2]-m._m[2][0])/(4.0f*_q3);
        _q2 = (m._m[1][0]-m._m[0][1])/(4.0f*_q3);
    }
    else
    {
        // Compute x, y, or z first:
        int j = (i+1)%3; 
        int k = (i+2)%3;
		 
        float* q = &_q0;
        // Compute first value:
        q[i] = sqrtf(m._m[i][i]-m._m[j][j]-m._m[k][k]+m._m[3][3])*0.5f;
		 
	  // And the others:
        q[j] = (m._m[j][i]+m._m[i][j])/(4.0f*q[i]);
        q[k] = (m._m[k][i]+m._m[i][k])/(4.0f*q[i]);
        _q3  = (m._m[k][j]-m._m[j][k])/(4.0f*q[i]);
    }
}

Quaternion::Quaternion(const Matrix4& m)
    : _q0(0.0f), _q1(0.0f), _q2(0.0f), _q3(0.0f)
{
    fromMatrix(m);
}

Vec3f
Quaternion::toEulerRadians() const
{
    // Using Euler angles (x,y,z), we get the following matrices:
    //
    //     | 1  0  0  |    | cy 0 sy |    | cz -sz 0 |
    //  Rx=| 0 cx -sx | Ry=|  0 1  0 | Rz=| sz  cz 0 |
    //     | 0 sx  cx |    |-sy 0 cy |    |  0   0 1 |
    //
    // Matrix multiplications gives
    //
    //            |    (cy.cz)           (-cy.sz)        sy   |
    //   Rx.Ry.Rz=|( sx.sy.cz+cx.sz) (-sx.sy.sz+cx.cz) -sx.cy |
    //            |(-cx.sy.cz+sx.sz) ( cx.sy.sz+sx.cz)  cx.cy |
    //
    // Thus, identifying with the quaternion matrix gives:
    //
    //  sy    = 2(q2q0 + q1q3)
    //
    //  -cy.sz = 2(q0q1 - q2q3)
    //   cy.cz = 1 - 2(q1q1 + q2q2)  => tan(z)
    //
    //  -sx.cy = 2(q1q2 - q0q3)
    //   cx.cy = 1 - 2(q1q1 + q0q0)  => tan(x)
    //
    // What if cy==0 ? (i.e. y= (+/-) PI/2 ????
    // 
    float sy = 2.0f*(_q0*_q2 + _q1*_q3);
    if (fabs(1-fabs(sy))<1e-5) // sy == 1 => cy==0
    {
        // => cy==0 !!! singular case
        // We get:
        //            |        0                0           1 |
        //   Rx.Ry.Rz=|( sx.sy.cz+cx.sz) (-sx.sy.sz+cx.cz)  0 |
        //            |(-cx.sy.cz+sx.sz) ( cx.sy.sz+sx.cz)  0 |
        // we can only exploit:
        //
        //    sin( x+z ) =   sx.cz + cxsz
        //    cos( x+z ) =   cx.cz - sx.sz 
        //  having:
        //            |     0         0      1 |
        //   Rx.Ry.Rz=|  sin(x+z)  cos(x+z)  0 |
        //            | -cos(x+z)  sin(x+z)  0 |
        // 
        // So, arbitrarily, we fix z=0, and deduce x !
        return Vec3f(atan2( 2.0f * (_q0*_q1 + _q2*_q3), 
                            -_q0*_q0 + _q1*_q1 - _q2*_q2 + _q3*_q3 ),
                     ( sy < 0 ? float(-M_PI/2) : float(M_PI/2) ),
                     0);
    }
    // La  fonction  atan2()  calcule l'Arc Tangente du rapport
    // des deux vari-ables x et y.  Cela revient a calculer l'Arc
    // Tangente de y  /  x,  mais les signes des deux variables
    // sont utilises pour determiner le quadrant du resultat.
    float rx=atan2(2.0f*(_q0*_q3 - _q1*_q2),
                   -_q0*_q0 - _q1*_q1 + _q2*_q2 + _q3*_q3);
    // Ensure rx is in the range [0:2PI]
    while (rx<0)      { rx += 2.0f*M_PI; }
    while (rx>2*M_PI) { rx -= 2.0f*M_PI; }
    float ry = asinf( sy );
    // Ry is in the range [-PI:PI] by definition (arcsin)
    float rz = atan2( 2.0f*(_q2*_q3 - _q0*_q1),
                      _q0*_q0 - _q1*_q1 - _q2*_q2 + _q3*_q3 );
    // Ensure rz is in the range [-PI:PI]
    while (rz<-M_PI) { rz += 2.0f*M_PI; }
    while (rz> M_PI) { rz -= 2.0f*M_PI; }
    return Vec3f(rx,ry,rz);
}


void
Quaternion::fromEulerRadians(const Vec3f& angles)
{
    // Composing Quaternions
    //   Rx = ( sin(x/2), 0, 0, cos(x/2) )
    //   Ry = ( 0, sin(y/2), 0, cos(y/2) )
    //   Rz = ( 0, 0, sin(z/2), cos(z/2) )
    float cx = cos(angles[0]/2.0f);
    float sx = sin(angles[0]/2.0f);
    float cy = cos(angles[1]/2.0f);
    float sy = sin(angles[1]/2.0f);
    float cz = cos(angles[2]/2.0f);
    float sz = sin(angles[2]/2.0f);
    
    _q0 = sx*cy*cz + cx*sy*sz;
    _q1 = cx*sy*cz - sx*cy*sz;
    _q2 = cx*cy*sz + sx*sy*cz;
    _q3 = cx*cy*cz - sx*sy*sz;
}

Vec3f
Quaternion::toEulerDegree() const
{
    return toEulerRadians()*(180.0f/M_PI);
}

void
Quaternion::fromEulerDegree(const Vec3f& a)
{
    fromEulerRadians(a*(M_PI/180.0f));
}

///////////////////////////////////////////////////////////////
//                                                           //
//           Matrix4 implementation                          //
//                                                           //
///////////////////////////////////////////////////////////////


void
Matrix4::setScale(const Vec3f& scale)
{
    _m[0][0] = scale._x;
    _m[1][1] = scale._y;
    _m[2][2] = scale._z;
    _m[3][3] = 1;
}

Vec3f
Matrix4::getScale() const
{
    return Vec3f(_m[0][0], _m[1][1], _m[2][2]);
}


void
Matrix4::setTranslation(const Vec3f& v)
{
    _m[3][0] = v._x;
    _m[3][1] = v._y;
    _m[3][2] = v._z;
}

void
Matrix4::addTranslation(const Vec3f& v)
{
    _m[3][0] += v._x;
    _m[3][1] += v._y;
    _m[3][2] += v._z;
}

Vec3f
Matrix4::getTranslation() const 
{
    return Vec3f( _m[3][0], _m[3][1], _m[3][2]);
}


Quaternion
Matrix4::getRotation() const
{
    return Quaternion(*this);
}

void
Matrix4::setRotation(const Quaternion& rot)
{
    _m[0][0] = rot._q0*rot._q0 - rot._q1*rot._q1 - rot._q2*rot._q2 + 
        rot._q3*rot._q3;
    _m[1][0] =        2.0f*(rot._q0*rot._q1 - rot._q2*rot._q3);
    _m[2][0] =        2.0f*(rot._q2*rot._q0 + rot._q1*rot._q3);

    _m[0][1] =        2.0f*(rot._q0*rot._q1 + rot._q2*rot._q3);
    _m[1][1] = - rot._q0*rot._q0 + rot._q1*rot._q1 - rot._q2*rot._q2 + 
        rot._q3*rot._q3;
    _m[2][1] =        2.0f*(rot._q1*rot._q2 - rot._q0*rot._q3);

    _m[0][2] =        2.0f*(rot._q2*rot._q0 - rot._q1*rot._q3);
    _m[1][2] =        2.0f*(rot._q1*rot._q2 + rot._q0*rot._q3);
    _m[2][2] = - rot._q0*rot._q0 - rot._q1*rot._q1 + rot._q2*rot._q2 + 
        rot._q3*rot._q3;

//  	 _m[0][3] = 0.0f;
//  	 _m[1][3] = 0.0f;
//  	 _m[2][3] = 0.0f;

//  	 _m[3][0] = 0.0f;
//  	 _m[3][1] = 0.0f;
//  	 _m[3][2] = 0.0f;
//  	 _m[3][3] = 1.0f;
}

Matrix4
Matrix4::scale( const Vec3f& s )
{
    Matrix4 tmp; tmp.setScale( s ); return tmp;
}

Matrix4
Matrix4::translate( const Vec3f& t )
{
    Matrix4 tmp; tmp.setTranslation( t ); return tmp;
}

Matrix4
Matrix4::rotate( const Quaternion& r )
{
    Matrix4 tmp; tmp.setRotation( r ); return tmp;
}

Vec3f
Matrix4::transformOf(const Vec3f& src) const
{
    // Rotate without translation
    return Vec3f( _m[0][0]*src._x + _m[1][0]*src._y + _m[2][0]*src._z,
                  _m[0][1]*src._x + _m[1][1]*src._y + _m[2][1]*src._z,
                  _m[0][2]*src._x + _m[1][2]*src._y + _m[2][2]*src._z);
}

Quaternion
Matrix4::transformOf(const Quaternion& src) const
{
    // Rotate without quaternion
    //  - Treat axis (q0,q1,q2) as vector xform.
    //  - Copy angle (q3).
    // Note: Post normalize left for programmer choice.
    Vec3f axis(src._q0, src._q1, src._q2);
    float sina = length(axis);
    axis = transformOf(axis);
    float len = length(axis);
    if (len != 0)
    {
        axis *= (sina/len);
    }
    return Quaternion(axis[0], axis[1], axis[2], src._q3);

//      return Quaternion( _m[0][0]*src._q0 + _m[1][0]*src._q1 + _m[2][0]*src._q2,
//                         _m[0][1]*src._q0 + _m[1][1]*src._q1 + _m[2][1]*src._q2,
//                         _m[0][2]*src._q0 + _m[1][2]*src._q1 + _m[2][2]*src._q2,
//                         src._q3 );
}

Vec3f 
Matrix4::coordinatesOf(const Vec3f& src) const
{
    // rotate *and* translate
    return Vec3f( _m[0][0]*src._x + _m[1][0]*src._y + _m[2][0]*src._z + _m[3][0],
                  _m[0][1]*src._x + _m[1][1]*src._y + _m[2][1]*src._z + _m[3][1],
                  _m[0][2]*src._x + _m[1][2]*src._y + _m[2][2]*src._z + _m[3][2]);
}

Matrix4
Matrix4::inverse() const
{
    //
    // Code adpat de SbMatrix L. 697, cf. SbMatrix::affine_inverse
    //

    if ( (_m[0][3] != 0.0) || (_m[1][3] != 0.0) ||
         (_m[2][3] != 0.0) || (_m[3][3] != 1.0))
    {
        // matrice non affine ... besoin de LU decomposition .. future

        // return inverse(); // general matrix inversion
        return Matrix4(); // identit en attendant
    }

    double det = det3();
    // matrice singulire
    if (det ==  0)
		return Matrix4();

    double m00 =  (_m[1][1]*_m[2][2] - _m[1][2]*_m[2][1])/det;
    double m10 = -(_m[1][0]*_m[2][2] - _m[2][0]*_m[1][2])/det;
    double m20 =  (_m[1][0]*_m[2][1] - _m[1][1]*_m[2][0])/det;

    double m01 = -(_m[0][1]*_m[2][2] - _m[0][2]*_m[2][1])/det;
    double m11 =  (_m[0][0]*_m[2][2] - _m[0][2]*_m[2][0])/det;
    double m21 = -(_m[0][0]*_m[2][1] - _m[0][1]*_m[2][0])/det;

    double m02 =  (_m[0][1]*_m[1][2] - _m[0][2]*_m[1][1])/det;
    double m12 = -(_m[0][0]*_m[1][2] - _m[0][2]*_m[1][0])/det;
    double m22 =  (_m[0][0]*_m[1][1] - _m[0][1]*_m[1][0])/det;

    return Matrix4( 
        float(m00), float(m10), float(m20), -float( _m[3][0]*m00 + _m[3][1]*m10 + _m[3][2]*m20 ),
        float(m01), float(m11), float(m21), -float( _m[3][0]*m01 + _m[3][1]*m11 + _m[3][2]*m21 ),
        float(m02), float(m12), float(m22), -float( _m[3][0]*m02 + _m[3][1]*m12 + _m[3][2]*m22 ),
        0.0f, 0.0f, 0.0f, 1.0f);
}




