////////////////////////////////////////////////////////////////////////
/*! \file   Math/Matrix4.h
 *  \brief  Another Matrix4 class
 *  \author Eric Ferley, Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Vec3f, Quaternion.
 *
 * $Id $
 */
////////////////////////////////////////////////////////////////////////
// File History:

#ifndef MATRIX4_H
#define MATRIX4_H

#include "Math.h"
#include "Vec3f.h"
#include "Quaternion.h"

/*! \brief another 4x4 Matrix class
	\ingroup Math

	Matrix4 implements 3D transformations.

	The matrix is stored in a GL fashion (transposed of Performer/Inventor).
	This means that the translation is stored in a30,a31,a32.

    \b Note:
	 - xformVec() only the rotation/scale xform is applied.
	 - xformPt() also translates the argument.
	 - invert() still un-tested.

	\sa Math.cpp for implementation of non-inline functions.
 */
class Matrix4
{
    public:
        float  _m[4][4];
        // default constructor of Identity Matrix4
        inline Matrix4();
        inline Matrix4(float a00, float a10, float a20, float a30,
                       float a01, float a11, float a21, float a31,
                       float a02, float a12, float a22, float a32,
                       float a03, float a13, float a23, float a33);
        // inline Matrix4(const Matrix4& mat);
        template <typename M>
        explicit inline Matrix4(M&);
	  
        inline void reset();                         // == reset to identity
        inline void setScale(float scale);     // only modifies scale part
        void setScale(const Vec3f& scale); 
        Vec3f getScale() const;
			
        void setTranslation(const Vec3f& transl); // only edits trans part
        void addTranslation(const Vec3f& transl);
        Vec3f getTranslation() const;

        void setRotation(const Quaternion& rot); // only edits rot part
        Quaternion getRotation() const;

        inline Matrix4& operator=(const Matrix4& a);


        // access
        inline const float* operator[](int i) const { return &(_m[i][0]); }
        inline float* operator[](int i) { return &(_m[i][0]); }
			
        // Cast
        inline operator const float*() const { return &_m[0][0]; }
        inline operator float*() { return &_m[0][0]; }

        Matrix4 inverse() const;
        inline Matrix4 transpose() const; 
        inline Matrix4 ortho_inverse() const; 
        inline float det3() const;

        /*!
          Vector transformation (do NOT use translationnal part)
         */
        Vec3f transformOf(const Vec3f& src) const;
        /*!
          Point transform (adds translationnal part)
         */
        Vec3f coordinatesOf(const Vec3f& src) const;
        /*! Rotate without quaternion
            - Treat axis (q0,q1,q2) as vector xform.
            - Copy angle (q3).
          Note: Post normalize left for programmer choice.
        */
        Quaternion transformOf(const Quaternion& src) const;
			
        // Explicit constructor of translation/scale/rot matrices
        static Matrix4 scale( const Vec3f& s );
        static Matrix4 translate( const Vec3f& t );
        static Matrix4 rotate( const Quaternion& r );
        // any use for explicit Matrix4::translate, rotate, etc ...

};

// Matrix composition
inline Matrix4 operator*(const Matrix4& a, const Matrix4& b);
inline Matrix4 operator+(const Matrix4& a, const Matrix4& b);
inline Matrix4 operator*(const Matrix4& a, float b);
inline Matrix4 operator*(float a, const Matrix4& b);
inline Matrix4 operator/(const Matrix4& a, float b);

////////////////////////////////////////////////////////////////////////
//
// Inline Matrix4 Implementation
//
////////////////////////////////////////////////////////////////////////

template<typename M>
inline Matrix4::Matrix4(M& m)
{
    for (int i=0;i<4;++i)
    {
        for (int j=0;j<4;++j)
        {
            _m[i][j] = m[i][j];
        }
    }
}


inline Matrix4& 
Matrix4::operator=(const Matrix4& a)
{
    for (int i=0; i<4; ++i)
    {
        for (int j=0; j<4; ++j)
        {
            _m[i][j] = a._m[i][j];
        }
    }
    return *this;
}

inline
Matrix4::Matrix4(float a00, float a10, float a20, float a30,
				 float a01, float a11, float a21, float a31,
				 float a02, float a12, float a22, float a32,
				 float a03, float a13, float a23, float a33)
{
    _m[0][0] = a00; _m[1][0] = a10; _m[2][0] = a20; _m[3][0] = a30;
    _m[0][1] = a01; _m[1][1] = a11; _m[2][1] = a21; _m[3][1] = a31;
    _m[0][2] = a02; _m[1][2] = a12; _m[2][2] = a22; _m[3][2] = a32;
    _m[0][3] = a03; _m[1][3] = a13; _m[2][3] = a23; _m[3][3] = a33;
}

inline
Matrix4::Matrix4()
{
    reset();
}

inline void
Matrix4::reset()
{
    _m[0][0] = 1.0f;
    _m[1][0] = _m[2][0] = _m[3][0] =_m[0][1] = 0.0f;
    _m[1][1] = 1.0f;
    _m[2][1] = _m[3][1] = _m[0][2] = _m[1][2] = 0.0f;
    _m[2][2] = 1.0f;
    _m[3][2] = _m[0][3] = _m[1][3] = _m[2][3] = 0.0f;
    _m[3][3] = 1.0f;
}

inline void
Matrix4::setScale(float scale)
{
    _m[0][0] = scale;
    _m[1][1] = scale;
    _m[2][2] = scale;
    _m[3][3] = 1;
}


inline Matrix4
operator*(const Matrix4& a, const Matrix4& b)
{
#define _MXMUL_(i,j) ( a._m[0][j]*b._m[i][0] + \
                       a._m[1][j]*b._m[i][1] + \
                       a._m[2][j]*b._m[i][2] + \
                       a._m[3][j]*b._m[i][3] )
    return Matrix4( _MXMUL_(0,0), _MXMUL_(1,0), _MXMUL_(2,0), _MXMUL_(3,0),
                    _MXMUL_(0,1), _MXMUL_(1,1), _MXMUL_(2,1), _MXMUL_(3,1), 
                    _MXMUL_(0,2), _MXMUL_(1,2), _MXMUL_(2,2), _MXMUL_(3,2), 
                    _MXMUL_(0,3), _MXMUL_(1,3), _MXMUL_(2,3), _MXMUL_(3,3) );
#undef _MXMUL_
}

inline Matrix4 
operator*(const Matrix4& a, float b)
{
    return Matrix4( a._m[0][0] * b, a._m[1][0] * b, a._m[2][0] * b, a._m[3][0] * b, 
                    a._m[0][1] * b, a._m[1][1] * b, a._m[2][1] * b, a._m[3][1] * b, 
                    a._m[0][2] * b, a._m[1][2] * b, a._m[2][2] * b, a._m[3][2] * b, 
                    a._m[0][3] * b, a._m[1][3] * b, a._m[2][3] * b, a._m[3][3] * b);
}

inline Matrix4 
operator*(float a, const Matrix4& b)
{
    return b*a;
}

inline Matrix4 
operator/(const Matrix4& a, float b)
{
    Q_ASSERT(b != 0.0f);

    return Matrix4( a._m[0][0] / b, a._m[1][0] / b, a._m[2][0] / b, a._m[3][0] / b, 
                    a._m[0][1] / b, a._m[1][1] / b, a._m[2][1] / b, a._m[3][1] / b, 
                    a._m[0][2] / b, a._m[1][2] / b, a._m[2][2] / b, a._m[3][2] / b, 
                    a._m[0][3] / b, a._m[1][3] / b, a._m[2][3] / b, a._m[3][3] / b);
}

inline Matrix4 
operator+(const Matrix4& a, const Matrix4& b)
{
    return Matrix4( a._m[0][0] * b._m[0][0], a._m[1][0] * b._m[1][0], a._m[2][0] * b._m[2][0], a._m[3][0] * b._m[3][0], 
                    a._m[0][1] * b._m[0][1], a._m[1][1] * b._m[1][1], a._m[2][1] * b._m[2][1], a._m[3][1] * b._m[3][1], 
                    a._m[0][2] * b._m[0][2], a._m[1][2] * b._m[1][2], a._m[2][2] * b._m[2][2], a._m[3][2] * b._m[3][2], 
                    a._m[0][3] * b._m[0][3], a._m[1][3] * b._m[1][3], a._m[2][3] * b._m[2][3], a._m[3][3] * b._m[3][3]);
}

inline Matrix4
Matrix4::transpose() const
{
    return Matrix4(
        _m[0][0], _m[0][1], _m[0][2], _m[0][3],
        _m[1][0], _m[1][1], _m[1][2], _m[1][3],
        _m[2][0], _m[2][1], _m[2][2], _m[2][3],
        _m[3][0], _m[3][1], _m[3][2], _m[3][3]);
}

inline Matrix4
Matrix4::ortho_inverse() const
{
    // inversion d'un matrice ortho[gonale|normale]
    // l'inverse de la rotation est la transpose, et la translation
    // est la translation oppose.
    return Matrix4(
        _m[0][0], _m[0][1], _m[0][2],
        -_m[3][0]*_m[0][0] - _m[3][1]*_m[0][1] - _m[3][2]*_m[0][2],
        _m[1][0], _m[1][1], _m[1][2],
        -_m[3][1]*_m[1][0] - _m[3][1]*_m[1][1] - _m[3][2]*_m[1][2],
        _m[2][0], _m[2][1], _m[2][2],
        -_m[3][2]*_m[2][0] - _m[3][1]*_m[2][1] - _m[3][2]*_m[2][2],
        _m[0][3], _m[1][3], _m[2][3],  _m[3][3]);
}

inline float 
Matrix4::det3() const
{
    float pos = 0.0;
    float neg = 0.0;

    // Determinant (Sarrus's method)
    float tmp = _m[0][0]*_m[1][1]*_m[2][2];
    if (tmp>=0.0) pos += tmp; else neg += tmp;
    tmp =  _m[0][1]*_m[1][2]*_m[2][0];
    if (tmp>=0.0) pos += tmp; else neg += tmp;
    tmp =  _m[0][2]*_m[1][0]*_m[2][1];
    if (tmp>=0.0) pos += tmp; else neg += tmp;
    tmp = -_m[0][2]*_m[1][1]*_m[2][0];
    if (tmp>=0.0) pos += tmp; else neg += tmp;
    tmp = -_m[0][1]*_m[1][0]*_m[2][2];
    if (tmp>=0.0) pos += tmp; else neg += tmp;
    tmp = -_m[0][0]*_m[1][2]*_m[2][1];
    if (tmp>=0.0) pos += tmp; else neg += tmp;

    float det = pos+neg;
    return ( fabs( det / (pos-neg) ) < 1.0e-15f ) ? 0.0f : det;
}

#endif // MATRIX4_H
