////////////////////////////////////////////////////////////////////////
/*! \file   Math/Quaternion.h
 *  \brief  Another Quaternion class
 *  \author Eric Ferley, Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * $Id: Quaternion.h,v 1.3 2008/02/20 11:49:46 gascuel Exp $
 */

#ifndef QUATERNION_H
#define QUATERNION_H

#include "Math.h"
#include "Vec3f.h"

class Matrix4;

/*! \brief another Quaternion class
	\ingroup Math

	Quaternion implements 3D rotations

	The first three components represent the rotation axis. Let Phi be the
	rotation angle.
	Then, (q[0],q[1],q[2]) is the rotation axis. Its norm is sin(Phi/2).
	q[3]==cos(Phi/2).

	Rotations composition operator are + and -

	Rotations can be constructed from Matrix4, couple (axis,angle) or couple
	(vector from, vector to).

	see Math.cpp for implementation of non-inline functions.
 */
class Quaternion
{
public:
    float _q0, _q1, _q2, _q3;
    
    inline Quaternion();
    inline Quaternion(float q0, float q1, float q2, float q3);
    Quaternion(const Vec3f& axis, float sin_half_a, float cos_half_a);
    Quaternion(const Vec3f& axis, float angle);
    Quaternion(const Matrix4&);
    Quaternion(const Vec3f& from, const Vec3f& to);

    // inline Quaternion(const Quaternion& src);
    
    // Template conversion from other Quaternion implementation
    // (e.g. qglviewer::Quaternion)
    template<class Q> explicit inline Quaternion(const Q& src)
        : _q0(src[0]), _q1(src[1]), _q2(src[2]), _q3(src[3]) {}
    
    template<class Q> inline Quaternion& operator=(const Q& src)
    {
        _q0=src[0]; _q1=src[1]; _q2=src[2]; _q3=src[3]; return (*this);
    }
    
    // set, reset
    inline Quaternion& operator=(const Quaternion& src);
    inline void operator*=(const Quaternion& src);
    inline void set(float q0, float q1, float q2, float q3);
    inline void set(const Vec3f& rfrom, const Vec3f& rto);
    inline void reset() { set(0.0f, 0.0f, 0.0f, 1.0f); }

    // indexed access
    inline float operator[](int i) const 
    {
        Q_ASSERT( i >= 0 && i < 4 );
        return (&_q0)[i];
    }
    inline float& operator[](int i) 
    {
        Q_ASSERT( i >= 0 && i < 4 );
        return (&_q0)[i];
    }
    
    // cast
    inline operator const float*() const { return &_q0;}
    inline operator float*() { return &_q0; }
    
    inline void invert() { _q3 = -_q3; }			
    inline Quaternion inverse() const{ return Quaternion(_q0,_q1,_q2, -_q3); }
    //
    inline void normalize(); 
    
    // Axis angle conversion
    void fromAxis(const Vec3f& v, float angle);
    void fromAxis(const Vec3f& v, float sina, float cosa); 
    
    inline Vec3f axis() const { return Vec3f(_q0, _q1, _q2); }
    float angle() const;
    
    // Euler Angle conversion
    // Note: euler angles expected in radians
    // Note: euler values returned are ensure in the ranges
    //    -  [0  .. 2PI] for rx
    //    -  [-PI .. PI] for ry
    //    -  [-PI .. PI] for rz
    Vec3f toEulerRadians() const;
    void fromEulerRadians(const Vec3f&);
    // Note: euler angles expected in degree
    // Note: euler values returned are ensure in the ranges
    //    -  [  0  .. 360] for rx
    //    -  [-180 .. 180] for ry
    //    -  [-180 .. 180] for rz
    Vec3f toEulerDegree() const;
    void fromEulerDegree(const Vec3f&);
    
    // Matrix conversion
    void fromMatrix(const Matrix4&);
    
    // vector xform
    template <class V>
    inline V rotate(const V& v) const;
    
    template <class V>
    inline V inverseRotate(const V& v) const;
    
};

// return a normalized Quaternion
inline Quaternion normalize(const Quaternion&);

inline Quaternion inverse(const Quaternion&);

inline float dot(const Quaternion&, const Quaternion&);

// Composition of 2 rotations ... non commutative op
inline Quaternion operator*(const Quaternion& a, const Quaternion& b);

// Scalar operation on quaternion
Quaternion operator^(const Quaternion& q, float s);

////////////////////////////////////////////////////////////////////////
//
// Inline Quaternion Implementation
//
////////////////////////////////////////////////////////////////////////

inline
Quaternion::Quaternion()
   : _q0(0.0f), _q1(0.0f), _q2(0.0f), _q3(1.0f)
{
}

inline
Quaternion::Quaternion(float q0, float q1, float q2, float q3)
    : _q0(q0), _q1(q1), _q2(q2), _q3(q3)
{
}

//  inline
//  Quaternion::Quaternion(const Quaternion& src)
//      : _q0(src._q0), _q1(src._q1), _q2(src._q2), _q3(src._q3)
//  {
//  }

inline Quaternion&
Quaternion::operator=(const Quaternion& src)
{
    _q0 = src._q0; _q1 = src._q1; _q2 = src._q2; _q3 = src._q3;
    return (*this);
}

inline void
Quaternion::operator*=(const Quaternion& src)
{
    (*this) = (*this) * src;
}

inline void
Quaternion::set(float q0, float q1, float q2, float q3)
{
    _q0 = q0; _q1 = q1; _q2 = q2; _q3 = q3;
}

inline Quaternion
normalize(const Quaternion& q)
{
    float  mag = q._q0*q._q0 + q._q1*q._q1 + q._q2*q._q2 + q._q3*q._q3;
    if (mag>0)
    {
	    float invn = 1.0f/sqrtf(mag);
	    return Quaternion(q._q0*invn, q._q1*invn, q._q2*invn, q._q3*invn);
    }
    return Quaternion();
}

inline
bool isnan(const Quaternion& v)
{
    return isnan(v._q0) || isnan(v._q1) || isnan(v._q2) || isnan(v._q3);
}


inline void
Quaternion::normalize()
{
    *this = ::normalize(*this);
}

inline Quaternion
operator*(const Quaternion& a, const Quaternion& b)
{
    return Quaternion
        (a._q3*b._q0 + a._q0*b._q3 + a._q1*b._q2 - a._q2*b._q1,
         a._q3*b._q1 + a._q1*b._q3 + a._q2*b._q0 - a._q0*b._q2,
         a._q3*b._q2 + a._q2*b._q3 + a._q0*b._q1 - a._q1*b._q0,
         a._q3*b._q3 - a._q0*b._q0 - a._q1*b._q1 - a._q2*b._q2 );
}

//! Log of a quaternion : returns the rotation vector.
inline Vec3f
log(const Quaternion& q)
{
    Vec3f u(q.axis());
    float sin2a = length2(u);   // sin(a/2)
    if (sin2a > 0)
    {
        sin2a = sqrtf(sin2a);
        return u * (q.angle()/sin2a);
    } else
        return Vec3f(0,0,0);
}

//! Exp of a rotation vector : returns the quaternion.
inline Quaternion
exp(const Vec3f& v)
{
    float a = length2(v);
    if( a > 0.0f )
    {
        a = sqrtf(a);
        return Quaternion(v/a, a); // Normalized axis, angle.
    }
    else
        return Quaternion(0,0,0, 1); // Null angle rotation.
}


//! Power of a quaternion : multiply its rotation angle by a given float.
inline Quaternion 
power(const Quaternion& q, float s)
{
    float a = q.angle(); 
    if (a > 0)
    {
        return Quaternion(q.axis(), a*s);
    } else
        return Quaternion(0,0,0, 1);    // Null rotation.
}

//! Spherical interpolation between two quaternions.
inline Quaternion
slerp(const Quaternion& qa, const Quaternion& qb, float u)
{
    return power(qa, 1.0f-u) * power(qb, u);
}

// vector xform
template <class V>
inline V
Quaternion::rotate(const V& v) const
{
    return V
        (( _q0*_q0 - _q1*_q1 - _q2*_q2 + _q3*_q3)*v[0] +
         (              2.0 * (_q0*_q1 - _q2*_q3))*v[1] +
         (              2.0 * (_q2*_q0 + _q1*_q3))*v[2],
	   
         (              2.0 * (_q0*_q1 + _q2*_q3))*v[0] +
         (-_q0*_q0 + _q1*_q1 - _q2*_q2 + _q3*_q3)*v[1] +
         (              2.0 * (_q1*_q2 - _q0*_q3))*v[2],
	   
         (              2.0 * (_q2*_q0 - _q1*_q3))*v[0] +
         (              2.0 * (_q1*_q2 + _q0*_q3))*v[1] +
         (-_q0*_q0 - _q1*_q1 + _q2*_q2 + _q3*_q3)*v[2]  );
}

template <class V>
inline V
Quaternion::inverseRotate(const V& v) const
{
    /// simply replace _q3 with -q3 in the above expr
    return V
        (( _q0*_q0 - _q1*_q1 - _q2*_q2 + _q3*_q3)*v[0] +
         (              2.0 * (_q0*_q1 + _q2*_q3))*v[1] +
         (              2.0 * (_q2*_q0 - _q1*_q3))*v[2],
	   
         (              2.0 * (_q0*_q1 - _q2*_q3))*v[0] +
         (-_q0*_q0 + _q1*_q1 - _q2*_q2 + _q3*_q3)*v[1] +
         (              2.0 * (_q1*_q2 + _q0*_q3))*v[2],
	   
         (              2.0 * (_q2*_q0 + _q1*_q3))*v[0] +
         (              2.0 * (_q1*_q2 - _q0*_q3))*v[1] +
         (-_q0*_q0 - _q1*_q1 + _q2*_q2 + _q3*_q3)*v[2]  );
//          ((1.0 - 2.0 * (_q1 * _q1 + _q2 * _q2))*v[0] +
//           (      2.0 * (_q0 * _q1 + _q2 * _q3))*v[1] +
//           (      2.0 * (_q2 * _q0 - _q1 * _q3))*v[2],
	   
//           (      2.0 * (_q0 * _q1 - _q2 * _q3))*v[0] +
//           (1.0 - 2.0 * (_q2 * _q2 + _q0 * _q0))*v[1] +
//           (      2.0 * (_q1 * _q2 + _q0 * _q3))*v[2],
	   
//           (      2.0 * (_q2 * _q0 + _q1 * _q3))*v[0] +
//           (      2.0 * (_q1 * _q2 - _q0 * _q3))*v[1] +
//           (1.0 - 2.0 * (_q1 * _q1 + _q0 * _q0))*v[2]  );
}

inline Quaternion
inverse(const Quaternion& q)
{
    return Quaternion( q._q0, q._q1, q._q2, -q._q3);
}

inline float
dot(const Quaternion& a, const Quaternion& b)
{
    return (a._q0*b._q0 + a._q1*b._q1 + a._q2*b._q2 + a._q3*b._q3);
}

#endif // QUATERNION_H
