////////////////////////////////////////////////////////////////////////
/*! \file   Vec2f.h
 *  \brief  A 2-float vector class
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Vec2f.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Vec2f.h,v 1.5 2008/02/20 11:49:46 gascuel Exp $
// File History:
// [jDG 2004/03/30] : Cut&Paste from Vec2f.

#ifndef VEC2F_H
#define VEC2F_H

#include "Math.h"

/*! \brief another 2D vector class
    \ingroup Math

	Vec2f implements 2D vector (floats)

	Sample usage:
    \code
    Vec2f a(12,12,);
    Vec2f b(3,9);

		float dot_product = dot(a,b); // computes the dot product of a and b
    \endcode
		... guess what the other are doing. Should we continue with writing twice
		the same thing in code and comments ... 

 */
class Vec2f
{
public:
    float _x,_y;
    	
    inline Vec2f();
    inline Vec2f(float x, float y);

    // Component operations
    inline Vec2f& operator=(const Vec2f&);
    inline void operator+=(const Vec2f&);
    inline void operator-=(const Vec2f&);

    // Scalar operations
    inline void operator+=(float);
    inline void operator-=(float);
    inline void operator*=(float);
    inline void operator/=(float);
    	
    inline void set(float x=0, float y=0);
    inline void reset() { set(); }

    inline float operator[](int i) const;
    inline float& operator[](int i);
    	
    inline void normalize();

    // Support for converting from any types that support operator[]
    // on at least 3 elements:
    template<class V>
    explicit inline Vec2f(const V& b)
	    : _x(b[0]), _y(b[1])
    {}


    // cast
    inline operator const float*() const { return &_x;}
    inline operator float*() { return &_x; }
			
    template<class V>
    inline Vec2f& operator=(const V& b)
    {
  		    _x = b[0];
  		    _y = b[1];
  		    return *this;
    }
};

// Mix Scalar operations
/// return of (const T) are also 'bad' => (T) instead
inline Vec2f operator*(float, const Vec2f&);
inline Vec2f operator*(const Vec2f&, float);
inline Vec2f operator/(const Vec2f&, float);

// Component operations
inline Vec2f operator+(const Vec2f&, const Vec2f&);
inline Vec2f operator-(const Vec2f&, const Vec2f&);
inline Vec2f operator*(const Vec2f&, const Vec2f&);
inline Vec2f operator/(const Vec2f&, const Vec2f&);

// equality
inline bool operator==(const Vec2f&, const Vec2f&);
inline bool operator!=(const Vec2f&, const Vec2f&);

// unary ops
inline Vec2f operator-(const Vec2f&);

// vector operation
inline float dot(const Vec2f&, const Vec2f&);
inline float cross(const Vec2f& a, const Vec2f& b);

// lengths computation
inline float length2(const Vec2f&);
inline float length(const Vec2f&);

// const unit length conversion
inline Vec2f normalize(const Vec2f&);

class QDataStream;
class QTextStream;

// Files operation
extern  QDataStream & operator<<(QDataStream&, const Vec2f&);
extern  QTextStream & operator<<(QTextStream&, const Vec2f&);
extern  QDataStream & operator>>(QDataStream&, Vec2f&);
extern  QTextStream & operator>>(QTextStream&, Vec2f&);



////////////////////////////////////////////////////////////////////////
//
// Inline Vec2f Implementation
//
////////////////////////////////////////////////////////////////////////

inline
Vec2f::Vec2f()
   : _x(0), _y(0)
{
}

inline 
Vec2f::Vec2f(float x, float y)
   : _x(x), _y(y)
{
}

inline Vec2f&
Vec2f::operator=(const Vec2f& src)
{
   _x = src._x;
   _y = src._y;
   return *this;
}

inline void 
Vec2f::operator+=(const Vec2f& src)
{
   _x += src._x;
   _y += src._y;
}

inline void 
Vec2f::operator-=(const Vec2f& src)
{
   _x -= src._x;
   _y -= src._y;
}

inline void 
Vec2f::operator+=(float s)
{
   _x += s;
   _y += s;
}
inline void 
Vec2f::operator-=(float s)
{
   _x -= s;
   _y -= s;
}
inline void 
Vec2f::operator*=(float s)
{
   _x *= s;
   _y *= s;
}
inline void 
Vec2f::operator/=(float s)
{
   _x /= s;
   _y /= s;
}
			
inline void 
Vec2f::set(float x, float y)
{
   _x = x; _y = y;
}

inline float 
Vec2f::operator[](int i) const
{
    Q_ASSERT( i >= 0 && i < 2 );
    return (&_x)[i];
}

inline float& 
Vec2f::operator[](int i)
{
    Q_ASSERT( i >= 0 && i < 2 );
    return (&_x)[i];
}

// equality
inline bool
operator==(const Vec2f& a, const Vec2f& b)
{
   return ((a._x==b._x) && (a._y==b._y));
}

inline bool
operator!=(const Vec2f& a, const Vec2f& b)
{
   return ((a._x!=b._x) || (a._y!=b._y));
}

// unary ops
inline Vec2f
operator-(const Vec2f& v)
{
   return Vec2f( -v._x, -v._y );
}

// vector operation

inline float
dot(const Vec2f& a, const Vec2f& b)
{
   return a._x*b._x + a._y*b._y ;
}

inline float
cross(const Vec2f& a, const Vec2f& b)
{
    return a._x*b._y - a._y*b._x;
}

// lengths computation
inline float
length2(const Vec2f& v)
{
   return dot(v,v);
}

inline float
length(const Vec2f& v)
{
   return sqrtf(length2(v));
}

inline void 
Vec2f::normalize()
{
   float len = length2(*this);
   if (len>0)
   {
	  len = sqrtf(len);
	  operator/=(len);
   }
}

// Mix Scalar operations
inline Vec2f
operator*(float a, const Vec2f& b)
{
   return Vec2f( a*b._x, a*b._y);
}

inline Vec2f
operator*(const Vec2f& a, float b)
{
   return Vec2f( a._x*b, a._y*b);
}

inline Vec2f
operator/(const Vec2f& a, float b)
{
   return Vec2f( a._x/b, a._y/b );
}

// const unit length conversion
inline Vec2f
normalize(const Vec2f& a)
{
   float len = length2(a);
   if (len>0)
   {
	  return a/sqrtf(len); // uses mix scalar ops ... 
   }
   return a;
}



// Component operations
inline Vec2f
operator+(const Vec2f& a, const Vec2f& b)
{
   return Vec2f( a._x+b._x, a._y+b._y );
}

inline Vec2f
operator-(const Vec2f& a, const Vec2f& b)
{
   return Vec2f( a._x-b._x, a._y-b._y );
}

inline Vec2f
operator*(const Vec2f& a, const Vec2f& b)
{
   return Vec2f( a._x*b._x, a._y*b._y );
}

inline Vec2f
operator/(const Vec2f& a, const Vec2f& b)
{
   return Vec2f( a._x/b._x, a._y/b._y );
}


						
#endif // Vec2f_H
