////////////////////////////////////////////////////////////////////////
/*! 
 *  \file   Vec3f.h
 *  \brief  A 3-float vector class
 *  \author Jean-Dominique Gascuel, Eric Ferley
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Vec4f, Vec3i, Vec2f.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Vec3f.h,v 1.6 2008/02/20 11:49:46 gascuel Exp $
// File History:

#ifndef VEC3F_H
#define VEC3F_H

#include "Math.h"

/*! \brief another 3D vector class
	\ingroup Math

	Vec3f implements 3D vector (floats)

	Sample usage:
    \code
    Vec3f a(12,12,12);
    Vec3f b(3,9,27);

    Vec3f cross_product = cross(a,b); // computes the cross product of a and b
	float dot_product = dot(a,b); // computes the dot product of a and b
    \endcode
		... guess what the other are doing. Should we continue with writing twice
		the same thing in code and comments ... 

 */
class Vec3f
{
public:
    float _x,_y,_z;
		
    inline Vec3f();
    inline Vec3f(float x, float y, float z);

    // Support for converting from any types that support operator[]
    // on at least 3 elements:
    template<class V>
    explicit inline Vec3f(const V& b)
        : _x(b[0]), _y(b[1]), _z(b[2]) 
    {}

    // Component operations
    inline Vec3f& operator=(const Vec3f&);
    inline void operator+=(const Vec3f&);
    inline void operator-=(const Vec3f&);
    inline void operator*=(const Vec3f&);
    inline void operator/=(const Vec3f&);

    // Scalar operations
    inline void operator+=(float);
    inline void operator-=(float);
    inline void operator*=(float);
    inline void operator/=(float);

    inline void set(float x=0, float y=0, float z=0);
    inline void reset() { set(); }

    inline float operator[](int i) const;
    inline float& operator[](int i);
		
    inline void normalize();

    // cast
    inline operator const float*() const { return &_x;}
    inline operator float*() { return &_x; }
		
    // Support for converting from any types that support operator[]
    // on at least 3 elements:
    template<class V>
    inline Vec3f& operator=(const V& b)
    {
        _x = b[0];
        _y = b[1];
        _z = b[2];
        return *this;
    }
};

// Mix Scalar operations
/// return of (const T) are also 'bad' => (T) instead
inline Vec3f operator+(const Vec3f&, float);
inline Vec3f operator+(float, const Vec3f&);
inline Vec3f operator-(const Vec3f&, float);
inline Vec3f operator-(float, const Vec3f&);
inline Vec3f operator*(float, const Vec3f&);
inline Vec3f operator*(const Vec3f&, float);
inline Vec3f operator/(const Vec3f&, float);

// Component operations
inline Vec3f operator+(const Vec3f&, const Vec3f&);
inline Vec3f operator-(const Vec3f&, const Vec3f&);
inline Vec3f operator*(const Vec3f&, const Vec3f&);
inline Vec3f operator/(const Vec3f&, const Vec3f&);

// equality
inline bool operator==(const Vec3f&, const Vec3f&);
inline bool operator!=(const Vec3f&, const Vec3f&);

// unary ops
inline Vec3f operator-(const Vec3f&);

// vector operation
inline Vec3f cross(const Vec3f&, const Vec3f&);
inline float dot(const Vec3f&, const Vec3f&);


// lengths computation
inline float length2(const Vec3f&);
inline float length(const Vec3f&);

// const unit length conversion
inline Vec3f normalize(const Vec3f&);

// Misc functions
//! Which component have the largest magnitude ?
inline int dominant(const Vec3f& v);

////////////////////////////////////////////////////////////////////////
//
// Inline Vec3f Implementation
//
////////////////////////////////////////////////////////////////////////

inline
Vec3f::Vec3f()
    : _x(0), _y(0), _z(0)
{
}

inline 
Vec3f::Vec3f(float x, float y, float z)
    : _x(x), _y(y), _z(z)
{
}

inline Vec3f&
Vec3f::operator=(const Vec3f& src)
{
    _x = src._x;
    _y = src._y;
    _z = src._z;
    return *this;
}

inline void 
Vec3f::operator+=(const Vec3f& src)
{
    _x += src._x;
    _y += src._y;
    _z += src._z;
}

inline void 
Vec3f::operator-=(const Vec3f& src)
{
    _x -= src._x;
    _y -= src._y;
    _z -= src._z;
}

inline void 
Vec3f::operator*=(const Vec3f& src)
{
    _x *= src._x;
    _y *= src._y;
    _z *= src._z;
}

inline void 
Vec3f::operator/=(const Vec3f& src)
{
    _x /= src._x;
    _y /= src._y;
    _z /= src._z;
}

inline void 
Vec3f::operator+=(float s)
{
    _x += s;
    _y += s;
    _z += s;
}
inline void 
Vec3f::operator-=(float s)
{
    _x -= s;
    _y -= s;
    _z -= s;
}
inline void 
Vec3f::operator*=(float s)
{
    _x *= s;
    _y *= s;
    _z *= s;
}
inline void 
Vec3f::operator/=(float s)
{
    _x /= s;
    _y /= s;
    _z /= s;
}
			
inline Vec3f
linear(const Vec3f&a, const Vec3f& b, float u)
{
    return (1.0f-u)*a + u*b;
}

inline Vec3f
bilinear(const Vec3f&a, const Vec3f& b, const Vec3f&c, const Vec3f&d, float u, float v)
{
    Vec3f ab = linear(a, b, u);
    Vec3f cd = linear(c, d, u);
    return linear(ab, cd, v);
}

inline void 
Vec3f::set(float x, float y, float z)
{
    _x = x; _y = y; _z = z;
}

inline float 
Vec3f::operator[](int i) const
{
    Q_ASSERT( i >= 0 && i < 3 );
    return (&_x)[i];
}

inline float& 
Vec3f::operator[](int i)
{
    Q_ASSERT( i >= 0 && i < 3 );
    return (&_x)[i];
}

// equality
inline bool
operator==(const Vec3f& a, const Vec3f& b)
{
    return ((a._x==b._x) && (a._y==b._y) && (a._z==b._z));
}
inline bool
operator!=(const Vec3f& a, const Vec3f& b)
{
    return ((a._x!=b._x) || (a._y!=b._y) || (a._z!=b._z));
}

// unary ops
inline Vec3f
operator-(const Vec3f& v)
{
    return Vec3f( -v._x, -v._y, -v._z );
}

// vector operation
inline Vec3f
cross(const Vec3f& a, const Vec3f& b)
{
    return Vec3f( a._y*b._z - a._z*b._y,
                  a._z*b._x - a._x*b._z,
                  a._x*b._y - a._y*b._x);
}

inline float
dot(const Vec3f& a, const Vec3f& b)
{
    return a._x*b._x + a._y*b._y + a._z*b._z;
}

// lengths computation
inline float
length2(const Vec3f& v)
{
    return dot(v,v);
}

inline float
length(const Vec3f& v)
{
    return sqrtf(length2(v));
}

inline void 
Vec3f::normalize()
{
    float len = length2(*this);
    if ( len > 1.0e-30 )
    {
        len = sqrtf(len);
        operator/=(len);
    }
    else
        reset();
}

inline
bool isnan(const Vec3f& v)
{
    return isnan(v._x) || isnan(v._y) || isnan(v._z);
}

// Mix Scalar operations
inline Vec3f
operator+(float a, const Vec3f& b)
{
    return Vec3f(a + b._x, a + b._y, a + b._z);
}

inline Vec3f
operator+(const Vec3f& b, float a)
{
    return a+b;
}

inline Vec3f
operator-(float a, const Vec3f& b)
{
    return Vec3f(a - b._x, a - b._y, a - b._z);
}

inline Vec3f
operator-(const Vec3f& b, float a)
{
    return Vec3f(b._x - a, b._y - a, b._z - a);
}

inline Vec3f
operator*(float a, const Vec3f& b)
{
    return Vec3f( a*b._x, a*b._y, a*b._z );
}

inline Vec3f
operator*(const Vec3f& a, float b)
{
    return b * a;
}

inline Vec3f
operator/(const Vec3f& a, float b)
{
    return Vec3f( a._x/b, a._y/b, a._z/b );
}

// const unit length conversion
inline Vec3f
normalize(const Vec3f& a)
{
    float len = length2(a);
    if (len>0)
    {
        return a/sqrtf(len); // uses mix scalar ops ... 
    }
    return a;
}



// Component operations
inline Vec3f
operator+(const Vec3f& a, const Vec3f& b)
{
    return Vec3f( a._x+b._x, a._y+b._y, a._z+b._z );
}

inline Vec3f
operator-(const Vec3f& a, const Vec3f& b)
{
    return Vec3f( a._x-b._x, a._y-b._y, a._z-b._z );
}

inline Vec3f
operator*(const Vec3f& a, const Vec3f& b)
{
    return Vec3f( a._x*b._x, a._y*b._y, a._z*b._z );
}

inline Vec3f
operator/(const Vec3f& a, const Vec3f& b)
{
    return Vec3f( a._x/b._x, a._y/b._y, a._z/b._z );
}

//! Which component have the largest magnitude ?
inline int dominant(const Vec3f& v)
{
    if( fabsf(v._x) >= fabsf(v._y) && fabsf(v._x) >= fabsf(v._z) ) return 0;
    if( fabsf(v._y) >= fabsf(v._x) && fabsf(v._y) >= fabsf(v._z) ) return 1;
    Q_ASSERT( fabsf(v._z) >= fabsf(v._x) && fabsf(v._z) >= fabsf(v._y) );

    return 2;
}

//! Which component have the smallest magnitude ?
// TODO : find a better name !
inline int domined(const Vec3f& v)
{
    if( fabsf(v._x) <= fabsf(v._y) && fabsf(v._x) <= fabsf(v._z) ) return 0;
    if( fabsf(v._y) <= fabsf(v._x) && fabsf(v._y) <= fabsf(v._z) ) return 1;
    Q_ASSERT( fabsf(v._z) <= fabsf(v._x) && fabsf(v._z) <= fabsf(v._y) );

    return 2;
}

#endif // VEC3F_H

