////////////////////////////////////////////////////////////////////////
/*! 
 *  \file   Vec4f.h
 *  \brief  A 4-float vector class
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Vec4f, Vec2f.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Vec4f.h,v 1.6 2008/02/20 11:49:46 gascuel Exp $
// File History:
// [jDG 2006-11-29] Creation.

#ifndef VEC4F_H
#define VEC4F_H

#include "Math.h"

/*! \brief another 4D vector class
	\ingroup Math

	Vec4f implements 4D vector (floats)

	Sample usage:
    \code
    Vec4f a(12,12,12,12);
    Vec4f b(3,9,27,81);

	float dot_product = dot(a,b); // computes the dot product of a and b
    \endcode
		... guess what the other are doing. Should we continue with writing twice
		the same thing in code and comments ... 

 */
class Vec4f
{
public:
    float _x,_y,_z,_w;
		
    inline Vec4f();
    inline Vec4f(float x, float y, float z, float w);

    // Support for converting from any types that support operator[]
    // on at least 3 elements:
    template<class V>
    explicit inline Vec4f(const V& b)
        : _x(b[0]), _y(b[1]), _z(b[2]), _w(b[3])
    {}

    // Component operations
    inline Vec4f& operator=(const Vec4f&);
    inline void operator+=(const Vec4f&);
    inline void operator-=(const Vec4f&);
    inline void operator*=(const Vec4f&);
    inline void operator/=(const Vec4f&);

    // Scalar operations
    inline void operator+=(float);
    inline void operator-=(float);
    inline void operator*=(float);
    inline void operator/=(float);

    inline void set(float x=0, float y=0, float z=0, float w=0);
    inline void reset() { set(); }

    inline float operator[](int i) const;
    inline float& operator[](int i);
		
    inline void normalize();

    // cast
    inline operator const float*() const { return &_x;}
    inline operator float*() { return &_x; }
		
    // Support for converting from any types that support operator[]
    // on at least 3 elements:
    template<class V>
    inline Vec4f& operator=(const V& b)
    {
        _x = b[0];
        _y = b[1];
        _z = b[2];
        _w = b[3];
        return *this;
    }
};

// Mix Scalar operations
/// return of (const T) are also 'bad' => (T) instead
inline Vec4f operator+(const Vec4f&, float);
inline Vec4f operator+(float, const Vec4f&);
inline Vec4f operator-(const Vec4f&, float);
inline Vec4f operator-(float, const Vec4f&);
inline Vec4f operator*(float, const Vec4f&);
inline Vec4f operator*(const Vec4f&, float);
inline Vec4f operator/(const Vec4f&, float);

// Component operations
inline Vec4f operator+(const Vec4f&, const Vec4f&);
inline Vec4f operator-(const Vec4f&, const Vec4f&);
inline Vec4f operator*(const Vec4f&, const Vec4f&);
inline Vec4f operator/(const Vec4f&, const Vec4f&);

// equality
inline bool operator==(const Vec4f&, const Vec4f&);
inline bool operator!=(const Vec4f&, const Vec4f&);

// unary ops
inline Vec4f operator-(const Vec4f&);

// vector operation
inline float dot(const Vec4f&, const Vec4f&);


// lengths computation
inline float length2(const Vec4f&);
inline float length(const Vec4f&);

// const unit length conversion
inline Vec4f normalize(const Vec4f&);

// Misc functions
//! Which component have the largest magnitude ?
inline int dominant(const Vec4f& v);

////////////////////////////////////////////////////////////////////////
//
// Inline Vec4f Implementation
//
////////////////////////////////////////////////////////////////////////

inline
Vec4f::Vec4f()
    : _x(0), _y(0), _z(0), _w(0)
{
}

inline 
Vec4f::Vec4f(float x, float y, float z, float w)
    : _x(x), _y(y), _z(z), _w(w)
{
}

inline Vec4f&
Vec4f::operator=(const Vec4f& src)
{
    _x = src._x;
    _y = src._y;
    _z = src._z;
    _w = src._w;
    return *this;
}

inline void 
Vec4f::operator+=(const Vec4f& src)
{
    _x += src._x;
    _y += src._y;
    _z += src._z;
    _w += src._w;
}

inline void 
Vec4f::operator-=(const Vec4f& src)
{
    _x -= src._x;
    _y -= src._y;
    _z -= src._z;
    _w -= src._w;
}

inline void 
Vec4f::operator*=(const Vec4f& src)
{
    _x *= src._x;
    _y *= src._y;
    _z *= src._z;
    _w *= src._w;
}

inline void 
Vec4f::operator/=(const Vec4f& src)
{
    _x /= src._x;
    _y /= src._y;
    _z /= src._z;
    _w /= src._w;
}

inline void 
Vec4f::operator+=(float s)
{
    _x += s;
    _y += s;
    _z += s;
    _w += s;
}
inline void 
Vec4f::operator-=(float s)
{
    _x -= s;
    _y -= s;
    _z -= s;
    _w -= s;
}
inline void 
Vec4f::operator*=(float s)
{
    _x *= s;
    _y *= s;
    _z *= s;
    _w *= s;
}
inline void 
Vec4f::operator/=(float s)
{
    _x /= s;
    _y /= s;
    _z /= s;
    _w /= s;
}
			
inline Vec4f
linear(const Vec4f&a, const Vec4f& b, float u)
{
    return (1.0f-u)*a + u*b;
}

inline Vec4f
bilinear(const Vec4f&a, const Vec4f& b, const Vec4f&c, const Vec4f&d, float u, float v)
{
    Vec4f ab = linear(a, b, u);
    Vec4f cd = linear(c, d, u);
    return linear(ab, cd, v);
}

inline void 
Vec4f::set(float x, float y, float z, float w)
{
    _x = x; _y = y; _z = z; _w = w;
}

inline float 
Vec4f::operator[](int i) const
{
    Q_ASSERT( i >= 0 && i < 4 );
    return (&_x)[i];
}

inline float& 
Vec4f::operator[](int i)
{
    Q_ASSERT( i >= 0 && i < 4 );
    return (&_x)[i];
}

// equality
inline bool
operator==(const Vec4f& a, const Vec4f& b)
{
    return (a._x==b._x) && (a._y==b._y) && (a._z==b._z) && (a._w == b._w);
}
inline bool
operator!=(const Vec4f& a, const Vec4f& b)
{
    return (a._x!=b._x) || (a._y!=b._y) || (a._z!=b._z) || (a._w != b._w);
}

// unary ops
inline Vec4f
operator-(const Vec4f& v)
{
    return Vec4f( -v._x, -v._y, -v._z, -v._w );
}

inline float
dot(const Vec4f& a, const Vec4f& b)
{
    return a._x*b._x + a._y*b._y + a._z*b._z + a._w*b._w;
}

// lengths computation
inline float
length2(const Vec4f& v)
{
    return dot(v,v);
}

inline float
length(const Vec4f& v)
{
    return sqrtf(length2(v));
}

inline void 
Vec4f::normalize()
{
    float len = length2(*this);
    if ( len > 1.0e-30 )
    {
        len = sqrtf(len);
        operator/=(len);
    }
    else
        reset();
}

inline bool 
isnan(const Vec4f& v)
{
    return isnan(v._x) || isnan(v._y) || isnan(v._z) || isnan(v._w);
}

// Mix Scalar operations
inline Vec4f
operator+(float a, const Vec4f& b)
{
    return Vec4f(a + b._x, a + b._y, a + b._z, a + b._w);
}

inline Vec4f
operator+(const Vec4f& b, float a)
{
    return a+b;
}

inline Vec4f
operator-(float a, const Vec4f& b)
{
    return Vec4f(a - b._x, a - b._y, a - b._z, a - b._w);
}

inline Vec4f
operator-(const Vec4f& b, float a)
{
    return Vec4f(b._x - a, b._y - a, b._z - a, b._w - a);
}

inline Vec4f
operator*(float a, const Vec4f& b)
{
    return Vec4f( a*b._x, a*b._y, a*b._z, a*b._w );
}

inline Vec4f
operator*(const Vec4f& a, float b)
{
    return b * a;
}

inline Vec4f
operator/(const Vec4f& a, float b)
{
    return Vec4f( a._x/b, a._y/b, a._z/b, a._w/b );
}

// const unit length conversion
inline Vec4f
normalize(const Vec4f& a)
{
    float len = length2(a);
    if (len>0)
    {
        return a/sqrtf(len); // uses mix scalar ops ... 
    }
    return a;
}



// Component operations
inline Vec4f
operator+(const Vec4f& a, const Vec4f& b)
{
    return Vec4f( a._x+b._x, a._y+b._y, a._z+b._z, a._w+b._w );
}

inline Vec4f
operator-(const Vec4f& a, const Vec4f& b)
{
    return Vec4f( a._x-b._x, a._y-b._y, a._z-b._z, a._w-b._w );
}

inline Vec4f
operator*(const Vec4f& a, const Vec4f& b)
{
    return Vec4f( a._x*b._x, a._y*b._y, a._z*b._z, a._w*b._w );
}

inline Vec4f
operator/(const Vec4f& a, const Vec4f& b)
{
    return Vec4f( a._x/b._x, a._y/b._y, a._z/b._z, a._w/b._w );
}

//! Which component have the largest magnitude ?
inline int dominant(const Vec4f& v)
{
    if( fabsf(v._x) >= fabsf(v._y) && fabsf(v._x) >= fabsf(v._z) && fabsf(v._x) >= fabsf(v._w)) return 0;
    if( fabsf(v._y) >= fabsf(v._x) && fabsf(v._y) >= fabsf(v._z) && fabsf(v._y) >= fabsf(v._w)) return 1;
    if( fabsf(v._z) >= fabsf(v._x) && fabsf(v._z) >= fabsf(v._y) && fabsf(v._z) >= fabsf(v._w)) return 2;
    Q_ASSERT( fabsf(v._w) >= fabsf(v._x) && fabsf(v._w) >= fabsf(v._y) && fabsf(v._w) >= fabsf(v._z));

    return 3;
}

//! Which component have the smallest magnitude ?
// TODO : find a better name !
inline int domined(const Vec4f& v)
{
    if( fabsf(v._x) <= fabsf(v._y) && fabsf(v._x) <= fabsf(v._z) && fabsf(v._x) <= fabsf(v._w)) return 0;
    if( fabsf(v._y) <= fabsf(v._x) && fabsf(v._y) <= fabsf(v._z) && fabsf(v._y) <= fabsf(v._w)) return 1;
    if( fabsf(v._z) <= fabsf(v._x) && fabsf(v._z) <= fabsf(v._y) && fabsf(v._z) <= fabsf(v._w)) return 2;
    Q_ASSERT( fabsf(v._w) <= fabsf(v._x) && fabsf(v._w) <= fabsf(v._y) && fabsf(v._w) <= fabsf(v._z));

    return 3;
}

#endif // VEC4F_H

