//////////////////////////////////////////////////////////////////////////////
/*! \file   OcclusionGrid.cpp
 *  \brief  Grid storing ambiant occlusion around an object.
 *  \author J.D. Gascuel
 *
 * \sa Scene, Environment, Shader.
 */
//////////////////////////////////////////////////////////////////////////////
// $Id: OcclusionGrid.cpp,v 1.4 2007/10/22 00:17:28 gascuel Exp $
// File History:
// [jDG 2007/06/14] : Creation from OcclusionGrid.

#include "OcclusionGrid.h"

#include <GL/glew.h>

#include <QVector>

#include "FrameBufferObject.h"
#include "Math/Matrix4.h"
#include "Utilities.h"
#include "Shader.h"
#include "Scene.h"

static bool unsupported = false;

#define QUERIES_NUMBER  64

//////////////////////////////////////////////////////////////////////////////

typedef QVector<OcclusionGrid*> OcclusionGrids;
static OcclusionGrids allOcclusionGrids;

Shader* OcclusionGrid:: phongShader = 0;
Shader* OcclusionGrid::sphereShader = 0;

//////////////////////////////////////////////////////////////////////////////

OcclusionGrid::OcclusionGrid()
: _nu(4),
  _nv(4),
  _nw(4),
  _data(0),
  _origin(0,0,0),
  _u(0.1,0,0),
  _v(0,0.1,0),
  _w(0,0,0.1),
  _size(128)
{
    allOcclusionGrids.push_back(this);
}

OcclusionGrid::~OcclusionGrid()
{
    int idx = allOcclusionGrids.indexOf(this);
    if( idx >= 0 )
        allOcclusionGrids.remove(idx);

    clean();
}

//////////////////////////////////////////////////////////////////////////////

void
OcclusionGrid::clean()
{
    if( _data )
    {
        delete[] _data;
        _data = 0;
    }

    if( _queries )
    {
        glDeleteQueries(QUERIES_NUMBER, _queries);
        delete[] _queries;
        _queries = 0;
    }
}


//////////////////////////////////////////////////////////////////////////////

bool
OcclusionGrid::reloadShaders(const Environment& env)
{
    if( ! GLEW_EXT_geometry_shader4
     || ! GLEW_EXT_framebuffer_object )
    {
        if( !unsupported )
            qWarning("**** Cannot do glossy map without missing \n"
                     "     GL_EXT_geometry_shader4 or \n"
                     "     GL_EXT_framebuffer_object\n"
                     "OpenGL extensions");
        unsupported = true;
        return false;
    }

    //------------------------------------------------------------------------

    QString envString = "#define SUBMODE_COVERAGE 1\n";
    envString += env.envString();

    //------------------------------------------------------------------------
    // Use a full featured phong shader to :
    //  - Apply glossy map to objects w/o special shaders...
    if( phongShader ) delete phongShader;
    phongShader = new GLShader(
        "phong_vertex.glsl",
        GL_TRIANGLES, "phong_geometry.glsl", GL_TRIANGLE_STRIP,
        "phong_fragment.glsl",
        envString.toLatin1().constData());

    if( ! phongShader->compile() )
    {
        delete phongShader;
        phongShader = 0;
        return false;
    }

    //------------------------------------------------------------------------
    // Use a most basic CosinusMap shader to :
    //  - Render scene to measure the occlusion coefficients.
    if( sphereShader ) delete sphereShader;
    sphereShader = new GLShader(
        "common_vertex.glsl",
        GL_TRIANGLES, "sphereBox_geometry.glsl", GL_TRIANGLE_STRIP,
        "sphere_fragment.glsl",
        envString.toLatin1().constData());

    if( ! sphereShader->compile() )
    {
        delete sphereShader;
        sphereShader = 0;
        return false;
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

bool 
OcclusionGrid::render(const Scene* scene, float zFar)
{
    if( unsupported ) return false;
    if( ! phongShader )
        if( ! reloadShaders( scene->env ) )
            return false;                   // Bad shader code.... retry later.

    //------------------------------------------------------------------------
    // Creates Querries.
    if( ! _queries )
    {
        _queries = new GLuint[QUERIES_NUMBER];
        Q_CHECK_PTR(_queries);
        glGenQueries(QUERIES_NUMBER, _queries);
        CHECK_GL_ERROR("alloc Queries");
    }

    //------------------------------------------------------------------------
    // Creates the FBO, render buffer (both color and Z) :
    FrameBufferObject fbo;
    if( ! fbo.bindColorImage(_size, _size, true) )
        return false;

    // ---- setup ------------------------------------------------------------

    glPushAttrib(GL_VIEWPORT_BIT);
    glViewport(0, 0, _size, _size);   // Origin & w*h

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    CHECK_GL_ERROR("push matrices");

    // ---- For each layer of the array --------------------------------------
    bool ok = true;
    for(int u=0; ok && u<_nu; ++u)
        for(int v=0; ok && v<_nv; ++v)
            for(int w=0; ok && v<_nw; ++w)
            {
                int idx = w + _nw * (v + u * _nv);

                Vec3f eye = _origin
                          + _u * u
                          + _v * v
                          + _w * w
                          ;

*** HERE ! *******************************************************************

                // Viewing direction for that layer :
	            glMatrixMode(GL_MODELVIEW);
	            glLoadIdentity();
                gluLookAt(eye._x, eye._y, eye._z,
                          n._x + eye._x, n._y + eye._y, n._z + eye._z,
                          _v._x,   _v._y,   _v._z);
	            CHECK_GL_ERROR("glossy LookAt");

                // Make FBO to directly write results there :
                glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
                glFramebufferTextureLayerEXT(GL_FRAMEBUFFER_EXT,    // target
                                          GL_COLOR_ATTACHMENT0_EXT, // attachment
                                          _texId, 0,                // dest texture & mipmap level
                                          idx);                     // dest layer (aka zoffset)
                CHECK_GL_ERROR("attach glossy map layer");

                // Use a single rendering target for that RGBA image :
                glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
                glReadBuffer(GL_NONE);
                CHECK_GL_ERROR("draw to FBO");

                // Check everything went well :
                uint complete = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
                if( complete != GL_FRAMEBUFFER_COMPLETE_EXT )
                {
                    qWarning("*** Uncomplete frame buffer. Code 0x%x", complete);
                    ok = false;
                    break;
                }

                // ---- Then Go --------------------------------------------------
                glClearColor( !(idx&1), !(idx&2), !(idx&4), 1.0);  // XXX debug colors
                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);   // Make sure we start from a clear texture.

                _currentRender = true;
                    scene->draw(sphereShader, zFar);
                _currentRender = false;
            }
    
    if(ok)
        qDebug("Rendered occlusion grid");

    // ---- Restore previous environment -------------------------------------
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    glClearColor( 0.0, 0.0, 0.0, 1.0);
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glPopAttrib();                                  // pop viewport.
    CHECK_GL_ERROR("unwind opengl matrices");

    return ok;
}

//////////////////////////////////////////////////////////////////////////////

void OcclusionGrid::renderAll(const Scene* scene, float zFar)
{
    if( unsupported ) return;

    int n=0, f=0;
    double time = TBender::time();

    for(OcclusionGrids::iterator it = allOcclusionGrids.begin(); it != allOcclusionGrids.end(); ++it, ++n)
    {
        if( (*it)->render(scene, zFar) )
        {
            f += (*it)->_nu * (*it)->_nv;
            if( scene->env.dumpImages )
            {
                time -= TBender::time();    // stop timer.

                QImage img;
                (*it)->getImage(img);
                img.save( QString("glossy%1Map.png").arg(n));

                time += TBender::time();    // restart timer.
            }
        }
        else
            break;
    }

    time = TBender::time() - time;
    if( n > 0 )
        qWarning("Mean glossy map rendering time = %s (mean on %d maps, with %d elements each)",
            TBender::timeCString(time / n).toLatin1().constData(),
            n, f/n);
}

//////////////////////////////////////////////////////////////////////////////

void OcclusionGrid::setSize(int nu, int nv)
{
    clean();
    _nu = nu; _nv = nv;
}

//////////////////////////////////////////////////////////////////////////////

void OcclusionGrid::setResolution(int r)
{
    // Make sure the image is correctly initialized to hold our depth
    // data.
    static int maxRBSize = 0;
    if( ! maxRBSize )
        glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &maxRBSize);
    Q_ASSERT( 0 < r && r <= maxRBSize );

    //------------------------------------------------------------------------

    clean();
    _size = r;
}

//////////////////////////////////////////////////////////////////////////////

void OcclusionGrid::setPosition(const Vec3f& center, const Vec3f& u, const Vec3f& v)
{
    _center = center;
    _u      = u;
    _v      = v;
}

//////////////////////////////////////////////////////////////////////////////

bool OcclusionGrid::enable(int texUnit, const Shader* shader) const
{
    if( unsupported ) return false;

    if( _currentRender )    // Object self-culling during render()
        return true;

    if( ! _texId )
    {
        qDebug("Skiping glossy map: not yet rendered");
        return false;
    }

    glActiveTexture( GL_TEXTURE0 + texUnit );
    glBindTexture(GL_TEXTURE_2D_ARRAY_EXT, _texId);
    CHECK_GL_ERROR("bind glossy tex array");

    if( ! shader )
    {
        shader = phongShader;
        shader->enable();   // If a scene shader is used, it is already enabled.
    }

    // For any point P = C + x U + y V, we do have :
    // glossyMatrix * P = (a .(C + x U + y V) - a.(C-U), b.(C + x U + y U) -b.(C-U))
    //                  = (a . (x+1)U, b . (y+1) V)
    //                  = ( (x+1)/2 nu, (y+1)/2 nv )
    // Hence it maps (x,y) in [-1..1]x[-1..1] to [0..nu, 0..nv]
    Matrix4 glossyMatrix;
    Vec3f a = _u * ( 0.5f * _nu / length2(_u));
    Vec3f b = _v * ( 0.5f * _nv / length2(_v));
    glossyMatrix[0][0] = a._x;
    glossyMatrix[0][1] = a._y;
    glossyMatrix[0][2] = a._z;
    glossyMatrix[0][3] = - dot(a, _center-_u);
    glossyMatrix[1][0] = b._x;
    glossyMatrix[1][1] = b._y;
    glossyMatrix[1][2] = b._z;
    glossyMatrix[1][3] = - dot(b, _center-_v);

    // TODO glossyMatrix *= camera()->modelViewMatrix().inverse();

    shader->setUniform("uHasOcclusionGrid", true, GL_BOOL);
    shader->setUniform("glossyMap", _nu);
    shader->setUniform("uGlossyNV", _nv);
    shader->setUniform("uGlossyMatrix", (float*)glossyMatrix, 0x8B69); // GL_FLOAT_MAT4x2_EXT ???

    return true;
}

//////////////////////////////////////////////////////////////////////////////

void OcclusionGrid::disable(int texUnit, const Shader* shader) const
{
    if( unsupported ) return;

    if( _currentRender )    // Object self-culling during render()
        return;

    if( ! _texId ) return;

    glActiveTexture( GL_TEXTURE0 + texUnit );
    glBindTexture(GL_TEXTURE_2D_ARRAY_EXT, 0);
    CHECK_GL_ERROR("unbind glossy tex array");

    if( shader )
        shader->setUniform("uHasOcclusionGrid", false, GL_BOOL);
    else
    {
        phongShader->setUniform("uHasOcclusionGrid", false, GL_BOOL);
        phongShader->disable();
    }
}

//////////////////////////////////////////////////////////////////////////////



bool
OcclusionGrid::getImage(QImage& image) const
{
    if( unsupported ) return false;

    image = QImage(_nu*_size, _nv*_size, QImage::Format_RGB32);
    image.fill( qRgba(128,0,128,255) );     // Violet where no write occurs.

    // Note : There is no GetTexSubImage() neither GetTexImageLayer OpenGL call.
    //        Using GetTexImage() returns all layers at once, creating a 
    //        size x (nu*nv*size) very tall image.
    //        To be able to have a (nu*size) x (nv*size) image, we have to
    //        read layers one at a time. This is done by attaching
    //        each layer to the FBO, then reading pixels from there...

    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
    glReadBuffer(GL_COLOR_ATTACHMENT0_EXT);
    CHECK_GL_ERROR("bind glossy map");

    glPixelStorei(GL_PACK_ROW_LENGTH, image.width());
    CHECK_GL_ERROR("pixelstore");

    for(int u=0; u<_nu; ++u)
        for(int v=0; v<_nv; ++v)
        {
            // Attach the corresponding layer of the texture array :
            int idx = v + u * _nv;
            glFramebufferTextureLayerEXT(GL_FRAMEBUFFER_EXT,    // target
                                      GL_COLOR_ATTACHMENT0_EXT, // attachment
                                      _texId, 0,                // dest texture & mipmap level
                                      idx);                     // dest layer (aka zoffset)
            CHECK_GL_ERROR("attach glossy map layer");

            void* p = image.scanLine(v*_size) + 4*u*_size;
            glReadPixels(0, 0,              // Raster position within this layer.
                         _size, _size,      // Raster size
                         GL_BGRA,           // Format
                         GL_UNSIGNED_BYTE,  // Type
                         p);                // Buffer ptr
            CHECK_GL_ERROR("get glossy mosaic");
        }

    glPixelStorei(GL_PACK_ROW_LENGTH, 0);
    glReadBuffer(GL_NONE);
    glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
    CHECK_GL_ERROR("unbind glossy map");

    return true;
}
