////////////////////////////////////////////////////////////////////////
/*! \file   OcclusionGrid.h
 *  \brief  Grid storing ambiant occlusion around an object.
 *  \author J.D. Gascuel
 *
 * \sa Scene, Environment, Shader.
 */
////////////////////////////////////////////////////////////////////////
// $Id: OcclusionGrid.h,v 1.3 2007/06/18 14:29:17 gascuel Exp $
// File History:
// [jDG 2007/06/14] : Creation from GlossyMap.

#include "Math/Vec3f.h"

class Shader;
class Scene;
class Environment;

class OcclusionGrid
{
    int _nu, _nv, _nw;          //!< 3D grid sizes.
    float* _data;               //!< Occlusions coefficients [nu * nv * nw].
    Vec3f _origin;              //!< Grid's base position. In Scene coordinates.
    Vec3f _u, _v, _w;           //!< grid cell edges. In Scene coordinates.

    int _size;                  //!< Texture size used for computing coefficients.
    GLuint* _queries;           //!< IDs of occlusions queries used to compute grid values.

    static Shader* phongShader; //!< Apply occlusion grids on shaderless objects.
    static Shader* sphereShader;//!< Render occlusion maps.

    void clean();               //!< Free internal texture array.
    bool render(const Scene*, float zFar);  //!< Allocate & compute internal texture array.

public:
    OcclusionGrid();
    ~OcclusionGrid();

    //! Reload internal shaders (Scene environment have changed).
    static bool reloadShaders(const Environment& env);

    //! Re-render all OcclusionGrids.
    static void renderAll(const Scene* scene, float zFar);

    //! Change number of textures in the grid. Reset all data.
    void setSize(int, int, int);

    //! Change resolution of individual textures. Reset all data.
    void setResolution(int);

    //! Change map's center, orientation and size.
    void setPosition(const Vec3f& center, const Vec3f& u, const Vec3f& v, const Vec3f& w);

    //! Setup a shader to start texturing with the glossy map.
    bool enable(int texUnit, const Shader*) const;

    //! Back to normal rendering mode.
    void disable(int texUnit, const Shader*) const;

    //! Returns a mosaic image of nu * nv * nw occlusion coefficients.
    bool getImage(QImage&) const;
};
