//////////////////////////////////////////////////////////////////////////////
/*! \file   Scene.h
 *  \brief  A vertex array based scene format.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Vec3f.
 */
//////////////////////////////////////////////////////////////////////////////
// $Id: Scene.cpp,v 1.80 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2006/11/29] : Creation.

#include "Scene.h"

#include <float.h>

#include <GL/glew.h>
#include <QtOpenGL>
#include <qglobal.h>
#include <QHash>

#include "Math/Matrix4.h"

#include "Utilities.h"
#include "Shader.h"
#include "ShadowMap.h"
#include "EnvironmentMap.h"
#include "HemiCube.h"
#include "GlossyMap.h"

#define SHADOW_MAP_SIZE 2048

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Light::Light()
: position(0, 0, 1, 0),     // Directional, //, in the Z direction.
  ambient(1,1,1,1),         // No default ambient color.
  diffuse(1,1,1,1),         // White colored.
  specular(1,1,1,1),        // With white high-lights.
  attenuation(1.0,0.0,0.0), // Without any attenuation.
  exponent(0.0),            // Uniform spot.
  cutoff(180.0)             // This is a directional spot.
{}

//////////////////////////////////////////////////////////////////////////////

void
Light::enable(int l) const
{
    glEnable(GL_LIGHT0+l);

    glLightfv(GL_LIGHT0+l, GL_POSITION, (const GLfloat*) position);

    glLightfv(GL_LIGHT0+l, GL_AMBIENT,  (const GLfloat*) ambient);
    glLightfv(GL_LIGHT0+l, GL_DIFFUSE,  (const GLfloat*) diffuse);
    glLightfv(GL_LIGHT0+l, GL_SPECULAR, (const GLfloat*) specular);

    glLightf(GL_LIGHT0+l, GL_CONSTANT_ATTENUATION,  attenuation._x);
    glLightf(GL_LIGHT0+l, GL_LINEAR_ATTENUATION,    attenuation._y);
    glLightf(GL_LIGHT0+l, GL_QUADRATIC_ATTENUATION, attenuation._z);

    // Additional SPOT parameters ?
    if( position._w != 0.0f && cutoff == 180.0f )
    {
        glLightfv(GL_LIGHT0+l, GL_SPOT_DIRECTION, (const GLfloat*) - normalize(position));
        glLightf (GL_LIGHT0+l, GL_SPOT_EXPONENT,  exponent);
        glLightf (GL_LIGHT0+l, GL_SPOT_CUTOFF,    cutoff);
    }

    CHECK_GL_ERROR("enable light");
}

//////////////////////////////////////////////////////////////////////////////

void
Light::disable(int l) const
{
    glDisable(GL_LIGHT0+l);
    CHECK_GL_ERROR("disable light");
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Environment::Environment()
: dumpImages(true),
  nbLights(1),
  nbShadows(1),
  dynamicShadows(false),
  shadowsDirty(true),
  primitiveFilter(-1),
  wireframe(RENDER_FACES),
  culling(BACK_FACE),
  depthTest(true),
  blending(true),
  colorMaps(true),
  normalMaps(false),
  envMaps(false),
  envMapMode(EnvironmentMap::FACES_MAP),
  dynamicEnvMap(false),
  envMapsDirty(true),
  glossyMaps(false),
  dynamicGlossyMaps(false),
  glossyMapsDirty(true),
  twoPass(false),
  testOverdraw(false),
  testShading(false),
  testNoFrags(false),
  testNoGeomShader(false)
{
    static Vec4f lightDir[8][2] = {
    //  position                diffuseColor
        { Vec4f(-1, 1,  1, 0),    Vec4f(1.0f,0.8f,0.8f,0) },   // Main spot, warm, from upper-left-behind.
		{ Vec4f( 1,-1, -1, 0),    Vec4f(0.2f,0.4f,0.4f,0) },   // Second spot : cold, back light.

		{ Vec4f( 0, 1,  0, 0),    Vec4f(0.5f,0.5f,0.4f,0) },		    // Third spot : top = Y
        { Vec4f( 0,-1,  0, 0),    Vec4f(0.1f,0.1f,0.2f,0) },		    //     and bottom back light.

        { Vec4f(-1, 0,  0, 0),    Vec4f(0.2f,0.2f,0.1f,0) },		    // All othe : left = -X, and front = -Z.
        { Vec4f( 1, 0,  0, 0),    Vec4f(0.1f,0.1f,0.2f,0) },
        { Vec4f( 0, 0,  1, 0),    Vec4f(0.2f,0.1f,0.2f,0) },
        { Vec4f( 0, 0, -1, 0),    Vec4f(0.1f,0.2f,0.1f,0) }
    };

    for(int l=0; l<MAX_LIGHTS; ++l)
    {
        lights[l].position = normalize(lightDir[l][0]);
        lights[l].ambient  = (l==0) ? Vec4f(0.5,0.5,0.5,1) : Vec4f(0,0,0,1);
        lights[l].diffuse  = lightDir[l][1];
        lights[l].specular = Vec4f(0,0,0,1); // Vec4f(0.1f,0.1f,0.1f,1) / (l+1.0f);
        lights[l].ambient._w = lights[l].diffuse._w = lights[l].specular._w = 1.0f;

        shadows[l] = 0;
    }
}

Environment::~Environment()
{
    for(int s=0; s<MAX_LIGHTS; ++s)
        if( shadows[s] )
            delete shadows[s];
}

//////////////////////////////////////////////////////////////////////////////

void Environment::prepare(const Shader* shader, const Scene* scene, int subMode)
{
    // ---- UIpdates scene bounds, if needed ---------------------------------
    Vec3f min, max;
    scene->getBBox(min, max);
    Vec3f center =  0.5f*(max+min);
    float radius = 0.5f * length(max-min);

    // ---- Recompute shadow maps, if needed ---------------------------------
	if( nbLights > 0 && nbShadows > 0 && (shadowsDirty || dynamicShadows) )
	{
        glEnable(GL_DEPTH_TEST);	// no test ==> don't write z buffer either !

		glDisable(GL_LIGHTING);		// Faster rendering. We just need depth here.
		glDisable(GL_BLEND);
		glEnable(GL_CULL_FACE);
		glShadeModel(GL_FLAT);
		glColorMask(0, 0, 0, 0);

        // Kill lighting and fancy stuff while drawing depth maps...
        bool savedColorMaps  = colorMaps;   colorMaps  = false;
        bool savedNormalMaps = normalMaps;  normalMaps = false;
        bool savedEnvMaps    = envMaps;     envMaps    = false;
        int  savedLights     = nbLights;    nbLights   = 0;

		for(int l=0; l<nbShadows; ++l)
		{
			// Make sure shadow maps are initialised :
			if( ! shadows[l] )
			{
				shadows[l] = new ShadowMap;
				Q_CHECK_PTR( shadows[l] );
				shadows[l]->bindDepthTexture(SHADOW_MAP_SIZE, SHADOW_MAP_SIZE);
			}

			// Dynamic redraw of shadow maps :
			shadows[l]->setupLightMatrix(2.5f * radius * Vec3f(lights[l].position), center, radius);
				scene->draw(0, 1.45f * radius, 3.55f * radius);             // No shader fir depth map
			shadows[l]->cleanMatrix();

            if( dumpImages )
            {
			    QImage img(SHADOW_MAP_SIZE, SHADOW_MAP_SIZE, QImage::Format_Indexed8);
			    shadows[l]->finish(img);
			    img.save( QString("light%1Map.png").arg(l) );
			}
		}
		glShadeModel(GL_SMOOTH);	// Back to normal rendering state.
		glColorMask(1, 1, 1, 1);

        colorMaps  = savedColorMaps;
        normalMaps = savedNormalMaps;
        envMaps    = savedEnvMaps;
        nbLights   = savedLights;

        shadowsDirty = false;       // done once.
	}

    // ---- Recompute env maps, if needed ------------------------------------
    if( envMaps && ( dynamicEnvMap || envMapsDirty ) )
    {
        EnvironmentMap::renderAllMaps(shader, scene, subMode);
        envMapsDirty = false;
    }

    // ---- Recompute glossy maps, if needed ---------------------------------
    if( glossyMaps && ( dynamicGlossyMaps || glossyMapsDirty ) )
    {
        GlossyMap::renderAll(scene, subMode);
        glossyMapsDirty = false;
    }
}

//////////////////////////////////////////////////////////////////////////////

QString 
Environment::envString() const
{
    QString env;
    if( culling )           env += "#define FACE_CULLING 1\n";
    if( nbLights )          env += "#define PER_PIXEL_LIGHTINGS 1\n";
	if( nbShadows )         env += "#define SHADOW_MAPPING 1\n";
	if( colorMaps )         env += "#define COLOR_TEXTURE 0\n";
	if( normalMaps )        env += "#define COLOR_TEXTURE 0\n" "#define BUMP_MAPPING 1\n";
    if( envMaps )           env += "#define ENV_MAP 1\n";
    if( glossyMaps )        env += "#define GLOSSY_MAP 1\n";
    if( testShading )       env += "#define TEST_SHADING 1\n";

    switch( envMapMode )
    {
    case EnvironmentMap::FACES_MAP:
        break;
    case EnvironmentMap::SPHERE_MAP:   env += "#define SPHERE_ENV_MAP 1\n";    break;  // NYI
    case EnvironmentMap::PARABOLA_MAP: env += "#define PARABOLIC_ENV_MAP 1\n"; break;
    case EnvironmentMap::LAMBERT_MAP:  env += "#define LAMBERT_ENV_MAP 1\n";   break;  // NYI
    }
    
    return env;
}

//////////////////////////////////////////////////////////////////////////////

bool
Environment::enable(const Shader* shader, float zNear, float zFar, int pass) const
{
    if( shader )
    {
        shader->enable();
        shader->setUniform("uPrimitiveFilter", primitiveFilter);
        shader->setUniform("uNear",            zNear);
        shader->setUniform("uFar",             zFar);
        shader->setUniform("uLights",          nbLights);
        shader->setUniform("uTestOverdraw",    testOverdraw, GL_BOOL);
    }

    // ---- Setup lighting environment ---------------------------------------
    bool castingShadows = false;
    if( nbLights > 0 )
    {
        Q_ASSERT(nbLights <= MAX_LIGHTS);

        for(int l=0; l<nbLights; ++l)
		{
            // Does this light casts shadow ?
            bool castShadow = l < nbShadows && shadows[l];

            if( shader || (pass == -1 && !castShadow) || (pass==l && castShadow) )
            {
                lights[l].enable(l);

			    // Then enable shadow map algorithm, if any.
			    if( castShadow )
                {
				    shadows[l]->enable(l+2, shader);    // Starts at texUnit 2.
                    castingShadows = true;
                }
            }
            else
                glDisable(GL_LIGHT0 + l);
		}

        for(int l=nbLights; l < MAX_LIGHTS; ++l)
            glDisable(GL_LIGHT0 + l);
    }
    else
    {
        // Use just light 0, but don't tell shader to do per-vertex lighting.
        lights[0].enable(0);
    }
    glEnable(GL_LIGHTING);

    if( depthTest ) glEnable( GL_DEPTH_TEST);
    else            glDisable(GL_DEPTH_TEST);

    if( castingShadows && ! shader )
    {
        // Multi-pass no-shader mode : use blending to accumulate rendering
        // from each pass.
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    }
    else if( blending )
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
    }
    else
        glDisable(GL_BLEND);

    if( testNoFrags )
    {
        glScissor(0, 0, 1, 1);
        glEnable(GL_SCISSOR_TEST);
    }

    return castingShadows;
}

//////////////////////////////////////////////////////////////////////////////

void Environment::disable(const Shader* shader) const
{
    for(int l=0; l < nbLights; ++l)
	{
        if( l < nbShadows && shadows[l] )
			shadows[l]->disable(l+2, shader);
        lights[l].disable(l);
	}

    if( shader )
        shader->disable();

    // Put back OpenGL state in some default mode (eg. for QGLViewer axis).
    glEnable (GL_LIGHT0);
    glDisable(GL_LIGHTING);
    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_CULL_FACE);
    glDisable(GL_BLEND);
    glEnable ( GL_DEPTH_TEST);
    glDisable(GL_SCISSOR_TEST);

    CHECK_GL_ERROR("disable render state");
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

Texture::Texture(const QString& fileName, Texture::Mode mode)
: _mode(mode),
  _texId(0)
{
    _image = QImage(fileName);
    if( _image.isNull() )
    {
        qWarning("Unable to find texture '%s'",
            fileName.toLatin1().constData());
        _image = QImage(16,16, QImage::Format_RGB32);
        _image.fill( 0x80808080 );
    }

    if( ! GLEW_ARB_texture_non_power_of_two )
    {
        int w = (int)powf(2.0f, ceilf(logf( _image.width () ) / logf(2.0f)));
        int h = (int)powf(2.0f, ceilf(logf( _image.height() ) / logf(2.0f)));
        _image = _image.scaled( w,h);
    }

    // Automatically convert bump maps to normal maps. Finite difference is done
    // once at load time, not inside fragment shader !
    //
    if( mode == BUMP_MAPPING )
    {
        qWarning("Converting bump map to normal map");

        QImage nMap( _image.width(), _image.height(), QImage::Format_RGB32);

        for(int j=0; j<nMap.height(); ++j)
        {
            int j0 = max(0, j-1);
            int j1 = min(nMap.height()-1, j+1);
            for(int i=0; i<nMap.width(); ++i)
            {
                int i0 = max(0, i-1);
                int i1 = min(nMap.width()-1, i+1);

                Vec3f n( float( qGray(_image.pixel(i1,j)) - qGray(_image.pixel(i0,j)))/float(i1-i0),
                         float( qGray(_image.pixel(i,j1)) - qGray(_image.pixel(i,j0)))/float(j1-j0),
                         255.0f); // Biased normal. More flat than the real value !
                n.normalize();
                nMap.setPixel(i, j, qRgb( 127 + int(255*n._x),
                                          127 + int(255*n._y),
                                          127 + int(255*n._z)) );
            }
        }
        nMap.save( QString(fileName).section('.', 0, -2) + "_convertedNormalMap.png" );
        _image = nMap;
        _mode = NORMAL_MAPPING;
    }
}

//////////////////////////////////////////////////////////////////////////////

static QHash<QString, const Texture*> textures;

const Texture* Texture::make(const QString& fileName, Texture::Mode mode)
{
    const Texture* tex = textures[fileName];
    
    if( !tex  )
    {
        tex = new Texture(fileName, mode);
        textures[fileName] = tex;
    }

    return tex;
}

//////////////////////////////////////////////////////////////////////////////

void Texture::enable(const Shader* shader) const
{
	glActiveTexture( _mode == COLOR_REPLACE ? GL_TEXTURE0	// Uses tex unit 0 for color.
		                                    : GL_TEXTURE1);	// And tex unit 1 for bump mapping.

    if( ! _texId )
    {
        glGenTextures(1, &_texId);
        glBindTexture(GL_TEXTURE_2D, _texId);

        if( _image.depth() == 32 && _image.hasAlphaChannel() )
        {
            // Make sure storage don't have gaps between lines of RGBA pixels:
            Q_ASSERT( _image.scanLine(1) == _image.scanLine(0) + 4*_image.width() );

            glTexImage2D(GL_TEXTURE_2D,
                         0,
                         GL_RGBA,                       // Internal format.
                         _image.width(),
                         _image.height(),
                         0,                             // No border.
                         GL_BGRA,                       // QImage format : B and R swapped.
                         GL_UNSIGNED_BYTE,
                         _image.scanLine(0));
            CHECK_GL_ERROR("GL_TEXTURE_2D RGBA");
        }
        else if( _image.depth() == 32 )  // NO aplha channel
        {
            glTexImage2D(GL_TEXTURE_2D,
                         0,
                         GL_RGB,                        // Internal format.
                         _image.width(),
                         _image.height(),
                         0,                             // No border.
                         GL_BGRA,                       // QImage format : B and R swapped. Skip A
                         GL_UNSIGNED_BYTE,
                         _image.scanLine(0));
            CHECK_GL_ERROR("GL_TEXTURE_2D RGB");
        }
        else if( _image.depth() == 8 )  // Grey level image.
        {
            // Make sure lines are packed...
            Q_ASSERT( _image.scanLine(1) == _image.scanLine(0) + _image.width() );
            glTexImage2D(GL_TEXTURE_2D,
                         0,
                         GL_RGB,                        // Internal format.
                         _image.width(),
                         _image.height(),
                         0,                             // No border.
                         GL_LUMINANCE,                  // QImage format : Suppose it is a grey value (not an index).
                         GL_UNSIGNED_BYTE,
                         _image.scanLine(0));
            CHECK_GL_ERROR("GL_TEXTURE_2D LUMINANCE");
        }
        else
            Q_ASSERT( ! "Texture should be 32 bits, w or w/o alpha" );

        // Don't use mip-maps. Or we should load them !
        glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        CHECK_GL_ERROR("GL_TEXTURE_2D NEAREST");
    }
    else
    {
        glBindTexture(GL_TEXTURE_2D, _texId);
        CHECK_GL_ERROR("bind tex");
    }

    // Enable texture, either in OpenGL std pipeline, or in Shader mode...
    if( shader )
    {
		if( _mode == COLOR_REPLACE )
		{
	        shader->setUniform("uHasColorMap", true, GL_BOOL);
			shader->setUniform("uColorMap", 0, GL_SAMPLER_2D);		// Use tex unit 0.
		}
		else if( _mode == NORMAL_MAPPING )
		{
			shader->setUniform("uHasBumpMap", true, GL_BOOL);
			shader->setUniform("uBumpMap", 1, GL_SAMPLER_2D);		// Use tex unit 1.
		}
    }
    else
    {
		if( _mode == COLOR_REPLACE )
		{
			glEnable(GL_TEXTURE_2D);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			CHECK_GL_ERROR("ENV_MODE MODULATE");
		}
        //else if( _mode == NORMAL_MAPPING )
        //{
        // XXX : No OpenGL normal map support. The following change texture coordinates,
        //       It does not re-generate normals from a normal map texture !
        //
        //    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP);
        //    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_NORMAL_MAP);
        //    glEnable(GL_TEXTURE_GEN_S);
        //    glEnable(GL_TEXTURE_GEN_T);
        //}
    }
}

//////////////////////////////////////////////////////////////////////////////

void Texture::disable(const Shader* shader) const
{
	glActiveTexture( _mode == COLOR_REPLACE ? GL_TEXTURE0	// Uses tex unit 0 for color.
											: GL_TEXTURE1);	// And tex unit 1 for bump mapping.
    glBindTexture(GL_TEXTURE_2D, 0);

    if( shader )
    {
		if( _mode == COLOR_REPLACE )
	        shader->setUniform("uHasColorMap", false, GL_BOOL);
		else if( _mode == NORMAL_MAPPING )
			shader->setUniform("uHasBumpMap", false, GL_BOOL);
    }
    else
    {
        glDisable(GL_TEXTURE_2D);
        CHECK_GL_ERROR("disable tex");
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

ScnObject::ScnObject()
: _vbo(0),
  _ibo(0),
  _size(0),
  _culling(Environment::BACK_FACE),
  _loaded(true),
  _haveColor(false), _haveNormal(false), _haveTexture(false),
  _envMap(0),
  _glossyMap(0),
  ambient( 0.2f, 0.2f, 0.2f, 1.0f),	// Default color is white(80%), with 20% in ambient.
  diffuse( 0.8f, 0.8f, 0.8f, 1.0f),
  specular(0.2f, 0.2f, 0.2f, 1.0f), // Default slight highlights
  shininess(100.0f),                // With a small hotspot.
  colorTexture(0),
  bumpTexture(0)
{}

ScnObject::~ScnObject()
{
    if( _vbo ) glDeleteBuffers(1, &_vbo);
    if( _ibo ) glDeleteBuffers(1, &_ibo);
    _vbo = _ibo = 0;
    CHECK_GL_ERROR(0);

    if( _envMap ) delete _envMap;
    _envMap = 0;

    if( _glossyMap ) delete _glossyMap;
    _glossyMap = 0;
}

//////////////////////////////////////////////////////////////////////////////

ScnObject* ScnObject::copy() const
{
    ScnObject* o = new ScnObject;

    // ---- Copy rendering attributes ----------------------------------------
    o->setCulling( _culling );

    o->setAmbientColor(ambient._x, ambient._y, ambient._z, ambient._w);
    o->setDiffuseColor(diffuse._x, diffuse._y, diffuse._z, diffuse._w);
    o->setSpecularColor(specular._x, specular._y, specular._z, specular._w);
    o->setColorShininess( shininess );

    // ----- Copy object's data ----------------------------------------------
    if( _haveColor )
    {
        o->_haveColor = true;
        o->colors = colors;
    }
    if( _haveNormal )
    {
        o->_haveNormal = true;
        o->normals = normals;
    }
    if( _haveTexture )
    {
        o->_haveTexture = true;
        o->texcoord = texcoord;
        o->colorTexture = colorTexture; // Share texture ptr.
		o->bumpTexture = bumpTexture;
    }

    o->indices = indices;

    return o;
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::transform(const Matrix4& xfm)
{
    for(int v=0; v<vertices.count(); ++v)
        vertices[v] = xfm.coordinatesOf( vertices[v] );

    if( _haveNormal )
        for(int n=0; n<normals.count(); ++n)
            normals[n] = normalize(xfm.transformOf( normals[n] ));
}

//////////////////////////////////////////////////////////////////////////////
//
// Provides an acceleration structure for mergeVertex, hence to find
// vertices with arbitrary V/N/T/C attributes.
//
// This structure do lives only between a beginEdit()/endEdit() pair.
// There is only one object allowed to be edited at any time.
// So use a clumsy implementation through a simple shared static stuff.

typedef QVector<uint>       VBucket;        //!< A list of vertex index.
typedef QVector<VBucket>    VArray;         //!< A list of buckets, indexed by x coordinates.

static int rebuildAccel;                    //!< When to rebuild the whole array.
static VArray accel;

void ScnObject::beginEdit()
{
    Q_ASSERT( accel.isEmpty() );

    // ---- Determine a scaling to map X coordinates into the array range...
    float min=0.0f, max=1.0f;               // Defaults if the scene is empty.
    for(int v=0; v<vertices.count(); ++v)
    {
        if( vertices[v]._x < min ) min = vertices[v]._x;
        if( vertices[v]._x > max ) max = vertices[v]._x;
    }

    // Use 
    int n = vertices.count() / 100;
    if( n < 100 )  n = 100;
    if( n > 5000 ) n = 5000;
    rebuildAccel = n;           // Should rebuild once we doubled population.
    accel.resize(n);

    // ---- Insert all vertices into the array -------------------------------

    for(int v=0; v<vertices.count(); ++v)
    {
        int i = hash(vertices[v]);
        accel[i].push_back(v);
    }
}

//////////////////////////////////////////////////////////////////////////////

unsigned int ScnObject::hash(const Vec3f& V) const
{
    unsigned int h  = qHash( int(V._x * 1000.0f) );
                 h ^= qHash( int(V._y * 1001.0f) );
                 h ^= qHash( int(V._z * 1002.0f) );
    return h % accel.count();
}

int ScnObject::findVertex(const Vec3f& V, 
                          const Vec3f& N,  
                          const Vec2f& T, 
                          const Vec4f& C) const
{
    Q_ASSERT( ! accel.isEmpty() );      // Should have at last 100 (possibly empty) buckets.

    int maxN = normals.count()-1;
    int maxT = texcoord.count()-1;
    int maxC = colors.count()-1;

    int i = hash(V);
    for(VBucket::ConstIterator it = accel[i].begin(); it != accel[i].end(); ++it)
    {
        int v = *it;
        if(       V != vertices[v] ) continue;

        // Vertex have normal/texture/color, but different from the
        // already known vertex ?
        if( &N && N != normals [v] ) continue;
        if( &T && T != texcoord[v] ) continue;
        if( &C && C != colors  [v] ) continue;

        // Vertex do not have normal/texture/color, but the known
        // vertex does ?
        if( !&N && v <= maxN ) continue;
        if( !&T && v <= maxT ) continue;
        if( !&C && v <= maxC ) continue;
        return v;
    }
    return -1;
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::endEdit()
{
    // ---- Make sure to free any space used by the acceleration list...
    accel.clear();

    _loaded = false;
    _haveColor   = colors  .count() == vertices.count();
    _haveNormal  = normals .count() == vertices.count();
    _haveTexture = texcoord.count() == vertices.count();

}

//////////////////////////////////////////////////////////////////////////////

int ScnObject::mergeVertex(const Vec3f& V, 
                           const Vec3f& N,  
                           const Vec2f& T, 
                           const Vec4f& C)
{
    // ---- Use accelerated search -------------------------------------------
    // Doing a linear search make a O(n^2) merge time, hich is a very bad idea
    // for 100K vertex objects 8-)
    int v = findVertex(V, N, T, C);
    if( v >= 0 ) return v;

    // ---- Insert a new vertex ----------------------------------------------
    v = vertices.count();
    vertices.push_back(V);

    // If we do have companion attributes, make sure they are always present...
    if( &N ) { normals .push_back(N); Q_ASSERT(vertices.count() == normals .count()); }
    if( &T ) { texcoord.push_back(T); Q_ASSERT(vertices.count() == texcoord.count()); }
    if( &C ) { colors  .push_back(C); Q_ASSERT(vertices.count() == colors  .count()); }

    // ---- Update acceleration structure ------------------------------------
    if( (--rebuildAccel)<0 )      // Out or obsolete table ?
    {
        endEdit();                      // Clean,
        beginEdit();                    // And rebuild the whole table.
    }
    int i = hash(V);
    accel[i].push_back(v);              // Just insert the new vertex.

    // ---- Done -------------------------------------------------------------
    return v;
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::insertFaceVertex(int idx)
{
    indices.push_back(idx);
}

int ScnObject::insertFace(int len)
{
    Q_ASSERT(len == 3);
    return indices.count()-3;
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::setAmbientColor(float r, float g, float b, float a)
{
    ambient._x = r;
    ambient._y = g;
    ambient._z = b;
    ambient._w = a;
}

void ScnObject::setDiffuseColor(float r, float g, float b, float a)
{
    diffuse._x = r;
    diffuse._y = g;
    diffuse._z = b;
    diffuse._w = a;
}

void ScnObject::setSpecularColor(float r, float g, float b, float a)
{
    specular._x = r;
    specular._y = g;
    specular._z = b;
    specular._w = a;
}

void ScnObject::setColorShininess(float s)
{
    shininess = s;

    if( s == 1000.0f )
    {
        Vec3f bmin(FLT_MAX, FLT_MAX, FLT_MAX), bmax(-FLT_MAX, -FLT_MAX, -FLT_MAX);
        enlargeBBox(bmin, bmax);
        Vec3f center = 0.5f  * (bmin+bmax);
        float innerRadius = 0.000001f;

        _envMap = new EnvironmentMap(EnvironmentMap::FACES_MAP, center, 512, innerRadius);
        Q_CHECK_PTR(_envMap);
    }
    else if( s == 999.0f )
    {
        Vec3f bmin(FLT_MAX, FLT_MAX, FLT_MAX), bmax(-FLT_MAX, -FLT_MAX, -FLT_MAX);
        enlargeBBox(bmin, bmax);
        Vec3f center = 0.5f  * (bmin+bmax);
        float innerRadius = 0.000001f;

        _envMap = new HemiCube(HemiCube::FACES_MAP, center, 512, innerRadius);
        Q_CHECK_PTR(_envMap);
    }
    else if( s == 998.0f )
    {
        Vec3f bmin(FLT_MAX, FLT_MAX, FLT_MAX), bmax(-FLT_MAX, -FLT_MAX, -FLT_MAX);
        enlargeBBox(bmin, bmax);
        Vec3f center = 0.5f  * (bmin+bmax);
        Vec3f u(bmax._x-center._x, 0, 0);       // X
        Vec3f v(0,0,center._z-bmax._z);         // -Z

        _glossyMap = new GlossyMap;
        Q_CHECK_PTR(_glossyMap);

        _glossyMap->setPosition(center, u, v);

    }
}

void ScnObject::setCulling(Environment::Culling c)
{
    _culling = c;
}

void ScnObject::setTexture(const QString& fileName, Texture::Mode mode)
{
	if( mode == Texture::COLOR_REPLACE )
		colorTexture = Texture::make(fileName, mode);
	else if( mode == Texture::BUMP_MAPPING || mode == Texture::NORMAL_MAPPING )
		bumpTexture = Texture::make(fileName, mode);
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::reload()
{
    // ---- Make sure we have a VBO with the correct size --------------------
    uint newSize = 3*vertices.count();
    if( _haveColor   ) newSize += 4*colors  .count();
    if( _haveNormal  ) newSize += 3*normals .count();
    if( _haveTexture ) newSize += 2*texcoord.count();

    if( newSize != _size || !_vbo)
    {
        glGenBuffers(1, &_vbo);
        glBindBuffer(GL_ARRAY_BUFFER, _vbo);
        glBufferData(GL_ARRAY_BUFFER, newSize * sizeof(float), NULL, GL_STATIC_DRAW);
        CHECK_GL_ERROR("alloc vbo");
        _size = newSize;
    }
    else
        glBindBuffer(GL_ARRAY_BUFFER, _vbo);

    // ---- Upload data into the VBO -----------------------------------------

    GLintptr ptr = 0;
    glBufferSubData(GL_ARRAY_BUFFER, ptr, vertices.count() * sizeof(Vec3f), vertices.data());
    ptr += vertices.count() * sizeof(Vec3f);

    if( _haveColor )
    {
        glBufferSubData(GL_ARRAY_BUFFER, ptr, colors.count() * sizeof(Vec4f), colors.data());
        ptr += colors.count() * sizeof(Vec4f);
    }
    if( _haveNormal )
    {
        glBufferSubData(GL_ARRAY_BUFFER, ptr, normals.count() * sizeof(Vec3f), normals.data());
        ptr += normals.count() * sizeof(Vec3f);
    }
    if( _haveTexture )
    {
        glBufferSubData(GL_ARRAY_BUFFER, ptr, texcoord.count() * sizeof(Vec2f), texcoord.data());
        ptr += texcoord.count() * sizeof(Vec2f);
    }
    CHECK_GL_ERROR("reload vbo");
    Q_ASSERT( ptr == GLintptr(newSize * sizeof(float)) );

    // ---- Defines & upload indixes into the IBO ----------------------------

    if( !_ibo )
        glGenBuffers(1, &_ibo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _ibo);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, indices.count() * sizeof(uint), indices.data(), GL_STATIC_DRAW);
    CHECK_GL_ERROR("reload ibo");

    _loaded = true;
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::draw(const Shader* shader, const Environment* env) const
{
    if( ! _vbo ) const_cast<ScnObject*>(this)->reload();
    Q_ASSERT( _vbo );

    // ---- Cull object for which we are recomputing the env map -------------
    if( _envMap && _envMap->currentRender() )
        return;

    // ---- Setup global state for this object -------------------------------

    if( _haveColor ) // Per-pixel color. 
    {
        // Allow per-vertex color to update these material :
        glEnable(GL_COLOR_MATERIAL);
        glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
        glColor4fv( (const float*) diffuse );   // Send initial value, before shader stuff.
        CHECK_GL_ERROR("vertex material");
    }
    else
    {
        // Setup default material for this VBO :
        glDisable(GL_COLOR_MATERIAL);
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT,  (const float*) ambient );
        glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE,  (const float*) diffuse );
        CHECK_GL_ERROR("object material");
    }

    // Give every object kind a metalo-plastic look :
    // (Except when there is an auto-envmap on that object)
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, (const float*) specular );
    glMaterialf (GL_FRONT_AND_BACK, GL_SHININESS, ( shininess < 900.0f ) ? shininess : 100.0f);
    CHECK_GL_ERROR("specular");
    
    // ---- Handle dynamic culling -------------------------------------------
    int cull = int(env->culling) * int(_culling);		// Both global and per-object choice.

	if( shader )
        shader->setUniform("uCulling", cull);			// Both global and local culling, if any.

    if( cull == Environment::NO_CULLING )
    {
        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, true);   // Use both front and back materials.
        glDisable(GL_CULL_FACE);
    }
    else
    {
        glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, false);  // Use only front material.
        glEnable(GL_CULL_FACE);
        glCullFace(GL_BACK);                            // Always cull back-face polygons.
        glFrontFace(cull == Environment::BACK_FACE ? GL_CCW : GL_CW);    // Who is back ?
    }
    CHECK_GL_ERROR("culling");

    // ---- Setup texture, if any --------------------------------------------
    if( env->envMaps && _envMap )
        _envMap->enable(0, shader);
    else if( env->glossyMaps && _glossyMap )
        _glossyMap->enable(0, shader);
    else if( env->colorMaps && colorTexture )
        colorTexture->enable(shader);

	if( env->normalMaps && bumpTexture )
		bumpTexture->enable(shader);

    // ---- Bind vertices and vertex data ------------------------------------
    Q_ASSERT( _vbo );
    glBindBuffer(GL_ARRAY_BUFFER, _vbo);

    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, 0, 0);
    uint ptr = vertices.count() * sizeof(Vec3f);

    if( _haveColor )
    {
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(4, GL_FLOAT, 0, (void *)ptr);
        ptr += colors.count() * sizeof(Vec4f);
    }
    if( _haveNormal )
    {
        glEnableClientState(GL_NORMAL_ARRAY);
        glNormalPointer(GL_FLOAT, 0, (void *)ptr);
        ptr += normals.count() * sizeof(Vec3f);
    }
    if( _haveTexture && ! _envMap )
    {
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        glTexCoordPointer(2, GL_FLOAT, 0, (void *)ptr);
        ptr += normals.count() * sizeof(Vec2f);
    }
    CHECK_GL_ERROR("enable vbo");

    // ---- Bind indices -----------------------------------------------------
    Q_ASSERT(_ibo);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, _ibo);

    glEnableClientState(GL_INDEX_ARRAY);
    glIndexPointer(GL_INT, 0, 0);
    CHECK_GL_ERROR("enable ibo");

    // ---- Draw all at once -------------------------------------------------

    glDrawRangeElements(GL_TRIANGLES, 0, vertices.count()-1,
                        indices.count(), GL_UNSIGNED_INT, 0);
    CHECK_GL_ERROR("draw range elements");

    // ---- Unbind all VBOs --------------------------------------------------

    glDisableClientState(GL_VERTEX_ARRAY);
    if( _haveColor   ) glDisableClientState(GL_COLOR_ARRAY);
    if( _haveNormal  ) glDisableClientState(GL_NORMAL_ARRAY);
    if( _haveTexture ) glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_INDEX_ARRAY);

    glBindBuffer(GL_ARRAY_BUFFER, 0);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0);
    CHECK_GL_ERROR("disable vbo/ibo");

    // ---- Unbind all Textures ----------------------------------------------
    if( env->envMaps && _envMap )
        _envMap->disable(0, shader);
    else if( env->glossyMaps && _glossyMap )
        _glossyMap->disable(0, shader);
    else if( env->colorMaps && colorTexture )
        colorTexture->disable(shader);

    if( env->normalMaps && bumpTexture )
        bumpTexture->disable(shader);
}

//////////////////////////////////////////////////////////////////////////////

void Scene::drawPass(const Shader* shader) const
{
    if( env.wireframe & Environment::RENDER_FACES )
    {
        for(int o=0; o<objects.count(); ++o)
            objects[o]->draw(shader, &env);
    }

    if( env.wireframe & Environment::RENDER_EDGES )
    {
        glLineWidth(1.0f);
        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	    glEnable(GL_POLYGON_OFFSET_LINE);
	    glDisable(GL_POLYGON_OFFSET_FILL);
        glPolygonOffset(0, 1);

        for(int o=0; o<objects.count(); ++o)
            objects[o]->draw(shader, &env);

	    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	    glDisable(GL_POLYGON_OFFSET_LINE);
    }
}

//////////////////////////////////////////////////////////////////////////////

void Scene::draw(const Shader* shader, float zNear, float zFar) const
{
    Q_ASSERT( zNear > 0.0f && zNear < zFar );

    if( shader )
    {
        // Shaders enable one pass for all lights (casting shadow or not)
        // at once.
        env.enable(shader, zNear, zFar);   // Enable all lights at once:
        drawPass(shader);
        env.disable(shader);       // and back to normal OpenGL state.
    }
    else
    {
        // Draw all lights that do not cast shadows.
        env.enable(0, zNear, zFar, -1);
        drawPass(0);

        // Then add drawing of all shadow-casting lights :
        for(int l=0; l<env.nbLights; ++l)
            if( env.enable(0, zNear, zFar, l) )
                drawPass(0);

        // and back to normal OpenGL state.
        env.disable(0);
    }
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::enlargeBBox(Vec3f& min, Vec3f& max) const
{
    for(int i=0; i<vertices.count(); ++i)
    {
        const Vec3f &v = vertices[i];

        if( v._x < min._x ) min._x = v._x;
        if( v._y < min._y ) min._y = v._y;
        if( v._z < min._z ) min._z = v._z;

        if( v._x > max._x ) max._x = v._x;
        if( v._y > max._y ) max._y = v._y;
        if( v._z > max._z ) max._z = v._z;
    }
}

void Scene::getBBox(Vec3f& min, Vec3f& max) const
{
    min.set( FLT_MAX,  FLT_MAX,  FLT_MAX);
    max.set(-FLT_MAX, -FLT_MAX, -FLT_MAX);

    for(int o=0; o<objects.count(); ++o)
        objects[o]->enlargeBBox(min, max);
}

//////////////////////////////////////////////////////////////////////////////

Scene::Scene()
{
}

void Scene::clear()
{
    for(int o=0; o<objects.count(); ++o)
        delete objects[o];
    objects.clear();
}

Scene::~Scene()
{
    clear();
}

//////////////////////////////////////////////////////////////////////////////

ScnObject*
Scene::makeBox(const Vec3f& min, const Vec3f& max)
{
    ScnObject *o = new ScnObject;

    o->beginEdit();

    const Vec3f& n = *((Vec3f*)0);

    int a, b, c, d;
#define INSERT_QUAD(a,b,c,d)    \
    o->insertFaceVertex( a );   \
    o->insertFaceVertex( b );   \
    o->insertFaceVertex( c );   \
    o->insertFace(3);           \
    o->insertFaceVertex( a );   \
    o->insertFaceVertex( c );   \
    o->insertFaceVertex( d );   \
    o->makeNormals();

    // Face Z = min.
    a = o->mergeVertex( Vec3f(min._x, min._y, min._z), n, Vec2f(0,0));
    b = o->mergeVertex( Vec3f(min._x, max._y, min._z), n, Vec2f(0,1));
    c = o->mergeVertex( Vec3f(max._x, max._y, min._z), n, Vec2f(1,1));
    d = o->mergeVertex( Vec3f(max._x, min._y, min._z), n, Vec2f(1,0));
    INSERT_QUAD(a,b,c,d);

    // Face Z = max.
    a = o->mergeVertex( Vec3f(min._x, min._y, max._z), n, Vec2f(0,0));
    b = o->mergeVertex( Vec3f(min._x, max._y, max._z), n, Vec2f(0,1));
    c = o->mergeVertex( Vec3f(max._x, max._y, max._z), n, Vec2f(1,1));
    d = o->mergeVertex( Vec3f(max._x, min._y, max._z), n, Vec2f(1,0));
    INSERT_QUAD(d,c,b,a);

    // Face X = min.
    a = o->mergeVertex( Vec3f(min._x, min._y, min._z), n, Vec2f(0,0));
    b = o->mergeVertex( Vec3f(min._x, min._y, max._z), n, Vec2f(0,1));
    c = o->mergeVertex( Vec3f(min._x, max._y, max._z), n, Vec2f(1,1));
    d = o->mergeVertex( Vec3f(min._x, max._y, min._z), n, Vec2f(1,0));
    INSERT_QUAD(a,b,c,d);

    // Face X = max.
    a = o->mergeVertex( Vec3f(max._x, min._y, min._z), n, Vec2f(0,0));
    b = o->mergeVertex( Vec3f(max._x, min._y, max._z), n, Vec2f(0,1));
    c = o->mergeVertex( Vec3f(max._x, max._y, max._z), n, Vec2f(1,1));
    d = o->mergeVertex( Vec3f(max._x, max._y, min._z), n, Vec2f(1,0));
    INSERT_QUAD(d,c,b,a);

    // Face Y = min.
    a = o->mergeVertex( Vec3f(min._x, min._y, min._z), n, Vec2f(0,0));
    b = o->mergeVertex( Vec3f(max._x, min._y, min._z), n, Vec2f(0,1));
    c = o->mergeVertex( Vec3f(max._x, min._y, max._z), n, Vec2f(1,1));
    d = o->mergeVertex( Vec3f(min._x, min._y, max._z), n, Vec2f(1,0));
    INSERT_QUAD(a,b,c,d);

    // Face Y = max.
    a = o->mergeVertex( Vec3f(min._x, max._y, min._z), n, Vec2f(0,0));
    b = o->mergeVertex( Vec3f(max._x, max._y, min._z), n, Vec2f(0,1));
    c = o->mergeVertex( Vec3f(max._x, max._y, max._z), n, Vec2f(1,1));
    d = o->mergeVertex( Vec3f(min._x, max._y, max._z), n, Vec2f(1,0));
    INSERT_QUAD(d,c,b,a);

#undef INSERT_QUAD

    o->endEdit();
    objects.push_back(o);
    return o;
}

//////////////////////////////////////////////////////////////////////////////

static void reccursiveSphere(ScnObject* o, 
                             float radius, int level, 
                             const Vec3f& a, const Vec3f& b, const Vec3f& c)
{
    if( level > 0 )
    {
        // Not enough subdivision. Redo some :
        Vec3f ab = radius * normalize(a+b);
        Vec3f bc = radius * normalize(b+c);
        Vec3f ca = radius * normalize(c+a);
        reccursiveSphere(o, radius, level-1,  a, ab, ca);
        reccursiveSphere(o, radius, level-1, ca, ab, bc);
        reccursiveSphere(o, radius, level-1, bc, ab,  b);
        reccursiveSphere(o, radius, level-1,  c, ca, bc);
    }
    else
    {
        // We do have a triangle. Merge it with the actual object.
        // Note : position == normal, 'coz it is a unit sphere centered on 0.
        o->insertFaceVertex( o->mergeVertex(a, a) );
        o->insertFaceVertex( o->mergeVertex(b, b) );
        o->insertFaceVertex( o->mergeVertex(c, c) );
        o->insertFace(3);
    }
}

ScnObject*
Scene::makeSphere(float r, int level)
{
    ScnObject *o = new ScnObject;

    o->beginEdit();
    reccursiveSphere(o, r, level, Vec3f( r, 0, 0), Vec3f( 0, r, 0), Vec3f( 0, 0, r));
    reccursiveSphere(o, r, level, Vec3f( r, 0, 0), Vec3f( 0,-r, 0), Vec3f( 0, 0,-r));
    reccursiveSphere(o, r, level, Vec3f( r, 0, 0), Vec3f( 0, 0, r), Vec3f( 0,-r, 0));
    reccursiveSphere(o, r, level, Vec3f( r, 0, 0), Vec3f( 0, 0,-r), Vec3f( 0, r, 0));

    reccursiveSphere(o, r, level, Vec3f(-r, 0, 0), Vec3f( 0,-r, 0), Vec3f( 0, 0, r));
    reccursiveSphere(o, r, level, Vec3f(-r, 0, 0), Vec3f( 0, r, 0), Vec3f( 0, 0,-r));
    reccursiveSphere(o, r, level, Vec3f(-r, 0, 0), Vec3f( 0, 0, r), Vec3f( 0, r, 0));
    reccursiveSphere(o, r, level, Vec3f(-r, 0, 0), Vec3f( 0, 0,-r), Vec3f( 0,-r, 0));
    o->endEdit();

    objects.push_back(o);
    return o;
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::makeNormals(bool force)
{
    if( force || normals.count() != vertices.count() )
    {
        int countBefore = normals.count();

        // ---- Reset new normals --------------------------------------------
        normals.resize(vertices.count());

        // ---- Sum-up a normal-per-face onto each vertex --------------------
        for(int t=0; t<indices.count(); t+=3)
        {
            int i = indices[t  ];
            int j = indices[t+1];
            int k = indices[t+2];
            if( (i<countBefore) && (j<countBefore) && (k<countBefore) )
                continue;

            Vec3f N = cross( vertices[j] - vertices[i],
                             vertices[k] - vertices[i] );
            N.normalize();
            if( length2(N) > 0.0000001f )
            {
                if( i >= countBefore ) normals[i] += N;
                if( j >= countBefore ) normals[j] += N;
                if( k >= countBefore ) normals[k] += N;
            }
        }

        // ---- Normalize everybody ------------------------------------------
        for(int n=countBefore; n<normals.count(); ++n)
            normals[n].normalize();
    }

    _haveNormal = true;
}

void Scene::makeNormals(bool force)
{
    for(int o=0; o<objects.count(); ++o)
        objects[o]->makeNormals(force);
}

//////////////////////////////////////////////////////////////////////////////

void ScnObject::checkNormals()
{
    // Make sure we do have all normals in this object.
    if( normals.count() != vertices.count() )
        makeNormals(false);

    uint count = 0;

    for(int t=0; t<indices.count(); t+=3)
    {
        // ---- Compute triangle's normal --------------------------------
        int i = indices[t  ];
        int j = indices[t+1];
        int k = indices[t+2];

        Vec3f N = cross( vertices[j] - vertices[i],
                         vertices[k] - vertices[i] );
        N.normalize();

        // ---- Not coherent with vertices's normal ? --------------------
        int bad = 0;
        if( dot(N, normals[i]) < -0.000001 ) bad++;
        if( dot(N, normals[j]) < -0.000001 ) bad++;
        if( dot(N, normals[k]) < -0.000001 ) bad++;
        
        if( bad >= 2)
        {
            // Change vertex ordering by swapping vertex 1 and 2 :
            indices[t+1] = k;
            indices[t+2] = j;
            ++count;
        }
    }

    if( count > 0 )
    {
        qWarning("%d triangles flipped [%3.0f%%]",
            count,
            (100.0f*count)/float(indices.count() / 3));
    }
}

//////////////////////////////////////////////////////////////////////////////

int Scene::vertices() const
{
	int count = 0;
    for(int o=0; o<objects.count(); ++o)
		count += objects[o]->vertices.count();
	return count;
}

int Scene::triangles() const
{
	int count = 0;
    for(int o=0; o<objects.count(); ++o)
		count += objects[o]->indices.count() / 3;
	return count;
}
