////////////////////////////////////////////////////////////////////////
/*! \file   Scene.h
 *  \brief  A vertex array based scene format.
 *  \author    Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa Vec3f.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Scene.h,v 1.42 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2006/11/29] : Creation.

#include <QImage>
#include <QVector>

#include "Math/Vec2f.h"
#include "Math/Vec3f.h"
#include "Math/Vec4f.h"

class Matrix4;
class Shader;
class EnvironmentMap;
class GlossyMap;
class ShadowMap;
class Scene;

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
/// Lighting of the scene
///
class Light
{
public:
    Vec4f   position;
    Vec4f   ambient, diffuse, specular;
    Vec3f   attenuation;

    float   exponent;           //!< For spot light. 0.0 means uniform.
    float   cutoff;             //!< 0 to 90(deg). 180 means directional spot.

    Light();                    //!< Initialize default values.

    void enable(int l) const;   //!< Enable it, as OpenGL LIGHTl
    void disable(int l) const;  //!< Disable it, as OpenGL LIGHTl
};

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
/// Global render state for a scene.
class Environment
{
public:
    enum { MAX_LIGHTS=8 };

    //! Face rendering options.
    enum Wireframe {
        RENDER_FACES=1,
        RENDER_EDGES=2,
        RENDER_BOTH = RENDER_FACES + RENDER_EDGES
    };

    //! Face culling options.
    enum Culling {
        BACK_FACE=1,                    //!< Cull back faces, keep front faces.
        NO_CULLING=0,                   //!< No culling. Force double-face render.
        FRONT_FACE=-1                   //!< Cull front faces, keep back faces.
    };

    bool            dumpImages;         //!< Dump shadow maps, env maps, glossy maps.

    int             nbLights;           //!< How many lights to use ?
    Light           lights[MAX_LIGHTS]; //!< Lighting data, for each light.

    int             nbShadows;          //!< How many lights cast shadows ?
	bool			dynamicShadows;     //!< Should we recompute shadows before each frame ?
	bool			shadowsDirty;	    //!< Should we recompute shadows next frame ?
    ShadowMap*      shadows[MAX_LIGHTS];//!< Shadows data, if needed.

    int             primitiveFilter;    //!< If >=0, which triangle is displayed (0..12).
    Wireframe       wireframe;          //!< Draw triangle's outline instead of filling them.
	Culling         culling;            //!< 1 = cull back face. 0 = none. -1 = front face.
    bool            depthTest;          //!< 
    bool            blending;

    bool            colorMaps;          //!< Should we apply color maps ?
	bool			normalMaps;	        //!< Should we apply normal maps ?

    bool            envMaps;            //!< Should we apply env maps ?
    int             envMapMode;         //!< cubemap, parabola, sphere, lambert, etc...
    bool            dynamicEnvMap;      //!< Env maps should be recomputed at each frame.
    bool            envMapsDirty;       //!< Env maps should be recomputed at next frame.

    bool            glossyMaps;         //!< Should we render glossy maps ?
    bool            dynamicGlossyMaps;  //!< Render at each frame ?
    bool            glossyMapsDirty;    //!< Render next frame ?

    bool            twoPass;            //!< Two pass rendering...

    bool            testOverdraw;       //!< Tmp disable discards, to measure overdraw.
    bool            testShading;        //!< Tmp disable shading, to measure ray-triangle intersection time.
    bool            testNoFrags;        //!< Tmp disable fragments, to measure vertex+geometry time.
    bool            testNoGeomShader;   //!< Tmp skip geometry shader, to measure vertex time.

    //! Make sure everything is initialized.
    Environment();

    //! Make sure everything is cleaned.
    ~Environment();

    //! Pre-compute all needed maps.
    void prepare(const Shader* shader, const Scene* scene, int subMode);

    //! return pre-processor defines to compile shaders according to current state.
    QString envString() const;

    //! Enable all options for given shader (or OpenGL state if NULL)
    /*!
     * \p pass is the pass number, needed when several lights cast
     * shadows in OpenGL mode. Default (-1) means to render all lights
     * (if a \p shader is given), or all non-shadow casting lights
     * when in OpenGL mode (\p shader is NULL).
     *
     * Returns true if enabled lights do cast shadows.
     */
    bool enable(const Shader* shader, float zNear, float zFar, int pass=-1) const;

    //! Back to standard rendering state.
    void disable(const Shader*) const;
};

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
/* \brief A shared texture class.
 * Texture an object. Loading it just once in the OpenGL renderer.
 */
class Texture
{
public:

    //! Known texturing mode.
    enum Mode {
        COLOR_REPLACE = 1,          //!< Texture replaces fragment's color.
        BUMP_MAPPING,               //!< Grey level reprensents a fake detail relief.
        NORMAL_MAPPING              //!< RGB biased values represent normal in (tan, bitan, normal) local coordinates.
    };
    
    //! Get a pointer to a possibly shared texture.
    static const Texture* make(const QString& fileName, Mode mode);

    //! Setup OpenGL & eventual shader for texture rendering.
    void enable(const Shader* = 0) const;

    //! Cleanup texturing env after drawing.
    void disable(const Shader* = 0) const;

private:
    Texture(const QString& fileName, Mode mode);

    QImage              _image;
    Mode                _mode;
    mutable uint        _texId;         //!< Texture OpenGL ID.
};

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
/* \brief A Scene object is a set of triangles.
 * With per vertex, or per object material/color/normal.
 */
class ScnObject
{
    uint                _vbo;           //!< Vertex buffer used for all vertex data.
    uint                _ibo;           //!< Index buffer used for DrawRangeElements
    uint                _size;          //!< Size of the allocated VBO.
    Environment::Culling _culling;      //!< 1 : back face culling. 0 : no culling (double face), -1 : back face:

    bool                _loaded;        //!< Shoudl the VBO be reloaded ?
    bool                _haveColor;     //!< Do we have per-vertex color ?
    bool                _haveNormal;    //!< Do we have per-vertex normals ?
    bool                _haveTexture;   //!< Do we have per-vertex texture coordinates ?

    EnvironmentMap*     _envMap;        //!< Should this object reflect its environment ?
    GlossyMap*          _glossyMap;     //!< Should this object have a glossy reflection ?

    void reload();                      //!< Updates VBO/IBO from actual data.
    void bindVertices();                //!< Bind all associated vertex data.
    void unbindVertices();              //!< Back to unbind state.

    /// Quick hash function to accelerate findVertex().
    unsigned int hash(const Vec3f& V) const;
 
public:
    ScnObject();
    virtual ~ScnObject();

    ScnObject* copy() const;            //!< Return a copy of the current object.

    void transform(const Matrix4& xfm); //!< Apply a transorfmation to an object.

    //========================================================================
    // Scene data
    //

    Vec4f          ambient;             //!< Overall material ambient color.
    Vec4f          diffuse;             //!< Overall material diffuse color.
    Vec4f          specular;            //!< Overall material specular color.
    float          shininess;           //!< Overall material shininess.
    const Texture* colorTexture;        //!< Shared texture, if any.
    const Texture*  bumpTexture;        //!< Shared texture, if any.

    QVector<Vec3f> vertices;            //!< Shared vertices for al triangles.
    QVector<Vec4f> colors;              //!< Per vertex interpolated colors.
    QVector<Vec3f> normals;             //!< Per vertex interpolated normals.
    QVector<Vec2f> texcoord;            //!< Per vertex 2D texture coordinates.

    QVector<uint>  indices;             //!< triplets of position within the vertices array.

    //========================================================================
    // Scene drawing features

    //! \brief Draw an object, within its context
    void draw(const Shader* shader, const Environment* env) const; 

    //========================================================================
    // Scene definition features
    //
    void beginEdit();                   //!< Should be called before any vertex/face editing.
    void endEdit();                     //!< Should be called once all editing are done.

    //! Search for a given vertex.
    /// Valid only between a beginEdit()/endEdit() pair, because it uses
    /// temporary acceleration structures.
    int findVertex(const Vec3f& V, 
                   const Vec3f& N = *(Vec3f*)0, 
                   const Vec2f& T = *(Vec2f*)0,
                   const Vec4f& C = *(Vec4f*)0) const;

    //! Add a vertex to the vertex buffer already stored.
    /// All vertices are searched (with findVertex(), in time ~ constant)
    /// for a vertex with the exact same (V/N/T/C) values.
    /// NOTE: N should be normalized *before* calling this function.
    int mergeVertex(const Vec3f& V, 
                    const Vec3f& N = *(Vec3f*)0, 
                    const Vec2f& T = *(Vec2f*)0,
                    const Vec4f& C = *(Vec4f*)0);

    void insertFaceVertex(int idx);

    //! Add a face.
    /// Returns the position in the indices table of the first vertex.
    int insertFace(int len);

    /// \brief Set material's ambient color.
    void setAmbientColor(float r, float g, float b, float a = 1.0f);

    /// \brief Set material's diffuse color, if not defined per-vertex.
    void setDiffuseColor(float r, float g, float b, float a = 1.0f);

    /// \brief Set material's specular color, if not defined per-vertex.
    void setSpecularColor(float r, float g, float b, float a = 1.0f);

    /// \brief Set material's shininess.
    void setColorShininess(float s);

    /// \brief Set culling policy. Defaults to BACK_FACE culling.
    void setCulling(Environment::Culling face);

    //========================================================================
    // Texturing
    //
    /// \brief Set texture associated with texcoordinates.
    void setTexture(const QString& fileName, Texture::Mode mode = Texture::COLOR_REPLACE);

    //========================================================================
    // Misc
    //
public:

    //! Update \p min and \p max so they include this object.
    void enlargeBBox(Vec3f& min, Vec3f& max) const;

    //! Add normals to every vertex that don't have one.
    //! Sharing a vertex creates smooth normals. Duplicating shared vertex
    //! make flat-shaded normals.
    void makeNormals(bool force = false);

    //! Re-order triangles if clock-wise order is different from vertex normals.
    // Calls makeNormals() first, if there is missing normals.
    void checkNormals();
};

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
///
/// \brief A set of ScnObject's
///
class Scene
{
    /// \brief The basic pass for scene rendering.
    void drawPass(const Shader *) const;

public:
    Scene();
    virtual ~Scene();

    /// \brief Global rendering state.
    Environment env;

    /// \brief All scene elements.
    QVector<ScnObject*> objects;

    /// \brief Remove all ScnObject from the scene.
    void clear();

    /// \brief Draw the entire scene, using Scene::drawPass(),
    /// using multi-pass lighting when in no-shader mode.
    void draw(const Shader*, float zNear, float zFar) const;

    /// \brief Compute the BBox of the entire scene.
    void getBBox(Vec3f& min, Vec3f& max) const;

    /// \brief Append to the scene a axis-aligned box.
    ScnObject* makeBox(const Vec3f& min, const Vec3f& max);

    /// \brief Append to the scene a tesselated sphere.
    ScnObject* makeSphere(float radius = 1.0f, int level = 4);

    /// \brief Re-generates per-vertex normals for the whole scene, if needed.
    void makeNormals(bool force = false);

	/// \brief Count how many triangles are in the whole scene.
	int triangles() const;
	int vertices() const;
};

extern bool loadOBJ(Scene*, const QString&);
extern bool loadVRML(Scene*, const QString&);
