///////////////////////////////////////////////////////////////////////////////
/*! \file OBJPlugin.cpp
 *  \brief I/O of obj (wavefront) format.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * $Id: Scene_obj.cpp,v 1.23 2008/02/20 11:49:45 gascuel Exp $
 */
///////////////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFileInfo>
#include <QString>
#include <QProgressDialog>
#include <QHash>
#include <QDir>

#include "Scene.h"
#include "Math/Vec2f.h"

///////////////////////////////////////////////////////////////////////////////
/*! \brief Minimal material definition for OBJ import.
 * Retain just what Scene can handle right now.
 */
struct OBJMaterial
{
    Vec4f   _Ka;                //!< Ambient color
    Vec4f   _Kd;                //!< Diffuse color
    Vec4f   _Ks;                //!< Specular color
    float   _Ns;                //!< Shininess
    QString _colorTexture;      //!< absolute file name. Rel to .mtl file.
    QString _bumpTexture;       // TODO bump maps.

    inline OBJMaterial()
    : _Ka(0.2f, 0.2f, 0.2f, 1.0f),
      _Kd(0.8f, 0.8f, 0.8f, 1.0f),
      _Ks(0.2f, 0.2f, 0.2f, 1.0f),
      _Ns(16.0f)
    {}

};

static QHash<QString,OBJMaterial*> materials;

///////////////////////////////////////////////////////////////////////////////
//
/// Utility fonction to load Wavefront material files (.mtl)
static bool
loadMTL(const QString& fileName)
{
    FILE* f = fopen(fileName.toLocal8Bit(), "r");
    if( ! f )
    {
        qWarning("Unable to open material lib \"%s\"",
            fileName.toLatin1().constData());
        return false;
    }

    int lineno = 0;
    char line[256];
    char name[121];
    OBJMaterial *mtl = 0;
    float r, g, b, a;

    while( fgets(line, sizeof line, f) )
    {
        ++lineno;

        // Skip leading white spaces:
        char *p = line;
        while( (*p==' ') || (*p == '\n') || (*p == '\r') ||(*p == '\t') )
            ++p;

        if( sscanf(p, "newmtl %120s", name) == 1 )
        {
            mtl = new OBJMaterial;
            materials[ QString(name) ] = mtl;
        }
        else if( sscanf(p, "Ka %f %f %f", &r, &g, &b) == 3)
        {
            if( mtl )
                mtl->_Ka = Vec4f(r, g, b, 1.0f);
        }
        else if( sscanf(p, "Ks %f %f %f", &r, &g, &b) == 3)
        {
            if( mtl )
            {
                if( mtl->_Ns == 0.0f )  // Shininess alredy set to 0 => no specular reflections.
                    mtl->_Ks.reset();
                else
                    mtl->_Ks = Vec4f(r, g, b, 1.0f);
            }
        }
        else if( sscanf(p, "Kd %f %f %f", &r, &g, &b) == 3)
        {
            if( mtl )
                mtl->_Kd = Vec4f(r, g, b, 1.0f);
        }
        else if( sscanf(p, "map_Kd %120s", name) == 1)
        {
            mtl->_colorTexture = QFileInfo(fileName).dir().absoluteFilePath(name);
        }
        else if( sscanf(p, "bump %120s", name) == 1)
        {
            mtl->_bumpTexture = QFileInfo(fileName).dir().absoluteFilePath(name);
        }
        else if( sscanf(p, "d %f", &a) == 1)
        {
            if( mtl )
                mtl->_Ka._w = mtl->_Kd._w = mtl->_Ks._w = a;
        }
        else if( sscanf(p, "Ns %f", &r) == 1)
        {
            if( mtl )
            {
                if( r > 0.0f )      // Shininess null means no specular reflections....
                    mtl->_Ns = r;
                else
                    mtl->_Ks.reset();
            }
        }
        else if( sscanf(p, "Ni %f", &r) == 1)
            ;   // XXX what is this ?
        else if( sscanf(p, "illum %f", &r) == 1)
            ;   // XXX what is this ?
        else if( *p == '\0' || *p == '#' )
            ;   // Skip empty lines.
        else
            qWarning("Ignoring line \"%s\" in material library \"%s\" line %d",
                QString(p).trimmed().toLatin1().constData(),
                fileName.toLatin1().constData(), lineno);
    }
    (void) fclose(f);

    return true;
}

///////////////////////////////////////////////////////////////////////////////

typedef QVector<int> OBJSmoothGroup;
static QHash<int, OBJSmoothGroup *> smoothGroups;

OBJSmoothGroup* startGroup(int name)
{
    Q_ASSERT(name);

    if( smoothGroups.contains(name) )
        return smoothGroups[name];

    OBJSmoothGroup*r = new OBJSmoothGroup;
    smoothGroups.insert(name, r);
    return r;
}

void applySmoothGroups(ScnObject *object)
{
    for( QHash<int,OBJSmoothGroup*>::iterator it = smoothGroups.begin();
        it != smoothGroups.end();
        ++it)
    {
        OBJSmoothGroup* group = *it;

        // ---- Reset all normals of this smooth group -----------------------

        // Extend normal vector at once. Hence we have to walk the entire
        // smooth group to find the max vertex indice...
        int maxVertex = 0;
        for(int f=0; f < group->count(); ++f)
        {
            int t = group->at(f);
            int i = object->indices[t  ];
            int j = object->indices[t+1];
            int k = object->indices[t+2];

            j = (j>k) ? j : k;
            i = (i>j) ? i : j;
            maxVertex = (maxVertex > i) ? maxVertex : i;
        }
        maxVertex++;
        if( object->normals.count() < maxVertex )
            object->normals.resize(maxVertex);

        // ---- Reset all normals we are going to accumulate into ------------
        for(int f=0; f < group->count(); ++f)
        {
            int t = group->at(f);
            int i = object->indices[t  ];
            int j = object->indices[t+1];
            int k = object->indices[t+2];
            object->normals[i].reset();
            object->normals[j].reset();
            object->normals[k].reset();
        }

        // ---- Accumulate normals for this smooth group ---------------------
        for(int f=0; f < group->count(); ++f)
        {
            int t = group->at(f);
            int i = object->indices[t  ];
            int j = object->indices[t+1];
            int k = object->indices[t+2];

            Vec3f N = cross( object->vertices[j] - object->vertices[i],
                             object->vertices[k] - object->vertices[i] );
            N.normalize();

            object->normals[i] += N;
            object->normals[j] += N;
            object->normals[k] += N;
        }

        // ---- Normalize normals of this smooth group -----------------------
        for(int f=0; f < group->count(); ++f)
        {
            int t = group->at(f);
            int i = object->indices[t  ];
            int j = object->indices[t+1];
            int k = object->indices[t+2];
            object->normals[i].normalize();
            object->normals[j].normalize();
            object->normals[k].normalize();
        }

        // Release this smooth group
        delete group;
        it.value() = 0;
    }

    // Once applyed, just remove it.
    smoothGroups.clear();
}

///////////////////////////////////////////////////////////////////////////////

void startNextObject(Scene* scene, ScnObject* &object, const OBJMaterial* mtl)
{
    if( object && object->vertices.count() > 0 )
    {
        // Are we closing a previously smooth group ?
        applySmoothGroups(object);

        // Check that triangle ordering is coherent with normal :
        object->checkNormals();

        // Then add cleaned object to the global scene :
        object->endEdit();
        scene->objects.push_back(object);

        // Apply global material, if any :
        if( mtl )
        {
            object->setAmbientColor ( mtl->_Ka._x, mtl->_Ka._y, mtl->_Ka._z, mtl->_Ka._w);
            object->setDiffuseColor ( mtl->_Kd._x, mtl->_Kd._y, mtl->_Kd._z, mtl->_Kd._w);
            object->setSpecularColor( mtl->_Ks._x, mtl->_Ks._y, mtl->_Ks._z, mtl->_Ks._w);
            object->setColorShininess( mtl->_Ns);

            if( ! mtl->_colorTexture.isEmpty() )
                object->setTexture(mtl->_colorTexture, Texture::COLOR_REPLACE);

			if( ! mtl->_bumpTexture.isEmpty() )
                object->setTexture(mtl->_bumpTexture, Texture::BUMP_MAPPING);
        }

        qDebug( "Found %d vertices, %d triangles",
            object->vertices.count(),
            object->indices.count()/3);

        object = 0;
    }

    // And start-up a new one :
    if( ! object )
    {
        object = new ScnObject;
        Q_CHECK_PTR(object);
        object->beginEdit();
    }
}

///////////////////////////////////////////////////////////////////////////////

/*! Plugin for Wavefront file format.
 *  http://netghost.narod.ru/gff/graphics/summary/waveobj.htm
 *  http://netghost.narod.ru/gff/vendspec/waveobj/obj_spec.txt
 *
 * Note: colorPerVertex are not saved (not supported ?)
 */
bool
loadOBJ(Scene *scene, const QString& fileName)
{
    FILE* f = fopen(fileName.toLocal8Bit(), "r");
    if (!f)
    {
        qWarning("Unable to open OBJ file \"%s\"", 
            fileName.toLatin1().constData());
        return false;
    }

    uint fileLength = QFileInfo(fileName).size();
    uint lineno = 0;
    char line[256];

    // Show progress :
    QProgressDialog *progress = new QProgressDialog("Reading OBJ format...", "Cancel", 0, fileLength);
    progress->show();

    OBJSmoothGroup* smoothGroup = 0;
    Vec3f fakeNormal;                       // Fake normal, different for each smooth group.
    QVector<Vec3f> tmpVertex;
    QVector<Vec3f> tmpNormal;
    QVector<Vec2f> tmp2DTex;
    QVector<Vec2f> tmp2DTex1;

    // Default optional parameters for mergeVertex()
    const Vec2f& NullTex = *((Vec2f*)0);
    const Vec3f& NullNormal = *((Vec3f*)0);
    const OBJMaterial *mtl = 0;

    // Start a surface, with defaults :
    ScnObject *object = 0;
    startNextObject(scene, object, mtl);

    while( fgets(line, sizeof line, f) )
    {
        ++lineno;

        // Skip leading white spaces:
        char *p = line;
        while( (*p==' ') || (*p == '\n') || (*p == '\r') ||(*p == '\t') )
            ++p;

        float x=0.0f, y=0.0f, z=0.0f;
        if( p[0] == 'f' )
        {
            int len = 0;    // Vertex nb in this face.
            ++p;
            while( *p )
            {
                int i=-1, it=-1, in=-1, dp=0;

                if( sscanf(p, "%d/%d/%d%n", &i, &it, &in, &dp) == 3 )
                    ;
                else if( sscanf(p, "%d//%d%n", &i, &in, &dp) == 2 )
                    ;
                else if( sscanf(p, "%d/%d%n", &i, &it, &dp) == 2 )
                    ;
                else if( sscanf(p, "%d%n", &i, &dp) == 1 )
                    ;
                else
                {
                    qWarning( "Invalid face description.\nLine %d: %s",
                            lineno,
                            line);
                    goto badVertex;
                }

                if( i < 0 )
                {
                    qWarning( "File '%s' contains unsupported backward references",
                        fileName.section('/', -2).toLatin1().constData());
                    return false;
                }

                // go to next charactere:
                p += dp;
                while( (*p==' ') || (*p == '\n') || (*p == '\r') ||(*p == '\t') )
                    ++p;

                // Bewhare: i/in/it are 1-based !
                if( i > (int)tmpVertex.count() )
                {
                    qWarning( "Error loading OBJ surface: vertex index %d > count %d.\nLine %d",
                            i, tmpVertex.count(),
                            lineno);
                    goto badVertex;
                }
                if( it > (int)tmp2DTex.count() )
                {
                    qWarning( "Error loading OBJ surface: texture index %d > count %d.\nLine %d",
                            it, tmp2DTex.count(),
                            lineno);
                    goto badVertex;
                }
                if( in > (int)tmpNormal.count() )
                {
                    qWarning( "Error loading OBJ surface: normal index %d > count %d.\nLine %d",
                            in, tmpNormal.count(),
                            lineno);
                    goto badVertex;
                }

                const Vec3f* normal = & NullNormal;
                     if( in >= 0 )     normal = & tmpNormal[in-1];
                else if( smoothGroup ) normal = & fakeNormal;

                if( it >= 0 )   // Texture coord presend ?
                {
                    Vec2f t = it <= (int)tmp2DTex1.count() // use texture unit 1 if present.
                            ? tmp2DTex1[it-1]
                            : tmp2DTex[it-1];
                    object->insertFaceVertex( object->mergeVertex(tmpVertex[i-1], *normal, t) );
                }
                else
                    object->insertFaceVertex( object->mergeVertex(tmpVertex[i-1], *normal, NullTex));

                ++len;
                if( (len == 3) && *p )  // 3 vertices, and the face *DO* continue ?
                {
                    // Close this triangle.
                    int f = object->insertFace(len);
                    if( smoothGroup ) smoothGroup->push_back(f);

                    //  Restart a triangle with the same first two vertices.
                    // Suppose we have already count==3 vertex. To re-emit the first,
                    // we have to re-insert vertex count()-3 and count()-1
                    object->insertFaceVertex( object->indices[ f ] );
                    object->insertFaceVertex( object->indices[ f+2 ] );
                    len = 2;
                }
            }
badVertex:
            // Insert only complete triangles.
            if(len == 3)
            {
                int f = object->insertFace(len);
                if( smoothGroup )
                    smoothGroup->push_back(f);
            }

            // If not a smooth group, generate flat FACE normals, if needed
            if( !smoothGroup )
                object->makeNormals();
        }
        else if( *p == 'v' )
        {
            if( sscanf(p+2, "%f %f %f", &x, &y, &z) != 3 )
            {
                qWarning( "Error reading OBJ file %s line %d: bad vertex\n\t%s",
                        fileName.toLatin1().constData(), lineno, line);
                x = y = z = 0.0f;
            }

            switch( p[1] )
            {
            case ' ':
                tmpVertex.append( Vec3f(x,y,z) );
                break;
            case 't':
                tmp2DTex.append( Vec2f(x, 1.0f-y) );    // Reverse Y direction, because OpenGL images have UpperLeft origin.
                break;
            case 'n':
                tmpNormal.append( normalize(Vec3f(x,y,z)) );
                break;
            default:
                goto badLine;
            }
        }
        else if( *p == '#' )
        {
            if( sscanf(p, "#_#vt1 %f %f", &x, &y) == 2 )       // Texture unit 1 : Melody kludge ?!
                tmp2DTex1.append( Vec2f(x,y) );
        }
        else if( *p == 'm' )
        {
            if( strncmp(p, "mtllib ", 7) == 0 )
            {
                // ---- Reset material list to just the default one :
                materials.clear();
                materials[ "default" ] = new OBJMaterial;

                // ---- Make .mtl file names relative to .obj location
                QStringList libs = QString(p+7).split(QRegExp("[ \t]+"));
                QDir dir = QFileInfo(fileName).dir();

                // ---- load all the materials, in reverse order (first match on search)
                materials.clear();
                for(int m=libs.count()-1; m>=0; --m)
                    loadMTL( dir.filePath(libs[m].trimmed()) );
            }
            else
                goto badLine;
        }
        else if( *p == 'u' )
        {
            char usemtl[71] = {0};
            if( sscanf(p, "usemtl %70s", usemtl) == 1 )
            {
                OBJMaterial* newM = materials[ QString(usemtl) ];
                if( ! newM )
                {
                    newM = materials[ "default" ];
                    qWarning("Unknown material %s file %s line %d",
                        usemtl, fileName.toLatin1().constData(), lineno);
                }

                // Do we have a different material ?
                if( mtl != newM )
                {
                    // Apply old material to the previous object.
                    startNextObject(scene, object, mtl);

                    // There was a running smooth group. Which have been deleted. Hence
                    // restart a new one, with the same name (thanks to fake normal kludge !).
                    if( smoothGroup )
                        smoothGroup = startGroup( int(fakeNormal._x) );

                    // And take the new material for the new object.
                    mtl = newM;
                }
            }
            else
                goto badLine;
        }
        else if( *p == 'g')
        {
            // Restart a new object.
            startNextObject(scene, object, mtl);
            smoothGroup = 0;
        }
        else if( *p == 's')
        {
            int name = 0;
            if( sscanf(line, "s %d", &name) == 1 && name != 0)
            {
                smoothGroup = startGroup(name);
                // Create a fake vector, uniq but consistant for each smooth group.
                fakeNormal = Vec3f(name, 0, 2);
            }
            else
                smoothGroup = 0;
        }
        else if( *p == 'o' )
            ;                                                   // skip names.
        else if( *p == 0 )
            ;                                                   // empty line
        else
        {
badLine:
            qWarning( "Error reading OBJ file %s line %d:\n%s",
                    fileName.toLatin1().constData(), lineno, line);
        }

        // Every ~ 256 lines, update progress bar.
        if( (lineno & 1023) == 0 )
        {
            if( progress->wasCanceled() )
                break;
            progress->setLabelText( QString("Loading OBJ surface (%1 vertices, %2 tris)")
                .arg(object->vertices.count())
                .arg(object->indices.count()/3));
            progress->setValue(ftell(f));
            qApp->processEvents();
        }
    }
    fclose(f);
    delete progress;

    // Close last object currently defined.
    startNextObject(scene, object, mtl);
    object->endEdit();
    delete object;  // Extra-object.

    qWarning( "Loaded OBJ surface '%s'", fileName.toLatin1().constData() );
    qWarning( "Total %d vertices, %d triangles", scene->vertices(), scene->triangles());
    return true;
}
