%{
////////////////////////////////////////////////////////////////////////
/*! \file      Scene_vrml.y
 *  \brief     Grammar to parse VRML files.
 *  \author    Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Scene_vrml.y,v 1.15 2008/02/20 11:49:45 gascuel Exp $

#include <stdlib.h>

#include <QVector>
#include <QStack>
#include <QFileInfo>
#include <QImage>
#include <QProgressDialog>
#include <QHash>

#include "Math/Vec4f.h"
#include "Math/Vec3f.h"
#include "Math/Vec2f.h"
#include "Math/Matrix4.h"
#include "Math/Quaternion.h"

#include "Scene.h"

#define YYERROR_VERBOSE 1

extern void  Scene_vrmlerror(const char *msg);
extern int   Scene_vrmllex();
extern char* Scene_vrmltext;
extern int   Scene_vrmlleng;
extern void  Scene_vrmlrestart(FILE*);

extern int Scene_vrml_version;
extern int Scene_vrml_lineno;
QString Scene_vrml_file;
int Scene_vrml_file_size = 0;
bool Scene_vrml_ok;
QProgressDialog *progress = 0;

static Matrix4 transform;
static Vec3f transformTranslation, transformScale, transformCenter;
static Quaternion transformRotation, transformScaleOrientation;
static void composeTransforms();
static QVector<float> floats;
static QVector<int>   ints;
static QVector<Vec2f> texCoord;
static QVector<int>   texIndices;
static QVector<Vec3f> coordinates;
static QVector<int>   coordIndices;
static QVector<Vec3f> normals;
static QVector<int>   normalIndices;
static QStack<Matrix4> transformStack;
static bool ccw = true,
            convex = false,
            solid = false,
            normalPerVertex = false;
static Vec3f size;            

#define YYPARSE_PARAM scene

// Minimal context to associate geometry, appearance, material, and texture.
static struct ShapeContext
{
    ScnObject *object;
    float      ambient;
    Vec4f      diffuse;
    Vec4f      specular;
    float      shininess;
    QImage     image;
    QString    imageFileName;
} shapeContext;

// Minimal storage to allow later sharing of a part of geometry.
// Because transformation stack is already applyed when storing
// the object, we should retain it to move any further instance(s).
struct Instance
{
    const ScnObject* o;       // Object to share.
    const Matrix4    xfm;     // It's location, to move it elsewhere afterward.
    
    Instance(const ScnObject* o, const Matrix4& xfm)
    : o(o), xfm(xfm)
    {}
};

// table to retain objects.
typedef QHash<QString, const Instance *> ObjectMap;
static ObjectMap objectMap;

%}

%token VRML_HEADER

%token VRML_NAME
%token VRML_VALUE
%token VRML_STRING
%token VRML_TRUE
%token VRML_FALSE

%token VRML_DEF
%token VRML_USE
%token VRML_SEPARATOR
%token VRML_GROUP
%token VRML_TRANSFORM
    %token VRML_TRANSLATION
    %token VRML_ROTATION
    %token VRML_SCALE
    %token VRML_SCALEORIENTATION
    %token VRML_CENTER
    %token VRML_CHILDREN

%token VRML_SHAPE
    %token VRML_APPEARANCE
    %token VRML_GEOMETRY

%token VRML_MATERIAL
    %token VRML_AMBIENTINTENSITY
    %token VRML_DIFFUSECOLOR
    %token VRML_SPECULARCOLOR
    %token VRML_TRANSPARENCY
    %token VRML_SHININESS

%token VRML_COORDINATE
    %token VRML_POINT

%token VRML_NORMALBINDING
    %token VRML_VALUE
    %token VRML_PER_VERTEX_INDEXED

%token VRML_NORMAL
    %token VRML_VECTOR
%token VRML_TEXCOORD

%token VRML_TEXTURE
    %token VRML_IMAGE
    %token VRML_URL

%token VRML_IFS
    %token VRML_CCW
    %token VRML_SOLID
    %token VRML_CONVEX
    %token VRML_COORD
    %token VRML_COORDINDEX
    %token VRML_NORMALPERVERTEX
    %token VRML_NORMALINDEX
    %token VRML_TEXCOORDINDEX

%token VRML_BOX
    %token VRML_SIZE

%token VRML_SPHERE
    %token VRML_RADIUS

// Note: an union do not support types with constructors.
//       so QString, Vec3f or Matrix4 types are not possible here.
%union {
    bool b;
    float f;
    int i;
    char* s;
    float vec3[3];
//  float vec4[4];
}

%type <b> bool
%type <f> value
%type <i> int
%type <s> name def
%type <vec3> vec3
// %type <vec4> vec4

%start file

%%

///////////////////////////////////////////////////////////////////////////////
// Basic data types
//

bool: VRML_TRUE     { $$=true; }
    | VRML_FALSE    { $$=false; }
    ;

// Manage 1/3/many float vector.
value:  VRML_VALUE              { $$ = atof(Scene_vrmltext); };
vec3:   value value value       { $$[0] = $1; $$[1] = $2; $$[2] = $3; };
// vec4:   value value value value { $$[0] = $1; $$[1] = $2; $$[2] = $3; $$[3] = $4; };

values:
    |   values value    { floats.append($2); }
    ;

int:    VRML_VALUE      {   // Beware of hexadecimal interegs (in Texture images).
                            if( Scene_vrmltext[0]=='0' && Scene_vrmltext[1]=='x' )
                                $$ = QString(Scene_vrmltext+2).toInt(0, 16);
                            else
                                $$ = QString(Scene_vrmltext).toInt();
                        }
    ;
ints:
    |   ints int        { ints.append($2); }
    ;

name: VRML_NAME { $$ = _strdup(Scene_vrmltext); }
    ;


///////////////////////////////////////////////////////////////////////////////
// VRML file structure
//

file: VRML_HEADER nodes
    ;

///////////////////////////////////////////////////////////////////////////////
// Parsing nodes
//

nodes:
    |   nodes def node
    ;

def:
        {
            $$ = 0;
        }
    |   VRML_DEF name
        {
            $$ = $2;
        }
    ;

node:
        group_node
    |   shape_node
    |   texCoord_node
    |   coordinate_node
    |   hint_node
    |   normal_node
    |   texture_node
    |   geometry_node
    |   appearance_node
    |   material_node
    |   name '{' others '}'
        {
            qDebug( "%d: Skipping unsupported node type '%s'",
                    Scene_vrml_lineno, $1 );
            free($1);
        }
    ;

group_node:
        VRML_SEPARATOR
            {   // Push current transform in transform-stack
                transformStack.push(transform);
            }
        '{' nodes '}'
            {   // Pop current transform from stack
                transform = transformStack.pop();
            }
    |   VRML_GROUP '{' nodes '}'
            {}
    |   VRML_TRANSFORM 
            {   // Push current transform in transform-stack,
                // so composeTransforms() can get it.
                transformStack.push(transform);
            }
        '{' transform_attributes '}'
            {   // pop back current transform
                if( Scene_vrml_version > 1 )
                    transform = transformStack.pop();
                else
                    transformStack.pop();
            }
    ;

shape_node:
        VRML_SHAPE '{'
            {
                shapeContext.object = 0;
                shapeContext.ambient  = 0.2f;
                shapeContext.diffuse  = Vec4f(0.8f, 0.8f, 0.8f, 1.0f);
                shapeContext.specular  = Vec4f(0,0,0,1);
                shapeContext.shininess  = 0.0;
                shapeContext.image  = QImage();
                shapeContext.imageFileName = QString::null;
            }
        shape '}'
            {
                if( shapeContext.object )
                {
                    Vec4f ambient = shapeContext.ambient * shapeContext.diffuse;
                    shapeContext.object->setAmbientColor ( ambient._x, ambient._y, ambient._z, shapeContext.diffuse._w);
                    shapeContext.object->setDiffuseColor ( shapeContext.diffuse._x, shapeContext.diffuse._y, shapeContext.diffuse._z, shapeContext.diffuse._w);
                    shapeContext.object->setSpecularColor( shapeContext.specular._x, shapeContext.specular._y, shapeContext.specular._z, shapeContext.specular._w);
                    shapeContext.object->setColorShininess( shapeContext.shininess );
                    if( ! shapeContext.imageFileName.isEmpty() )
                        shapeContext.object->setTexture( shapeContext.imageFileName );
                        
                    shapeContext.object = 0;    // No one is allowed to color it afterward.
                }
            }
        ;

texCoord_node:
        VRML_TEXCOORD '{' VRML_POINT texCoord '}';

coordinate_node:
        VRML_COORDINATE '{' point '}'
            {   // Set current coord3 vector
                if( floats.count()%3 )
                    yyerror("Wrong number of values in 3D coordinates");
                else
                {
                    int n = floats.count()/3;
                    coordinates.resize( n );
                    for(int i=0; i<n; ++i)
                        coordinates[i] = Vec3f(floats[3*i], floats[3*i+1], floats[3*i+2]);
                }
                floats.clear();
            }
    ;

hint_node:
      VRML_NORMALBINDING '{' hints '}';

normal_node:
      VRML_NORMAL '{' VRML_VECTOR normal '}';

texture_node:
      VRML_TEXTURE '{' image '}';

geometry_node:
      VRML_BOX '{' geomParam '}'
            {
                ScnObject *o = ((Scene*)scene)->makeBox( -0.5f * size, 0.5f * size);
                o->transform( transform );
                shapeContext.object = o;
            }

    | VRML_SPHERE '{' geomParam '}'
            {
                ScnObject *o = ((Scene*)scene)->makeSphere();
                o->transform( transform );
                shapeContext.object = o;
            }

    | VRML_IFS '{' ifs '}'
            {   // Create faces from indices.
                ScnObject *o = new ScnObject;
                o->beginEdit();
                
                for(int i=0; i<coordIndices.count(); ++i)
                {
                    if( (i&127) == 0 )
                    {
                        progress->setLabelText(QString("Created %1 triangles")
                            .arg(o->indices.count()/3));
                    }

                    int len = 0;
                    while( (i<coordIndices.count()) && (coordIndices[i] != -1) )
                    {
                        int j = coordIndices[i];    // Index of the vertex
                        if( j > (int)coordinates.count() )
                            qWarning( "%d: Vertex index %d out of range [0..%d]",
                                      Scene_vrml_lineno,
                                      j, coordinates.count());
                        
                        int k=-1, l=-1;             // Indices for normals and texCoord.

                        // Get (optional) vertex normal:
                        if( normalPerVertex && normals.count() > 0 )
                            if( normalIndices.count() == coordIndices.count() )
                                k = normalIndices[i];
                            else
                                k = j;                            
                        if( k >= (int)normals.count() )
                            qWarning( "%d: Normal index %d out of range [0..%d])",
                                      Scene_vrml_lineno,
                                      k, normals.count());
                        
                        // Get (optional) vertex 2D texture coord:
                        if( texCoord.count() > 0 )
                            if( texIndices.count() == coordIndices.count() )
                                l = texIndices[i];
                            else
                                l = j;
                        if( l >= (int)texCoord.count() )
                            qWarning( "%d: Texture index %d out of range [0..%d]",
                                      Scene_vrml_lineno,
                                      l, texCoord.count());
                                
                        // Apply current transform to the vertex.
                        Vec3f v = transform.coordinatesOf(coordinates[j]);
 
                        if( k<0 )
                        {
                            if( l<0 )
                                o->insertFaceVertex( o->mergeVertex(v) );
                            else
                                o->insertFaceVertex( o->mergeVertex(v, *((Vec3f*)0), texCoord[l]) );
                        }
                        else
                        {
                            Vec3f n = transform.transformOf(normals[k]);

                            if( l<0 )
                                o->insertFaceVertex( o->mergeVertex(v, n) );
                            else
                                o->insertFaceVertex( o->mergeVertex(v, n, texCoord[l]) );
                        }
                        
                        ++len;
                        ++i;
                    }
                    o->insertFace(len);
                }
                o->endEdit();
                o->makeNormals();       // Generate all normals. This make a global smooth group for the object !
                o->checkNormals();
                o->setCulling( solid ? ccw ? Environment::BACK_FACE : Environment::FRONT_FACE : Environment::NO_CULLING );

                qWarning( "Found %d vertices, %d triangles",
                    o->vertices.count(),
                    o->indices.count()/3);
                
                if( Scene_vrml_version > 1 )
                {
                    coordinates.clear();
                    coordIndices.clear();
                    normals.clear();
                    normalIndices.clear();
                    texCoord.clear();
                    texIndices.clear();
                }
                
                shapeContext.object = o;    // Keep pointer to enable Shape to apply color onto it.
                ((Scene*)scene)->objects.push_back(o);
            }
    ;

appearance_node:
        VRML_APPEARANCE '{' appearance '}';

material_node:
        VRML_MATERIAL '{' material '}';

///////////////////////////////////////////////////////////////////////////////
// Skip everything else
// Tentative to have a grammar of any valid VRML1/VRML97 sub-tree.
//
others:
    |   others other
    |   others error        // Recover here on error.
    ;

other:
        name skiped_value
            {
                qDebug( "%d: Skipping unsupported '%s' attribute",
                        Scene_vrml_lineno, $1);
                free($1);
            }
    |   VRML_USE name
    |   anyKeyword skiped_value
    ;

skiped_value:
        bool                                        {}
    |   value values                                { floats.clear(); }
    |   '[' values ']'                              { floats.clear(); }
    |   node                                        {}
    |   VRML_STRING
    ;


anyKeyword:
        VRML_SCALE
    |   VRML_ROTATION
    |   VRML_TRANSLATION
    |   VRML_CENTER
    |   VRML_SCALEORIENTATION
    ;

///////////////////////////////////////////////////////////////////////////////
// Nodes attributes
//
transform_attributes:
        {
            transformCenter = transformTranslation = Vec3f(0,0,0);
            transformScale = Vec3f(1,1,1);
            transformScaleOrientation = transformRotation = Quaternion();
        }
    | transform_attributes VRML_TRANSLATION vec3
        {   
            transformTranslation = Vec3f($3[0], $3[1], $3[2]);
            composeTransforms();
        }
    | transform_attributes VRML_SCALE vec3
        {
            transformScale = Vec3f($3[0], $3[1], $3[2]);
            composeTransforms();
        }
    | transform_attributes VRML_ROTATION vec3 value
        {
            transformRotation = Quaternion( Vec3f($3[0], $3[1], $3[2]), $4);
            composeTransforms();
        }
    | transform_attributes VRML_SCALEORIENTATION vec3 value
        {
            transformScaleOrientation = Quaternion( Vec3f($3[0], $3[1], $3[2]), $4);
            composeTransforms();
        }
    | transform_attributes VRML_CENTER vec3
        {
            transformCenter = Vec3f($3[0], $3[1], $3[2]);
            composeTransforms();
        }
    | transform_attributes VRML_CHILDREN '[' nodes ']'
    | transform_attributes error    // Recover here on error.
    ;


shape:
    | shape VRML_APPEARANCE def appearance_node
    | shape VRML_GEOMETRY def geometry_node
        {
            if( $3 )
                objectMap.insert($3, new Instance(shapeContext.object, transform) );
        }
    | shape VRML_GEOMETRY VRML_USE name
        {
            if( const Instance *i = objectMap[$4] )
            {
                // Compose transform to place the instance at its new position :
                Matrix4 xfm = i->xfm.inverse() * transform;
                shapeContext.object = i->o->copy();     // Keep source unchanged.
                shapeContext.object->transform(xfm);    // Move geometry.
                // Note: appearance are applyed later, so that's all for now.
            }
            else
                qDebug("%d: Skipping unknown reference to '%s'",
                    Scene_vrml_lineno, $4);
        }
    | shape VRML_TEXTURE def texture_node
    | shape error    // Recover here on error.
    ;

point:
      VRML_POINT '[' values ']'
    | VRML_POINT '[' error ']'    // Recover here on error.
    ;

geomParam:
        {   size = Vec3f(1,1,1); }
    | geomParam VRML_SIZE value value value       // Only for Box node.
        {   size = Vec3f($3, $4, $5); }
    | geomParam VRML_RADIUS value
        {   size = Vec3f($3, $3, $3); }             // Only for Sphere node.
    ;

ifs:
        {   // Initialise state.
            if( Scene_vrml_version > 1 )
            {
                coordinates.clear();
                coordIndices.clear();
                normals.clear();
                normalIndices.clear();
                texCoord.clear();
                texIndices.clear();
                ccw = true;
                convex = false;
                solid = false;
                normalPerVertex = false;
            }
        }
    | ifs VRML_CCW bool
        {   ccw = $3; }
    | ifs VRML_CONVEX bool
        {   convex = $3; }
    | ifs VRML_SOLID bool
        {   solid = $3; }
    | ifs VRML_NORMALPERVERTEX bool
        {   normalPerVertex = $3; }
    | ifs VRML_COORD def coordinate_node
        {}
    | ifs VRML_COORDINDEX '[' ints ']'
        {   coordIndices = ints;
            ints.clear();
        }
    | ifs VRML_NORMAL def normal_node
        {}
    | ifs VRML_NORMAL normal
        {}
    | ifs VRML_NORMALINDEX '[' ints ']'
        {   // Set current normal indices
            normalIndices = ints;
            ints.clear();
        }
    | ifs VRML_TEXCOORD def texCoord_node
        {}
    | ifs VRML_TEXCOORD texCoord
    | ifs VRML_TEXCOORDINDEX '[' ints ']'
        {   // Set current texture indices
            texIndices = ints;
            ints.clear();
        }
    | ifs other
    | ifs error    // Recover here on error.
    ;

hints:
    | hints VRML_VALUE VRML_PER_VERTEX_INDEXED
        { normalPerVertex = true; }
    | hints error    // Recover here on error.
    ;

normal:
    '[' values ']'
        {   // Set current normals
            if( floats.count()%3 )
                yyerror("Wrong number of values in normales");
            else
            {
                int n = floats.count()/3;
                normals.resize( n );
                for(int i=0; i<n; ++i)
                    normals[i] = Vec3f(floats[3*i], floats[3*i+1], floats[3*i+2]);
            }
            floats.clear();
        }
    |   '[' error ']'    // Recover here on error.
    ;

texCoord:
    '[' values ']'
        {   // Set current texCoord2 vector
                if( floats.count() %2 )
                    yyerror("Wrong number of points in 2D texture coordinates");
                else
                {
                    int n = floats.count()/2;
                    texCoord.resize( n );
                    for(int i=0; i<n; ++i)
                    {
                        // Revert Y-axis on texture coord, so image order matches OpenGL order.
                        texCoord[i] = Vec2f(floats[2*i], 1.0f - floats[2*i+1]);
                    }
                }
                floats.clear();
            }
    |   '[' error ']'    // Recover here on error.
    ;

image:
    | image VRML_IMAGE vec3 ints
        {
            int w = int($3[0]), h = int($3[1]), n = ints.count();
            if( n != w*h )
                yyerror("Wrong number of pixels in image.");
            shapeContext.image = QImage(w, h, ($3[2]==1) ? QImage::Format_RGB32  :
                                              ($3[2]==2) ? QImage::Format_ARGB32 :
                                              ($3[2]==3) ? QImage::Format_RGB32  :
                                                           QImage::Format_ARGB32 );
            int i=0;
            for(int y=0; y<h; ++y)
                for(int x=0; x<w; ++x)
                {
                    uint p = ints[i++];
                    switch( int($3[2]) )
                    {
                    case 1:             // Gray as 0xgg
                        shapeContext.image.setPixel(x,y, qRgb(255&p, 255&p, 255&p) );
                        break;
                    case 2:             // Gray+Alpha as 0xggAA
                        shapeContext.image.setPixel(x,y, qRgba(255&(p>>8), 255&(p>>8), 255&(p>>8), 255&p) );
                        break;
                    case 3:             // RGB as 0xRRGGBB
                        shapeContext.image.setPixel(x,y, qRgb((p>>16), 255&(p>>8), 255&p) );
                        break;
                    default:            // RGB as 0xRRGGBBAA
                        shapeContext.image.setPixel(x,y, qRgba((p>>24), 255&(p>>16), 255&(p>>8), 255&p) );
                        break;
                    }
                }
            ints.clear();
        }
    | image VRML_URL VRML_STRING
        {
            QString fileName = QString::fromUtf8(Scene_vrmltext+1, Scene_vrmlleng-2);
            shapeContext.imageFileName = QDir( Scene_vrml_file.section('/', 0, -2))
                                        .absoluteFilePath(fileName);
        }
    | image other
    | image error    // Recover here on error.
    ;

appearance:
    | appearance VRML_MATERIAL def material_node
    | appearance VRML_TEXTURE def texture_node
    | appearance other
    | appearance error    // Recover here on error.
    ;

material:
    | material VRML_AMBIENTINTENSITY value
        {
            shapeContext.ambient=$3
        }
    | material VRML_DIFFUSECOLOR vec3
        {
            shapeContext.diffuse._x = $3[0];
            shapeContext.diffuse._y = $3[1];
            shapeContext.diffuse._z = $3[2];
        }
    | material VRML_SPECULARCOLOR vec3
        {
            shapeContext.specular._x = $3[0];
            shapeContext.specular._y = $3[1];
            shapeContext.specular._z = $3[2];
        }
    | material VRML_TRANSPARENCY value
        {
            shapeContext.diffuse._w = shapeContext.specular._w = (1.0f-$3);
        }
    | material VRML_SHININESS value
        {
            shapeContext.shininess = $3;
        }
    | material other
    | material error    // Recover here on error.
    ;

%%

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

// Beware: The order of operations is independant of the file order.
// Hence, we have to store everything, to compute from scratch.
static void
composeTransforms()
{
    Matrix4& tos = transformStack.top();

    // From the VRML97 spec (AnnotatedVrmlRef/ch3-352.htm):
    // P_parent = T  C  R  SR  S  -SR  -C  P_child 
    transform = tos
              * Matrix4::translate(transformTranslation)
              * Matrix4::translate(transformCenter)
              * Matrix4::rotate(transformRotation)
              * Matrix4::rotate(transformScaleOrientation)
              * Matrix4::scale(transformScale)
              * Matrix4::rotate(transformScaleOrientation.inverse())
              * Matrix4::translate(-transformCenter)
              ;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void 
Scene_vrmlerror(const char *msg)
{
    qWarning( "While importing VRML object:\n%s(%d): %s\nnear '%s ...'",
              Scene_vrml_file.toLatin1().constData(),
              Scene_vrml_lineno,
              msg, Scene_vrmltext);
    Scene_vrml_ok = false;
}

bool
loadVRML(Scene *scene, const QString& fileName)
{
    FILE *f = fopen(fileName.toLocal8Bit(), "r");
    if( ! f )
    {
        qWarning( "Cannot open input file '%s'", fileName.toLatin1().data() );
        return false;
    }

    Scene_vrml_lineno = 1;
    Scene_vrml_file = fileName;
    Scene_vrml_file_size = QFileInfo(fileName).size();
    Scene_vrml_ok = true;
    Scene_vrmlrestart(f);
    progress = new QProgressDialog("Loading VRML object...", "Cancel", 0, Scene_vrml_file_size);
    
    yyparse(scene);
    
    delete progress; progress = 0;
    fclose(f);
    texCoord.clear();
    texIndices.clear();
    coordinates.clear();
    coordIndices.clear();
    normals.clear();
    normalIndices.clear();
    objectMap.clear();

    qWarning( "Loaded VRML surface '%s'", fileName.toLatin1().constData() );
    qWarning( "Total %d triangles", scene->triangles());

    return Scene_vrml_ok;
}
