////////////////////////////////////////////////////////////////////////
/*! \file   Shader.cpp
 *  \brief  Simple class to run shaders.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Shader.cpp,v 1.37 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2006/12/03] : Creation.

#include "Shader.h"

#include <QTextStream>

#include <GL/glew.h>
#include <QtOpenGL>

#include "Math/Matrix4.h"
#include "Utilities.h"

//////////////////////////////////////////////////////////////////////////////

Shader::~Shader()
{}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

ARBShader::ARBShader(const char *vertexFile,
                     const char *fragmentFile,
                     const char *environment)
: _env(environment ? _strdup(environment) : 0),
  _vertexFile(_strdup(vertexFile)),
  _fragmentFile(_strdup(fragmentFile)),
  _vp(0),
  _fp(0)
{
	if( ! GLEW_ARB_vertex_program && _vertexFile )
        qWarning("*** OpenGL extension ARB_vertex_program missing.");

	if( ! GLEW_ARB_fragment_program && _fragmentFile )
        qWarning("*** OpenGL extension ARB_fragment_program missing.");
}

//////////////////////////////////////////////////////////////////////////////

ARBShader::~ARBShader()
{
    disable();
    if( _vp )
	{
        glDeleteProgramsARB(1, &_vp);
		free( (void*) _vertexFile );
	}

    if( _fp )
	{
        glDeleteProgramsARB(1, &_fp);
		free( (void*)_fragmentFile );
	}

	if( _env )
		free( (void*)_env );
}

//////////////////////////////////////////////////////////////////////////////

bool 
ARBShader::checkLog()
{
    int pos = 0;
    glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &pos);
    if( pos >= 0 )
    {
        const char* list = (char *)glGetString(GL_PROGRAM_ERROR_STRING_ARB);

        if( pos )
        {
            qWarning("*** Error at byte %d:\t%s\n",
                pos,
                list);
            return false;
        }
    }
    CHECK_GL_ERROR("check log");

    return true;
}

//////////////////////////////////////////////////////////////////////////////

static const char LINE[] = "#line 1\n\r";

bool
ARBShader::loadNCompile(int type,
                        unsigned int program,
                        const char *fileName,
                        const char *environment)
{
    // ---- Read shader source -----------------------------------------------
    QFile in(fileName);
    if( !in.open(QIODevice::ReadOnly) )
    {
        qWarning("Unable to find shader source %s",
            fileName);
        return false;
    }
    qint64 len = in.size();
    qint64 envLen = environment ? strlen(environment)+strlen(LINE) : 0;
    char *shaderSource = new char[envLen+len+1];
    Q_CHECK_PTR(shaderSource);

    if(environment) 
    {
        strcpy(shaderSource, environment);
        strcat(shaderSource, LINE);
    }
    if( in.read(shaderSource+envLen, len) != len )
    {
        qWarning("Error reading shader file %s",
            fileName);
        in.close();
        delete shaderSource;
        return false;
    }
    in.close();
    shaderSource[envLen+len] = '\0';       // Close string, just for debug purpose.

    // ---- Load and compile source into shader object -----------------------
    glBindProgramARB(type, program);
    glProgramStringARB(type, GL_PROGRAM_FORMAT_ASCII_ARB, envLen+len, shaderSource);

    if( ! checkLog() )
    {
        qWarning("*** Compile shader %s failed.",
               fileName);
        return false;
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

bool ARBShader::compile()
{
    if( _vertexFile )
    {
        if( ! GLEW_ARB_vertex_program )
            qWarning("*** OpenGL extension ARB_vertex_program missing.");
        else
            qDebug("OpenGL GL_ARB_vertex_program ok");

        glGenProgramsARB(1, &_vp);
        if( !ARBShader::loadNCompile(GL_VERTEX_PROGRAM_ARB, _vp, _vertexFile, _env) )
            return false;

        qDebug("Vertex program ARB %s loaded.", _vertexFile);
    }

    if( _fragmentFile )
    {
        if( ! GLEW_ARB_fragment_program )
            qWarning("*** OpenGL extension ARB_fragment_program missing.");
        else
            qDebug("OpenGL GL_ARB_fragment_program ok");

        glGenProgramsARB(1, &_fp);
        if( !ARBShader::loadNCompile(GL_FRAGMENT_PROGRAM_ARB, _fp, _fragmentFile, _env) )
            return false;

        qDebug("Vertex fragment ARB %s loaded.", _fragmentFile);
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

static int getGLLocation(GLint program, const char *name, GLint type)
{
	int loc = glGetUniformLocationARB(program, name);
#ifdef DEBUG
	if( loc >= 0 )
	{
		char buffer[120]; int nameLen = 0; int uniformSize = 0; GLenum uniformType = 0;
		glGetActiveUniform(program, loc, sizeof buffer, &nameLen, &uniformSize, &uniformType, buffer);
        // SAMPLER_2D_ARRAY returns 0 type... So don't cry in that case...
		if( uniformType != type )
			qWarning("*** Bad uniform type 0x%X should be 0x%X", uniformType, type);
	}
#endif
	return loc;
}

bool ARBShader::setUniform(const char*name, float value) const
{
    int loc = getGLLocation(prog(), name, GL_FLOAT);
    if( loc >= 0 )
    {
        glUniform1fARB(loc, (GLfloat)value);
        CHECK_GL_ERROR(name);
        return true;
    }
    return false;
}

bool ARBShader::setUniform(const char*name, int value, int type) const
{
	int loc = getGLLocation(prog(), name, (type < 0) ? GL_INT : type);
    if( loc >= 0 )
    {
        if( type == GL_UNSIGNED_INT )
        {
            glUniform1uiEXT(loc, value);
            CHECK_GL_ERROR(name);
        }
        else
        {
            glUniform1i(loc, value);
            CHECK_GL_ERROR(name);
        }
        return true;
    }
    return false;
}

bool ARBShader::setUniform(const char* name, const void* value, int type, unsigned int count) const
{
    Q_ASSERT( count > 0 );
    Q_ASSERT( type > 0 );

	int loc = getGLLocation(prog(), name, type);
    if( loc >= 0 )
    {
        switch( type )
        {
        case GL_FLOAT_VEC2: glUniform2fv(loc, count, (GLfloat*)value); break;
        case GL_FLOAT_VEC3: glUniform3fv(loc, count, (GLfloat*)value); break;
        case GL_FLOAT_VEC4: glUniform4fv(loc, count, (GLfloat*)value); break;

        case GL_FLOAT_MAT2: glUniformMatrix2fv(loc, count, false, (GLfloat*)value); break;
        case GL_FLOAT_MAT3: glUniformMatrix3fv(loc, count, false, (GLfloat*)value); break;
        case GL_FLOAT_MAT4: glUniformMatrix4fv(loc, count, false, (GLfloat*)value); break;

        case GL_BOOL_VEC2:
        case GL_INT_VEC2: glUniform2iv(loc, count, (GLint*)value); break;
        case GL_BOOL_VEC3:
        case GL_INT_VEC3: glUniform3iv(loc, count, (GLint*)value); break;
        case GL_BOOL_VEC4:
        case GL_INT_VEC4: glUniform4iv(loc, count, (GLint*)value); break;

        case GL_UNSIGNED_INT_VEC2_EXT: glUniform2uivEXT(loc, count, (GLuint*)value); break;
        case GL_UNSIGNED_INT_VEC3_EXT: glUniform3uivEXT(loc, count, (GLuint*)value); break;
        case GL_UNSIGNED_INT_VEC4_EXT: glUniform4uivEXT(loc, count, (GLuint*)value); break;

        // XXX not yet in glew.h ...
        case 0x8B69:        glUniformMatrix4x2fv(loc, count, false, (GLfloat*)value); break;

        default:
            qDebug("Unknown float type 0x%04x, for %s", type, name);
            return false;
        }
        CHECK_GL_ERROR(name);
        return true;
    }

    return false;
}

//////////////////////////////////////////////////////////////////////////////

int ARBShader::prog() const
{
    return _vp;		// Arg. Must choose between vp and fp here...
}

//////////////////////////////////////////////////////////////////////////////

void ARBShader::enable() const
{
    if( !_vp && !_fp )
    {
        if( ! const_cast<ARBShader*>(this)->compile() )
            return;
    }

    if( _vp )
    {
        glBindProgramARB(GL_VERTEX_PROGRAM_ARB, _vp);
        glEnable(GL_VERTEX_PROGRAM_ARB);
    }
    if( _fp )
    {
        glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, _fp);
        glEnable(GL_FRAGMENT_PROGRAM_ARB);
    }

    checkLog();
}

void ARBShader::disable() const
{
    if( _vp )
    {
        glDisable(GL_VERTEX_PROGRAM_ARB);
        glBindProgramARB(GL_VERTEX_PROGRAM_ARB, 0);
    }
    if( _fp )
    {
        glDisable(GL_FRAGMENT_PROGRAM_ARB);
        glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, 0);
    }

    checkLog();
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

GLShader::GLShader(const char *vertexFile,
                   int inType, const char *geometryFile, int outType,
                   const char *fragmentFile,
                   const char *environment)
: ARBShader(vertexFile, fragmentFile, environment),
  _geometryFile( _strdup(geometryFile) ),
  _inType(inType), _outType(outType), 
  _nbOut(3),                                // Defaults to a single triangle.
  _gs(0),
  _program(0)
{
    Q_ASSERT(inType);
    Q_ASSERT(outType);

    if( ! GLEW_ARB_vertex_shader && _vertexFile )
        qWarning("*** OpenGL extension ARB_vertex_shader missing.");

    if( ! GLEW_EXT_geometry_shader4 && _geometryFile )
        qWarning("*** OpenGL extension EXT_geometry_shader4 missing.");

    if( ! GLEW_ARB_fragment_shader && _fragmentFile )
        qWarning("*** OpenGL extension ARB_fragment_shader missing.");
}

GLShader::GLShader(const char *vertexFile,
                   const char *fragmentFile,
                   const char *environment)
: ARBShader(vertexFile, fragmentFile, environment),
  _geometryFile(0),
  _inType(0), _outType(0),
  _nbOut(0),
  _gs(0),
  _program(0)
{
    if( ! GLEW_ARB_vertex_shader && _vertexFile )
        qWarning("*** OpenGL extension ARB_vertex_shader missing.");

    if( ! GLEW_ARB_fragment_shader && _fragmentFile )
        qWarning("*** OpenGL extension ARB_fragment_shader missing.");
}

//////////////////////////////////////////////////////////////////////////////

GLShader::~GLShader()
{
    if( _geometryFile ) free( (void *)_geometryFile );

    if( _program )
    {
        if( _vp )
        {
            glDetachShader(_program, _vp);
            glDeleteShader(_vp);
			_vp = 0;
        }
        if( _gs )
        {
            glDetachShader(_program, _gs);
            glDeleteShader(_gs);
        }
        if( _fp )
        {
            glDetachShader(_program, _fp);
            glDeleteShader(_fp);
			_fp = 0;
        }
        glDeleteProgram(_program);
        CHECK_GL_ERROR(0);
    }
}

//////////////////////////////////////////////////////////////////////////////

bool GLShader::checkLog(const char *fileName, unsigned int program, unsigned int shader, int type)
{
    int rc = 0;
    int infologLength = 0;
    if( shader )
    {
        glGetShaderiv(shader, type, &rc);
	    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infologLength);
        CHECK_GL_ERROR("shader log");

        if( (infologLength > 0)  && (rc == GL_FALSE) )
	    {
	        int charsWritten  = 0;
	        char *infoLog;

	        infoLog = new char[infologLength];
	        glGetShaderInfoLog(shader, infologLength, &charsWritten, infoLog);
            CHECK_GL_ERROR("shader log");
            qWarning("*** GL shader %s info log:", fileName);
		    qWarning(infoLog);
	        delete infoLog;

            return false;
	    }
    }
    if( program )
    {
        glGetProgramiv(program, type, &rc);
	    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infologLength);
        CHECK_GL_ERROR("program log");

        if( (infologLength > 0) && (rc == GL_FALSE) )
	    {
	        int charsWritten  = 0;
	        char *infoLog;

	        infoLog = new char[infologLength];
	        glGetProgramInfoLog(program, infologLength, &charsWritten, infoLog);
            CHECK_GL_ERROR("program log");
            qWarning("*** GL program %s info log:", fileName);
		    qWarning(infoLog);
	        delete infoLog;

            return false;
	    }
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

bool GLShader::loadNCompile(int type,
                          unsigned int &shader,
                          const char *fileName,
                          const char *environment)
{
    // ---- Read shader source -----------------------------------------------
    QFile in(fileName);
    if( !in.open(QIODevice::ReadOnly) )
    {
        qWarning("Unable to find shader source %s",
            fileName);
        return false;
    }
    qint64 len = in.size();

    char *shaderSource = new char[len+1];
    Q_CHECK_PTR(shaderSource);

    if( in.read(shaderSource, len) != len )
    {
        qWarning("Error reading shader file %s",
            fileName);
        in.close();
        delete shaderSource;
        return false;
    }
    in.close();
    shaderSource[len] = '\0';       // Close string, just for debug purpose.

    // ---- Create shader object ---------------------------------------------
    shader = glCreateShader(type);
    if( ! shader )
    {
        qWarning("*** Unable to create shader object for %s",
                 fileName);
        return false;
    }
    CHECK_GL_ERROR("create shader");

    // ---- Load and compile source into shader object -----------------------
    const char *lines[3] = { environment, LINE, shaderSource };
    const int   lens [3] = { environment ? (int)strlen(environment) : 0,
                            (int)strlen(LINE),
                            len };
    glShaderSource(shader, 3, lines, lens);
    glCompileShader(shader);
    delete[] shaderSource;
    CHECK_GL_ERROR("compile shader");

    if( ! GLShader::checkLog(fileName, 0, shader, GL_COMPILE_STATUS) )
    {
        qWarning("*** Compile shader %s failed.",
               fileName);
        return false;
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

bool GLShader::compile()
{
    _program = glCreateProgram();
    CHECK_GL_ERROR("create program");

    if( _vertexFile )
    {
        if( !GLShader::loadNCompile(GL_VERTEX_SHADER, _vp, _vertexFile, _env) )
            return false;

        qDebug("GLSL vertex shader %s loaded.", _vertexFile);
    }

    if( _fragmentFile )
    {
        if( !GLShader::loadNCompile(GL_FRAGMENT_SHADER, _fp, _fragmentFile, _env) )
            return false;

        qDebug("GLSL fragment shader %s loaded.", _fragmentFile);
    }

#ifdef GL_GEOMETRY_SHADER_EXT
    if( _geometryFile )
    {
        if( !GLShader::loadNCompile(GL_GEOMETRY_SHADER_EXT, _gs, _geometryFile, _env) )
            return false;

        qDebug("GLSL geometry shader %s loaded.", _geometryFile);
    }
#else
    Q_ASSERT( _geometryFile = NULL );
#endif

    if( _vp ) glAttachShader(_program, _vp);
	if( _fp ) glAttachShader(_program, _fp);
    CHECK_GL_ERROR("attach shaders");

#ifdef GL_GEOMETRY_SHADER_EXT
    if( _gs )
    {
        glAttachShader(_program, _gs);
        CHECK_GL_ERROR("attach geom shader");

        glProgramParameteriEXT(_program, GL_GEOMETRY_INPUT_TYPE_EXT,   _inType);
        glProgramParameteriEXT(_program, GL_GEOMETRY_VERTICES_OUT_EXT, _nbOut);
        glProgramParameteriEXT(_program, GL_GEOMETRY_OUTPUT_TYPE_EXT,  _outType);
        CHECK_GL_ERROR("param geom shader");
    }
#endif

	glLinkProgram(_program);
    if( ! GLShader::checkLog("LINK", _program, 0, GL_LINK_STATUS) )
    {
        qWarning("*** Link GLSL program failed.");
        return false;
    }

	glValidateProgram(_program);
    if( ! GLShader::checkLog("validate", _program, 0, GL_VALIDATE_STATUS) )
    {
        qWarning("*** Unable to validate GLSL shader program");
        return false;
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

int GLShader::prog() const
{
    return _program;
}

void GLShader::enable() const
{
    if( _program == 0 )
    {
        if( !const_cast<GLShader*>(this)->compile() )
            return;
    }

    glUseProgram(_program);
    CHECK_GL_ERROR("enable program");
}

void GLShader::disable() const
{
    glUseProgram(0);
    CHECK_GL_ERROR("disable program");
}

