////////////////////////////////////////////////////////////////////////
/*! \file   Shader.H
 *  \brief  Simple class to run shaders.
 *  \author    Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 */
////////////////////////////////////////////////////////////////////////
// $Id: Shader.h,v 1.19 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2006/12/03] : Creation.

class Matrix4;

////////////////////////////////////////////////////////////////////////
/// \brief Yet another base class to interface to shader objects.
///
/// This class provides loading, compiling, and setup of 
/// shaders programm.
/// Because there is subtile variations for Cg, GLSH, or ARB assembleurs,
/// there is derivated classes for each type.
///
/// \sa CgShader, ARBShader, GLShader, DXShader.
class Shader
{
public:
    virtual ~Shader();

    /// \brief Setup maximum number of vertices output by an optional geometry shader.
    virtual void setMaxOutput(int) = 0;

    /// \brief Load and compile shaders.
    /// Returns true if everything is ok. Fire qWarning() if it get
    /// trouble compiling.
    virtual bool compile() = 0;

    /// \brief Set a uniform float variable.
    /// Should be called after a successfull compile, at anytime before
    /// an enable() or after a disable().
    /// Returns true if evrything is ok, false if not.
    virtual bool setUniform(const char*name, float value) const = 0;

    /// \brief Set a uniform int (or derived) variable.
    /// Should be called after a successfull compile, at anytime before
    /// an enable() or after a disable().
    /// Returns true if evrything is ok, false if not.
    ///
    /// Use with type=GL_BOOL to set bool variables, type=GL_SAMPLER_2D to
    /// set texture units, etc. etc.
    virtual bool setUniform(const char*name, int value, int type = -1) const = 0;

    /// \brief Set a conpound float variable.
    /// Should be called after a successfull compile, at anytime before
    /// an enable() or after a disable().
    /// Returns true if evrything is ok, false if not.
    ///
    /// Use with type=GL_FLOAT_VEC3 to set vec3 variable, type-GL_FLOAT_MAT4 to
    /// set mat4 variable, etc. etc.
    ///
    /// Supported basic types are :
    ///     GL_FLOAT_VEC{2,3,4}
    ///     GL_FLOAT_MAT{2,3,4}
    ///     GL_INT_VEC{2,3,4}
    ///     GL_UNSIGNED_INT_VEC{2,3,4}_EXT
    ///     GL_BOOL_VEC{2,3,4}
    ///
    /// Default \p count =1 sets just one variable. Use higher values to set array
    /// of type variable.
    /// \code
    ///     GLSH: uniform vec3 p[16];
    ///           uniform mat4 m;
    ///     C++:  Vec3f p[16] = {...};
    ///           Matrix4 m;
    ///           shader->setUniform("p", p, GL_FLOAT_VEC3, 16);
    ///           shader->setUniform("m", m, GL_FLOAT_MAT4, 1);
    /// \endcode
    virtual bool setUniform(const char*name, const void* value, int type, unsigned int count=1) const = 0;

    /// \brief Setup this shaders active.
    /// Note: automatically calls compile() on first draw.
    virtual void enable() const = 0;

    /// \brief Put back OpenGL pipeline in default state.
    /// Note: automatically called when destroyed.
    virtual void disable() const = 0;

    /// \brief Returns an OpenGL identifier naming for this shader.
    virtual int prog() const = 0;
};

////////////////////////////////////////////////////////////////////////
/// \brief Derived class to manage ARB assembleur shader objects.
///
class ARBShader : public Shader
{
    /// Utility to check'n print compile or link error log.
    static bool checkLog();

    /// Utility for any of the three shader
    static bool loadNCompile(int type,
                             unsigned int program,
                             const char *fileName,
                             const char *environment);

protected:
    const char* _env;               //!< Lines to prepend to source, before loadind into GL.
    const char* _vertexFile;        //!< Where the vertex shader source is, if any.
    const char* _fragmentFile;      //!< Where the gragment shader source is, if any.

    unsigned int _vp, _fp;          //!< Vertex and Fragmnent programs id.

public:
    ARBShader(const char *vertexFile,
              const char *fragmentFile,
              const char *environment = 0);
    virtual ~ARBShader();

    /// \brief Load and compile shaders.
    /// Returns true if everything is ok. Fire qWarning() if it get
    /// trouble compiling.
    virtual bool compile();

    /// \brief Set a uniform float variable.
    /// Should be called after a successfull compile, at anytime before
    /// an enable() or after a disable().
    /// Returns true if evrything is ok, false if not.
    virtual bool setUniform(const char*name, float value) const;
    virtual bool setUniform(const char*name, int value, int type = -1) const;
    virtual bool setUniform(const char*name, const void* value, int type, unsigned int count=1) const;

    /// \brief Setup this shaders active.
    /// Note: automatically calls compile() on first draw.
    virtual void enable() const;

    /// \brief Put back OpenGL pipeline in default mode.
    /// Note: automatically called when destroyed.
    virtual void disable() const;

    /// \brief Returns an OpenGL identifier naming for this shader.
    /// Returns vertex's shader name, to be able to trace it with the
    /// Feedback class.
    virtual int prog() const;
};


////////////////////////////////////////////////////////////////////////
/// \brief Derived class to manage GLSL shader objects.
///
class GLShader : public ARBShader
{
    /// Utility for any of the three shader
    static bool loadNCompile(int type,
                             unsigned int &shader,
                             const char *fileName,
                             const char *environment);

    /// Utility to check'n print compile or link error log.
    static bool checkLog(const char *fileName, unsigned int program, unsigned int shader, int type);

protected:
	const char* _geometryFile;      //!< Where the geometry shader source is, if any.
    int _inType, _outType;          //!< Geometry shader i/o types, if any.
    int _nbOut;                     //!< Max number of output vertices.

    unsigned int _gs;				//!< ShaderObject id.
    unsigned int _program;          //!< The complete shader program.

public:
    GLShader(const char *vertexFile,
             int inType, const char *geometryFile, int outType,
             const char *fragmentFile,
             const char *environment = 0);
    GLShader(const char *vertexFile,
             const char *fragmentFile,
             const char *environment = 0);
    virtual ~GLShader();

    /// \brief Setup maximum number of vertices output by the geometry shader.
    /// Default to 3.
    virtual void setMaxOutput(int nbOut) { _nbOut = nbOut; }

    /// \brief Load and compile shaders.
    /// Returns true if everything is ok. Fire qWarning() if it get
    /// trouble compiling.
    virtual bool compile();

    /// \brief Setup this shaders active.
    /// Note: automatically calls compile() on first draw.
    virtual void enable() const;

    /// \brief Put back OpenGL pipeline in default state.
    /// Note: automatically called when destroyed.
    virtual void disable() const;

    /// \brief Returns an OpenGL identifier naming for this shader.
    virtual int prog() const;
};

////////////////////////////////////////////////////////////////////////
/// \brief Derived class to manage Cg shader objects.
///
class CgShaderPrivate;
class CgShader : public Shader
{
    CgShaderPrivate* p;
public:
    CgShader(const char *vertexFile,
             const char *fragmentFile);
    virtual ~CgShader();

    /// \brief Load and compile shaders.
    /// Returns true if everything is ok. Fire qWarning() if it get
    /// trouble compiling.
    virtual bool compile();

    /// \brief Set a uniform float variable.
    /// Should be called after a successfull compile, at anytime before
    /// an enable() or after a disable().
    /// Returns true if evrything is ok, false if not.
    virtual bool setUniform(const char*name, float value) const;
    virtual bool setUniform(const char*name, int value, int type = -1) const;
    virtual bool setUniform(const char*name, const float* value, int type, unsigned int count = 1) const;
    virtual bool setUniform(const char*name, const int*   value, int type, unsigned int count = 1) const;

    /// \brief Setup this shaders active.
    /// Note: automatically calls compile() on first draw.
    virtual void enable() const;

    /// \brief Get back to default pipeline.
    /// Note: automatically called when destroyed.
    virtual void disable() const;

    /// \brief Returns an OpenGL identifier naming for this shader.
    virtual int prog() const;
};
