//////////////////////////////////////////////////////////////////////////////
/*! \file   phong_fragment.glsl
 *  \brief  Basic per-pixel lighting, juste to test relative speed.
 *  \author J.D. Gascuel
 *  \sa phong_vertex.glsl, phong_geometry.glsl
 */
//////////////////////////////////////////////////////////////////////////////
// $Id: phong_fragment.glsl,v 1.53 2007/12/08 18:47:49 gascuel Exp $
// File History:
// [jDG 2006/12/05] : Creation from Cg version

#version 120
#extension EXT_texture_array    : enable
#extension EXT_gpu_shader4      : enable

//----------------------------------------------------------------------------
// Per Scene data :

uniform float uFar  = 1.0;      // Scene far distance.
uniform float uNear = 1.0;      // Scene near distance.

// Lighting. If 0, use interpolated per-vertex-colors.
uniform int uLights;

// Color texture, if any.
uniform bool uHasColorMap;
uniform sampler2D uColorMap;

// Additional environment or glossy mapping, if any.
#if defined(PARABOLIC_ENV_MAP)
    uniform bool uHasParabolicEnvMap;
#   ifdef PARABOLIC_ENV_MAP_ARRAY
        uniform sampler2DArray uEnvMapParabolic;
#   else
        uniform sampler2D uEnvMapParabolic;
#   endif
#elif defined(ENV_MAP)
    uniform bool uHasEnvMap;
    uniform samplerCube uEnvMap;
#endif

// Bump texture, if any.
#ifdef BUMP_MAPPING
uniform bool uHasBumpMap;
uniform sampler2D uBumpMap;
#endif

// Shadow map data :
#ifdef SHADOW_MAPPING
uniform bool uHasShadow;			// Should we apply shadow map to this object ?
uniform sampler2DShadow uShadowMap;	// Projection from world to light frame.
#endif

// Per pixel data :
varying in vec4 oP;     // Surface point, in camera coordinates (after modelview, before proj).
varying in vec3 oN;     // Surface normal, in camera coordinates.
varying in vec4	oS;	    // Position in light's frame.

//////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////
///
/// Directional-light calculation.
/// Derived from OpenGL Shading Language, second edition, p 237.

void
directionalLight(int  i,         //!< Light number.
                 vec3 E,         //!< Eye direction (usually +z)
                 vec3 N,         //!< normal vector, in camera coordinates.
                 inout vec4 color)
{
    vec3 V = normalize(gl_LightSource[i].position.xyz); // Light direction.
    vec3 H = normalize(V + E);  // mid between surface-to-light and look direction.

    float nDotV = max(0.0, dot(N, V));
    float nDotH = max(0.0, dot(N, H));

    float pf = 0.0;
    if( nDotV > 0.0 && gl_FrontMaterial.shininess > 0.0 )   // Shinny only on the front side.
        pf = pow(nDotH, gl_FrontMaterial.shininess);

    color += gl_FrontLightProduct[i].ambient
          +  gl_FrontLightProduct[i].diffuse  * nDotV
          +  gl_FrontLightProduct[i].specular * pf;
}

//////////////////////////////////////////////////////////////////////////////
///
/// Point-light calculation.
/// Derived from OpenGL Shading Language, second edition, p 238.

void
pointLight(int  i,         //!< Light number.
           vec3 P,         //!< pixel position, in camera coordinates.
           vec3 E,         //!< Eye direction (usually +z)
           vec3 N,         //!< normal vector, in camera coordinates.
           inout vec4 color)
{
    vec3 V =  gl_LightSource[i].position.xyz - P;   // Surface to ligth vector.
    float d = length(V);        // Distance from light source.
    V = normalize(V);           // Light direction (in camera coordinates).

    float atn = 1.0 / ( gl_LightSource[i].constantAttenuation
                      + gl_LightSource[i].linearAttenuation    * d
                      + gl_LightSource[i].quadraticAttenuation * d * d);

    vec3 H = normalize(V + E);  // mid between surface-to-light and look direction.
    float nDotV = max(0.0, dot(N, V));
    float nDotH = max(0.0, dot(N, H));

    float pf = 0.0;
    if( nDotV > 0.0 && gl_FrontMaterial.shininess > 0.0 ) // Shinny only on the front side.
        pf = pow(nDotH, gl_FrontMaterial.shininess);

    color  += gl_FrontLightProduct[i].ambient
           +  gl_FrontLightProduct[i].diffuse  * nDotV * atn
           +  gl_FrontLightProduct[i].specular * pf * atn;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

void main()
{
#ifdef SUBMODE_COVERAGE //////////////////////////////////////////////////////

	gl_FragColor = vec4(0.5,0.5,0.5, 1);
    gl_FragDepth = 0.999;     // forced at infinity.

#elif defined(SUBMODE_DEPTH) /////////////////////////////////////////////////

    float depth = (abs(oP.z)-uNear)/(uFar-uNear);			// Z in [0..-1] before projection.
    gl_FragColor = vec4(vec3(depth), 1);

#elif defined(SUBMODE_DEBUG) /////////////////////////////////////////////////

    vec3 normal = vec3(0.5) + 0.5 * normalize(oN);
    vec3 P = vec3(0.5) + (0.5/ oP.w) * oP.xyz ;
    gl_FragColor = vec4(P * normal, 0.5);

#else // SUBMODE_COLOR ///////////////////////////////////////////////////////

#ifdef TEST_SHADING //////////////////////////////////////////////////////////

    // Fake shading: use parameters (to avoid over-optimisations),
    // but don't do any time-consumming computations...
    
    // Interpolat RGBA coordinates
#ifdef PER_PIXEL_LIGHTINGS
    gl_FragColor = gl_FrontLightModelProduct.sceneColor
                 + gl_FrontLightProduct[0].ambient
                 + gl_FrontLightProduct[0].diffuse;

    // Interpolat Normal coordinates.
    gl_FragColor.rgb *= vec3(0.5) + 0.5 * oN;
#   ifdef SHADOW_MAPPING
    // If needed, interpolate shadow coordinates.
    if( uHasShadow ) // per object shadow-mapping enable.
        gl_FragColor.rgb *= oS.xyz / oS.w;
#   endif
#else // Interpolated colors...
    gl_FragColor = gl_Color;
#endif

#ifdef COLOR_TEXTURE
    // If needed, use interpolat texture coordinates.
    if( uHasColorMap )
        gl_FragColor.rg *= gl_TexCoord[COLOR_TEXTURE].xy;
#endif

#else // TEST_SHADING ////////////////////////////////////////////////////////

#if defined(PER_PIXEL_LIGHTINGS) || defined(PARABOLIC_ENV_MAP) || defined(ENV_MAP)
    // Material property lighting : p 241 :
    vec3 P = oP.xyz / oP.w;

    vec3 normal;
#if defined(BUMP_MAPPING) && defined(COLOR_TEXTURE)
	if( uHasBumpMap )
	{
		// BUG : coz' dUdx is a constant over a triangle, this algo
        //       produces badly <<flat>> shaded triangles !
        vec3 V = normalize( cross(  normal,
                                    vec3(1, // dFdx(gl_TexCoord[COLOR_TEXTURE].x),
                                         0, // dFdy(gl_TexCoord[COLOR_TEXTURE].x),
                                         0.0) ) );
        vec3 U = cross(normal, V);
        normal = texture2D(uBumpMap, gl_TexCoord[COLOR_TEXTURE].xy).xyz * mat3(U, V, normal); 
		normal = normalize(normal);
	}
    else
#endif
    normal = normalize(oN);

#if defined(PER_PIXEL_LIGHTINGS)
    // Viewing direction is fragment depedent.
    vec3 E = -normalize(P);

#ifdef SHADOW_MAPPING
	// TODO : percentage closer filtering:
	//   1- shadow --> float.
	//   2- have space-sized jittering. That is before w division ?
	//   3- change oS too. Add dS/dU * dU ?
	//   4- make N lookups. average all comparaison results.
    // See shadow2DProjGrad() ?
	bool shadowed = false;
    if( uHasShadow ) // per object shadow-mapping enable.
	{
	   shadowed = oS.w * shadow2DProj(uShadowMap, oS).z <= oS.z;
    }
#endif

    gl_FragColor = gl_FrontLightModelProduct.sceneColor;	// emissive + global ambient light.
    for(int l=0; l < uLights; ++l)
    {
#ifdef SHADOW_MAPPING
	    if( l==0 && shadowed )
			gl_FragColor += gl_FrontLightProduct[0].ambient;
		else
#endif
		{
			if( gl_LightSource[l].position.w == 0.0 )
				directionalLight(l, E, normal, gl_FragColor);
			else    // Let's suppose it is not a spot (ie. spotCutoff == 180.0)
				pointLight(l, P, E, normal, gl_FragColor);
		}
    }
#else // PARABOLIC_ENV_MAP, but no PER_PIXEL_LIGHTINGS
    gl_FragColor = gl_Color;
#endif 
#else // no PER_PIXEL_LIGHTINGS and no PARABOLIC_ENV_MAP
    gl_FragColor = gl_Color;
#endif

#ifdef COLOR_TEXTURE
    if( uHasColorMap )
        gl_FragColor.xyz *= texture2D( uColorMap, gl_TexCoord[COLOR_TEXTURE].xy).xyz;
#endif // COLOR_TEXTURE

#if defined(PARABOLIC_ENV_MAP)
    if( uHasParabolicEnvMap )
    {
        vec3 r = reflect(normalize(P), normal);         // In camera's coordinates.
        r = mat3(gl_ModelViewMatrixTranspose) * r;      // Back to world orientation.

        // The paraboloid transform, for a unit vector:
        r /= 1.0 + abs(r.z);

#   ifdef PARABOLIC_ENV_MAP_ARRAY
        // Use 2 layers of a 2D texture array : map to [0..1]x[0..1]x{0,1}
        if( r.z < 0.0 )
            r = vec3(0.5, 0.5, 1.0) + vec3(-0.49,-0.498, 0.0) * r;
        else
            r = vec3(0.5, 0.5, 0.0) + vec3(-0.49, 0.498, 0.0) * r;

        // Get the reflected color, and mix-up with lighted pixel:
        gl_FragColor.rgb = mix(gl_FragColor.rgb, texture2DArray(uEnvMapParabolic, r).rgb, 0.5);
#   else
        // Use a single texture, twice as height : map to [0..1]x[0..2/2]
        if( r.z < 0.0 )
            r.xy = vec2(0.5, 0.75) + vec2(-0.49,-0.248) * r.xy;
        else
            r.xy = vec2(0.5, 0.25) + vec2(-0.49, 0.248) * r.xy;

        // Get the reflected color, and mix-up with lighted pixel:
        gl_FragColor.rgb = mix(gl_FragColor.rgb, texture2D(uEnvMapParabolic, r.xy).rgb, 0.5);
#   endif
    }
#elif defined(ENV_MAP)
    if( uHasEnvMap )
    {
        vec3 r = reflect(normalize(P), normal);         // In camera's coordinates.
        r = mat3(gl_ModelViewMatrixTranspose) * r;      // Back to world orientation.

        // Get the reflected color, and mix-up with lighted pixel:
        gl_FragColor.rgb = mix(gl_FragColor.rgb, textureCube(uEnvMap, r).rgb, 0.5);
    }
#endif
	// Keep alpha channel safe...
    gl_FragColor.a = gl_Color.a;

#endif  // TEST_SHADING ////////////////////////////////////////////////////////

#endif // SUBMODE_COLOR ////////////////////////////////////////////////////////
}