////////////////////////////////////////////////////////////////////////
/*! \file   sphereBox_geometry.glsl
 *  \brief  Cosinus-Sphere projection. BBox flavor.
 *  \author J.D. Gascuel
 *  \sa common_vertex.glsl, sphere_fragment.glsl
 */
////////////////////////////////////////////////////////////////////////
// $Id: sphereBox_geometry.glsl,v 1.31 2007/12/09 16:39:06 gascuel Exp $
// File History:
// [jDG 2007/03/08] : Initial copy from cosinusMap, triangulated flavor.

#version 120
#extension EXT_geometry_shader4 : require
#extension EXT_gpu_shader4      : require

//----------------------------------------------------------------------------

const float epsilon0  = 0.0001f;
const float epsilon1  = 1 + epsilon0;
const float epsilonm1 = 1 - epsilon0;

//----------------------------------------------------------------------------

// Per rendering data :
uniform int  uPrimitiveFilter = -1; // Retain only one triangle if >= 0.
uniform int uCulling = 1;           // Retain only front facing triangles.

uniform bool uHasBumpMap = false;
uniform bool uHasColorMap = false;

flat varying out vec3[4] oColors;   // RGBA components, so that RGBA.i = dot(coord, oColors[i])
flat varying out vec3    oU, oV;    // If uHasTexture is true, tex coordinate U = dot(coord, oU)

// Per pixel data :
varying out vec2 oP;  // Raster position in 3D space.

// For per-pixel-lighting:
varying in  vec3        iN[];       // Normal, output from vertex shader
flat varying out vec3   oNp[3];     // Normal interpolation data (for per-pixel-lighting).

// Shadow map data :
varying in  vec4 iS[];				// Position, in light[1] frame.
flat varying out vec3 oSp[4];		// light space interpolation data.

////////////////////////////////////////////////////////////////////////

#define EMIT_VERTEX(v2D)                    \
{                                           \
    oP = v2D;                               \
    gl_Position = vec4(v2D, 0, 1);          \
    EmitVertex();                           \
}

//----------------------------------------------------------------------------
/// Send an axis aligned rectangle to the fragment processors.
void
emitBBox(vec2 bmin, vec2 bmax)
{
    // Clip by the [-1,+1] square :
    bmin = max( vec2(-1.0,-1.0), bmin);
    bmax = min( vec2( 1.0, 1.0), bmax);

    // Note: TRIANGLE_FAN mode is bogus, it does STRIPs anyway...
    EMIT_VERTEX( vec2(bmin.x, bmin.y) );  // 0
    EMIT_VERTEX( vec2(bmax.x, bmin.y) );  // 1
    EMIT_VERTEX( vec2(bmin.x, bmax.y) );  // 2
    EMIT_VERTEX( vec2(bmax.x, bmax.y) );  // 3
    EndPrimitive();
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/// Reset a BBox to an empty state.
void
resetBBox(inout vec2 bmin, inout vec2 bmax)
{
    bmin = vec2( 2.0);     // We are in -1..+1 space anyway...
    bmax = vec2(-2.0);
}

//----------------------------------------------------------------------------
/// Add a 2D point to an axis aligned bounding-box.
void
addToBBox(inout vec2 bmin, inout vec2 bmax, const vec2 P)
{
    bmin = min(bmin, P);
    bmax = max(bmax, P);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/// Projection functions. From a eye-space point, to the 2D screen space.

vec3
projection(vec3 P)
{
    return normalize(P);
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/// 2D line intersection.
///
/// Solve in (alpha,beta)
///     A + alpha U = B + beta V
/// That is :
///   (B-A).x = alpha U.x - beta V.x
///   (B-A).y = alpha U.y - beta V.y
///
vec2
intersect(vec2 A, vec2 U,
          vec2 B, vec2 V)
{
    float d = U.y*V.x - U.x*V.y;

    if( d != 0.0 )      // Oups, U and V parallel... Just keep vertex A.
    {
        B -= A;
        d = (V.x * B.y - V.y * B.x) / d;
        A += d * U;
    }
    return A;
}

//----------------------------------------------------------------------------
/// Add a centered ellipse arc to a bounding box.

void
addArcToBBox(inout vec2 bmin, inout vec2 bmax, vec3 A, vec3 B, float backFace)
{
    // Make computation about a sub-bbox that is aligned to the elipse,
    // Then add that bbox to the current bbox...
	addToBBox(bmin, bmax, A.xy);

	// Extremal at ellipse's small axis :
    vec3 n = normalize(cross(A, B));        // OAB normal.
    if( backFace * n.z > epsilon0 )
    {
	    vec2 X = n.xy;	                        // Small axis direction. max at (n.z * X)
        float l = length(X);
        if( l < epsilon0 )
            X = normalize(A.xy + B.xy);
        else
            X /= l;

	    vec2 Y = vec2(-X.y, X.x);	            // Big axis, always norm 1

        // coordinates along small/big axis.
        float ax = dot(A.xy, X);
        float bx = dot(B.xy, X);
        float ay = dot(A.xy, Y);
        float by = dot(B.xy, Y);

        // Flip if we get wrong side.
        // Beware we can have ax ~ 0, and/or bx ~ 0 (eg. if A and/or B are at
        // the horizon after a clipping)...
        if( dot(A.xy+B.xy, X) < 0.0 )
        {
            X = -X;
            ax = -ax;
            bx = -bx;
        }

        if( ay * by <= 0.0 ) // Span over max value ?
        {
            X *= abs(n.z);                      // Position of the extremal point.
            addToBBox(bmin, bmax, X + ay * Y);
            addToBBox(bmin, bmax, X + by * Y);
        }
        else // Both positives or negatives ?
        {
            if( abs(ay) < abs(by) )
                addToBBox(bmin, bmax, ax * X + by * Y);
            else
                addToBBox(bmin, bmax, bx * X + ay * Y);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
/// Which is the biggest componant of a vector (absolute value).
/// Returns -1 for X, 0 for Y, and +1 for Z (best for if()).
float
dominant(in vec3 v)
{
    v = abs(v);
    return ( v.x >= v.y && v.x >= v.z ) ? -1.0
         : ( v.y >= v.x && v.y >= v.z ) ?  0.0
         :                                 1.0;
}

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
/// Compress vec3 to vec2 by skipping coordinates designated by dominant.
vec2
crunsh(in vec3 v, in float dominant)
{
    if( dominant <  0.0 ) return v.yz;
    if( dominant == 0.0 ) return v.xz;
                          return v.xy;
}


//////////////////////////////////////////////////////////////////////////////

#if !defined(ISECT_BADOUEL) && !defined(ISECT_MTH)
#	define	ISECT_BADOUEL
#endif

#ifdef ISECT_BADOUEL

// Per-triangle intersection data :
flat varying float   oDominant; 
flat varying vec3    oA;        // Original triangle world coordinates, projected on the
flat varying vec2    oAB;       // unit 3D sphere.
flat varying vec2    oAC;
flat varying vec3    oN;        // Triangle's unit normal vector.

void
intersectSetup(vec3 A, vec3 B, vec3 C)
{
	B -= A;
	C -= A;

    oN         = cross(B, C);
    oA         = A;                 // Keep (oA, N) in 3D, to enable plane projection.

    float d = dominant( oN );
    oDominant  = d;
    oAB        = crunsh(B, d);
    oAC        = crunsh(C, d);
}

#endif // ISECT_BADOUEL

//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

#ifdef ISECT_MTH

// Per-triangle intersection data :
flat varying vec3    oN;        // AB cross AC
flat varying vec3    oQ1;		// AO cross AB
flat varying vec3    oQ2;       // AO cross AC
flat varying float   oT;		// Q1 dot AC

void
intersectSetup(vec3 A, vec3 B, vec3 C)
{
	B -= A;
	C -= A;
	oN  = cross(B, C);	// AB ^ AC
	oQ1 = cross(B, A);	// AB^A == AO^AB
	oQ2 = cross(C, A);	// AC^A == AO^AC
	oT  = dot(oQ1, C);	// C = AC
}

#endif // ISECT_MTH

//////////////////////////////////////////////////////////////////////////////

void main()
{
    // Filter primitive option :
#ifdef PRIMITIVE_FILTERING
    if( (uPrimitiveFilter >= 0) && ( gl_PrimitiveIDIn != uPrimitiveFilter) )
        return;
#endif

    vec3 A = gl_PositionIn[0].xyz;
    vec3 B = gl_PositionIn[1].xyz;
    vec3 C = gl_PositionIn[2].xyz;

    // ---- Is this triangle fully on the wrong side of the camera plane ?
    int code =     int( A.z > 0.0 )
             + 2 * int( B.z > 0.0 )
             + 4 * int( C.z > 0.0 );
    if( code == 7 ) return;

    // ---- Or back-facing ? -------------------------------------------------
    // Triangles front-facing the eye are those one so that 
    //    dot(view, N) < 0
    // Because view = A + alpha AB + beta AC (point on plane),
    // and dot(N,AB) == dot(N,AC) == 0, it is sufficient to test
    // just any corner of the triangle :
	vec3 N  = cross(B-A, C-A);
    float backFace = (dot(A,N) > 0.0) ? -1.0 : 1.0;
#ifdef FACE_CULLING   
	if( (uCulling * backFace) < 0.0 ) return;
#endif

	// ---- Setup attributes shared by all generated triangles ---------------
    intersectSetup(A, B, C);
    
    // ---- Prepare all attributes too ---------------------------------------
#ifdef SUBMODE_COLOR
#ifdef COLOR_TEXTURE
    if( uHasColorMap || uHasBumpMap )
    {
        vec4 d = vec4(gl_TexCoordIn[1][COLOR_TEXTURE].xy, 
                      gl_TexCoordIn[2][COLOR_TEXTURE].xy)
              - gl_TexCoordIn[0][COLOR_TEXTURE].xyxy;
        oU = vec3(gl_TexCoordIn[0][COLOR_TEXTURE].x, d.x, d.z);
        oV = vec3(gl_TexCoordIn[0][COLOR_TEXTURE].y, d.y, d.w);
    }
#endif

#ifdef PER_PIXEL_LIGHTINGS
    oNp[0] = vec3(iN[0].x, iN[1].x - iN[0].x, iN[2].x - iN[0].x);
    oNp[1] = vec3(iN[0].y, iN[1].y - iN[0].y, iN[2].y - iN[0].y);
    oNp[2] = vec3(iN[0].z, iN[1].z - iN[0].z, iN[2].z - iN[0].z);

#ifdef SHADOW_MAPPING
    oSp[0] = vec3(iS[0].x, iS[1].x - iS[0].x, iS[2].x - iS[0].x);
    oSp[1] = vec3(iS[0].y, iS[1].y - iS[0].y, iS[2].y - iS[0].y);
    oSp[2] = vec3(iS[0].z, iS[1].z - iS[0].z, iS[2].z - iS[0].z);
    oSp[3] = vec3(iS[0].w, iS[1].w - iS[0].w, iS[2].w - iS[0].w);
#endif

#else // no PER_PIXEL_LIGHTINGS
    oColors[0] = vec3(gl_FrontColorIn[0].r, gl_FrontColorIn[1].r - gl_FrontColorIn[0].r, gl_FrontColorIn[2].r - gl_FrontColorIn[0].r);
    oColors[1] = vec3(gl_FrontColorIn[0].g, gl_FrontColorIn[1].g - gl_FrontColorIn[0].g, gl_FrontColorIn[2].g - gl_FrontColorIn[0].g);
    oColors[2] = vec3(gl_FrontColorIn[0].b, gl_FrontColorIn[1].b - gl_FrontColorIn[0].b, gl_FrontColorIn[2].b - gl_FrontColorIn[0].b);
    oColors[3] = vec3(gl_FrontColorIn[0].a, gl_FrontColorIn[1].a - gl_FrontColorIn[0].a, gl_FrontColorIn[2].a - gl_FrontColorIn[0].a);
#endif // PER_PIXEL_LIGHTINGS
#endif // SUBCOLOR_MODE

    // ---- Cover the bended triangle with some rendering area -----------
    //

    vec2 bmin, bmax;
    resetBBox(bmin, bmax);

    vec3 As, Bs, Cs, Ds;
    if( code != 0 )
    {
         //Make everybody ~ case 1 : cut A, keep B & C,
         //         or like case 6 : keep corner A, cut B & C.
        if( (code-3) * (code-4) == 0 ) { vec3 tmp = A; A = C; C = B; B = tmp; }
        if( (code-2) * (code-5) == 0 ) { vec3 tmp = A; A = B; B = C; C = tmp; }

        // Exactly two edge intersection do exists.
        // Because of the previous permuttations, we do know that they are
        // indeed intersections on the A-B and the A-C edges :
        Bs = projection( mix(A, B, A.z/(A.z-B.z)) );
        Cs = projection( mix(A, C, A.z/(A.z-C.z)) );

        if( (code-1) * (code-2) * (code-4) == 0 )   // Just one corner clipped ?
        {
            // Keep Quad C/Cp/Bp/B
            As = Bs; Bs = Cs; Cs = As;  // Swap Bs/Cs to keep poly ccw.
            As = projection(C);
            Ds = projection(B);
            addArcToBBox(bmin, bmax, Ds, As, backFace);
        }
        else    // 3, 5, 6                          // Two corner clipped ?
        {
            // Keep A/Bp/Cp triangle. Just close the Quad:
            Ds = As = projection(A);
        }
    }
    else        // 0
    {
        // Keep original triangle
        Ds = As = projection(A);
        Bs = projection(B);
        Cs = projection(C);
    }

    // Make a box, from the three points. The assumption here is that
    // ellipse arcs will be included in their bounding sphere.
    // Which is wrong when we have more than half an ellipse...
    // (partially clipped triangles).
    addArcToBBox(bmin, bmax, As, Bs, backFace);
    addArcToBBox(bmin, bmax, Bs, Cs, backFace);
    addArcToBBox(bmin, bmax, Cs, Ds, backFace);
    emitBBox(bmin, bmax);
}
