//////////////////////////////////////////////////////////////////////////////
/*! \file   ShadowMap.cpp
 *  \brief  Implement OpenGL shadow maps.
 *  \author    Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa FrameBufferObject, Scene, Shader.
 */
//////////////////////////////////////////////////////////////////////////////
// $Id: ShadowMap.cpp,v 1.27 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2007/03/15] : Creation.

#include "ShadowMap.h"

#include <GL/glew.h>

#include <qglobal.h>
#include <qimage.h>

#include "Math/Vec3f.h"
#include "Math/Matrix4.h"
#include "Utilities.h"
#include "Shader.h"

//////////////////////////////////////////////////////////////////////////////

ShadowMap::ShadowMap()
: _width(-1),
  _height(-1),
  _texId(0)
{}

ShadowMap::~ShadowMap()
{
    if( _texId )
    {
        glDeleteTextures(1, &_texId);
        _texId = 0;
    }
}

//////////////////////////////////////////////////////////////////////////////

bool
ShadowMap::bindDepthTexture(int width, int height)
{
    Q_ASSERT( ! _texId );                  // Should be bound just once.

    // Make sure the image is correctly initialized to hold our depth
    // data.
    static int maxRBSize = 0;
    if( ! maxRBSize )
    {
        glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE_EXT, &maxRBSize);
        qDebug("Max renderbuffer size is %d", maxRBSize);
    }
    Q_ASSERT( 0 < width  && width  < maxRBSize );
    Q_ASSERT( 0 < height && height < maxRBSize );

    // Sanity check, before crashing on null function pointers...
    if( ! GLEW_EXT_framebuffer_object )
    {
        qWarning("*** Missing GL_EXT_framebuffer_object");
        return false;
    }

    // Creates the FBO id :
    if( ! _id )
    {
	    glGenFramebuffersEXT(1, &_id);
        CHECK_GL_ERROR("gen shadow FBO");
    }

    // Creates texture id :
    if( ! _texId )
    {
        glGenTextures(1, &_texId);
        CHECK_GL_ERROR("gen shadow map");
    }

    // Make sure the texture is correctly initialized to hold our depth
    // data.
    _width  = width;
    _height = height;
	glActiveTexture(GL_TEXTURE2);
	glBindTexture(GL_TEXTURE_2D, _texId);
	glTexImage2D(GL_TEXTURE_2D, 0,      // Target and mip-map level.          
            GL_DEPTH_COMPONENT,         // Internal format: Choose best size of hardware
            _width, _height,            // texture size. Should be a power of two ???
            0,                          // no border.
            GL_DEPTH_COMPONENT,         // What pixel data should be loaded there ?
            GL_UNSIGNED_INT,            // As 32bit unsigned.
            0);                         // And no initialisation data (yet).

    // Only linear mapping is supported (or u get uncomplete buffer !)
    // This may be used by implementation to do multiple depth comparaisons...
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// Keep z value on the border of the shadow map. Other options are:
	//	CLAMP_TO_BORDER --> should define a border color as an outside depth
	//	CLAMP           --> provides a 50% shadow out of the depth map.
    //  CLAMP_TO_EDGE   --> keep depthmap border value: nice if bigger than shadow.
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    CHECK_GL_ERROR("clamp shadow map");

    // Tells OpenGL the FBO is to be mapped to that texture
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, 
                              GL_DEPTH_ATTACHMENT_EXT, 
                              GL_TEXTURE_2D,    // Destination type,
                              _texId, 0);       // id and mip-map level.
    CHECK_GL_ERROR("attach shadow map");

    // Remember (*FOR THIS FRAMEBUFFER only*) not to draw onto anything else.
	glDrawBuffer(GL_NONE);              // Don't draw to screen too.
	glReadBuffer(GL_NONE);              // Don't read from screen ???
    CHECK_GL_ERROR("don't draw colors");

    // Make sure everything went well :
    uint complete = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
    if( complete != GL_FRAMEBUFFER_COMPLETE_EXT)
    {
        qWarning("*** Uncomplete frame buffer. Code 0x%x", complete);
        return false;
    }

	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);

    return true;
}

//////////////////////////////////////////////////////////////////////////////

void
ShadowMap::setupLightMatrix(const Vec3f& eye, 
                            const Vec3f& center,
                            float sceneRadius,
                            float fovy, 
                            float aspectRatio) const
{
	glPushAttrib(GL_VIEWPORT_BIT);
	glViewport(0, 0, _width, _height);

    // ---- Automatic near/far computation and projection matrix -------------
    float zFar  = length(eye-center) + 1.01f*sceneRadius;
    float zNear = zFar - 2.02f * sceneRadius;
    if( zNear < 0.01f ) zNear = 0.01f;

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	gluPerspective(fovy,        // y-angle (degrees).
                   aspectRatio, // fovy/fovx
                   zNear,       // Near clip plane
                   zFar);       // Far clip plane.
    glGetFloatv(GL_PROJECTION_MATRIX, (GLfloat*)_lightProjectionMatrix._m);
	CHECK_GL_ERROR("shadow map perspective");
	
    // ---- Automatic up calculation and modelview matrix --------------------
    Vec3f up(0,1,0);
    if( length2(cross(up,eye-center)) < 0.01f ) up = Vec3f(0,0,1);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
    gluLookAt(eye._x, eye._y, eye._z,           // Eye position.
              center._x, center._y, center._z,  // Look direction.
              up._x, up._y, up._z);             // Up vector.
    glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat*)_lightViewMatrix._m);
	CHECK_GL_ERROR("shadow map LookAt");

    // ---- Use offset so that face won't self shadow themselfes -------------
	glPolygonOffset(2.0f, 4.0f);	
	glEnable(GL_POLYGON_OFFSET_FILL);

    // ---- Make sure depth is initialized -----------------------------------
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, _id);
    glClear(GL_DEPTH_BUFFER_BIT);   // Make sure we start from a clear texture.
	CHECK_GL_ERROR("clear shadow map");
}

void
ShadowMap::cleanMatrix() const
{
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);    // Back to screen.
	glDisable(GL_POLYGON_OFFSET_FILL);				// Back to normal Z calculation.
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();                                  // Pop light view matrix.
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();                                  // Pop light projection matrix.
    glPopAttrib();                                  // pop viewport.
}

//////////////////////////////////////////////////////////////////////////////

void ShadowMap::enable(int texUnit, const Shader* shader) const
{
	// Beware: Matrix4 constructor make a transpose !
    static Matrix4 biasMatrix(0.5f, 0.0f, 0.0f, 0.5f,	// Maps [-1..1] to [0..1], in X, Y, Z.
                              0.0f, 0.5f, 0.0f, 0.5f,
                              0.0f, 0.0f, 0.5f, 0.5f,
                              0.0f, 0.0f, 0.0f, 1.0f);
    Matrix4 textureMatrix = biasMatrix * _lightProjectionMatrix * _lightViewMatrix  ;

	glActiveTexture(GL_TEXTURE0 + texUnit);
	glBindTexture(GL_TEXTURE_2D, _texId);

	if( shader )
	{
		// Note : shader should be alreay enabled for this to work...
		shader->setUniform("uHasShadow", true, GL_BOOL);
        shader->setUniform("uShadowMatrix", textureMatrix, GL_FLOAT_MAT4);
        shader->setUniform("uShadowMap", texUnit, GL_SAMPLER_2D_SHADOW); // BUG when driver < 100 : w/o _SHADOW
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_COMPARE_R_TO_TEXTURE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_FUNC, GL_LEQUAL);    // (default)
        // Note : GL_LUMINANCE + GL_MODULATE multiply the current pixel color
        //        by the shadow function, in [0..r].
		glTexParameteri(GL_TEXTURE_2D, GL_DEPTH_TEXTURE_MODE, GL_LUMINANCE);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
		CHECK_GL_ERROR("shadow map GL MODULATE");

		Matrix4 transpose = textureMatrix.transpose();

		// Uses TexGen to generate coordinates in light coordinate system :
		glTexGeni (GL_S, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGenfv(GL_S, GL_EYE_PLANE, transpose[0]);
		glEnable(GL_TEXTURE_GEN_S);

		glTexGeni (GL_T, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGenfv(GL_T, GL_EYE_PLANE, transpose[1]);
		glEnable(GL_TEXTURE_GEN_T);

		glTexGeni (GL_R, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGenfv(GL_R, GL_EYE_PLANE, transpose[2]);
		glEnable(GL_TEXTURE_GEN_R);

		glTexGeni (GL_Q, GL_TEXTURE_GEN_MODE, GL_EYE_LINEAR);
		glTexGenfv(GL_Q, GL_EYE_PLANE, transpose[3]);
		glEnable(GL_TEXTURE_GEN_Q);
		CHECK_GL_ERROR("shadow map GL coord gen");
	}

	glEnable(GL_TEXTURE_2D);
}

void ShadowMap::disable(int texUnit, const Shader* shader) const
{
	if( shader )
	{
		shader->setUniform("uHasShadow", false, GL_BOOL);
	}
	else
	{
		glActiveTexture(GL_TEXTURE0 + texUnit);

		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_COMPARE_MODE, GL_NONE);
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glDisable(GL_TEXTURE_GEN_R);
		glDisable(GL_TEXTURE_GEN_Q);
		glBindTexture(GL_TEXTURE_2D, 0);
	    glDisable(GL_TEXTURE_2D);

		CHECK_GL_ERROR("shadow map GL disable");
	}
}

//////////////////////////////////////////////////////////////////////////////

bool
ShadowMap::finish(QImage& img) const
{
    // ---- Some sanity checking ---------------------------------------------
    Q_ASSERT(_texId);
    if( img.depth() != 8 || img.width() != _width || img.height() != _height )
    {
        qWarning("Wrong image depth or size.");
        return false;
    }

    // ---- Read texture data ------------------------------------------------
    float *buffer = new float[_width*_height];
    Q_CHECK_PTR(buffer);

    glBindTexture(GL_TEXTURE_2D, _texId);
    glGetTexImage(GL_TEXTURE_2D,        // What.
                  0,                    // level
                  GL_DEPTH_COMPONENT,   // format
                  GL_FLOAT,             // type
                  buffer);
    glBindTexture(GL_TEXTURE_2D, 0);
    CHECK_GL_ERROR("get shadow map image");

    // ---- Auto contrast ---------------------------------------------------
    float min = buffer[0], max = buffer[0];
    for(int i=1; i < _width*_height; ++i)
    {
        if( min > buffer[i] ) min = buffer[i];
        if( max < buffer[i] ) max = buffer[i];
    }
    float scale = (max > min) ? 255.0f / float(max-min)
                              : 1.0f;

    // ---- 8 bit conversion & saving ----------------------------------------

    // Have a Qt's gray scale image.
    img.setNumColors(256);
    for(int c=0; c<256; ++c)
        img.setColor(c, qRgb(c, c, c));

    // Copy data
    int p = 0;                          // Uniform access through OpenGL's data.
    for(int y=0; y<_height; ++y)
    {
        // Faster access. And do the vertical mirror all-in-one.
        unsigned char *q = img.scanLine(_height - 1 - y);
        for(int x=0; x<_width; ++x)
        {
            unsigned char c = (unsigned char) (scale * (buffer[p++] - min));
            q[x] = c;
        }
    }
    delete[] buffer;

    return true;
}
