////////////////////////////////////////////////////////////////////////
/*! \file   ShadowMap.h
 *  \brief  Implement pure OpenGL and shaders shadow maps.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 * \sa FrameBufferObject, Scene, Shader.
 */
////////////////////////////////////////////////////////////////////////
// $Id: ShadowMap.h,v 1.10 2008/02/20 11:49:45 gascuel Exp $
// File History:
// [jDG 2007/03/15] : Creation.

#ifndef SHADOW_MAP_H
#define SHADOW_MAP_H

#include "FrameBufferObject.h"

#include "Math/Matrix4.h"

class Vec3f;
class Shader;

/*! \brief Implement OpenGL shadow maps.
 *
 * Uses frame buffer objects and fragment shader to do modern 
 * OpenGL shadow map.
 */
class ShadowMap : private FrameBufferObject
{
    int             _width;     //!< Depth map width. Default to 512.
    int             _height;    //!< Depth map height. Default to 512.
    unsigned int    _texId;     //!< Depth texture OpenGL name. 0 if FBO not yet inited.
    Matrix4         _lightProjectionMatrix;
    Matrix4         _lightViewMatrix;

public:
    ShadowMap();
    virtual ~ShadowMap();

    /// \brief Was bindDepthTexture() called ?
    virtual bool bound() const { return _texId != 0; }

    /*! \brief Attach a FBO to the depth texture (eg. to do shadow maps).
     * \code
     *      FrameBufferObject fbo;
     *      if( ! fbo.bindDepthTexture(_texId, 512, 512) )
     *          return;     // Unsupported feature !
     * \endcode
     *
     * Note: render buffer is not let bound. Yoy should call setupLightMatrix() afterward.
     */
    virtual bool bindDepthTexture(int width, int height);

    /*! \brief Push, then load the current matrices with the shadowmap matrix.
     * Let the framebuffer bound, and clear z-buffer so rendering can start right away.
     */
    virtual void setupLightMatrix(const Vec3f& eye, 
                                  const Vec3f& center,
                                  float sceneRadius,
                                  float fovy = 40.0f, 
                                  float aspectRatio = 1.0f) const;

    /// \brief pop matrices.
    void cleanMatrix() const;

    /// \brief setup OpenGL texture unit \p texUnit for shadow map.
    void enable(int texUnit, const Shader* shader = 0) const;

    /// \brief back to standard light rendering.
    void disable(int texUnit, const Shader* shader = 0) const;

    //! Stop render redirection, and wait for the image to safely by
    //! in the image.
    virtual bool finish(QImage& img) const;
};

#endif // SHADOW_MAP_H

