////////////////////////////////////////////////////////////////////////
/*! \file   main.cpp
 *  \brief  Demo program for Spherical mapping.
 *  \author J.D. Gascuel
 *
 */
////////////////////////////////////////////////////////////////////////
// $Id: glslc.cpp,v 1.8 2007/04/18 16:24:26 gascuel Exp $
// File History:
// [jDG 2006/11/29] : Creation.

#include <GL/glew.h>
#include <QApplication>
#include <QGLWidget>
#include <QFile>
#include <QTextStream>
#include <QStringList>

#include "Utilities.h"

//////////////////////////////////////////////////////////////////////////////

static bool
checkLog(const char *fileName, unsigned int program, unsigned int shader, int type)
{
    GLint rc = 0;
    GLint infologLength = 0;
    if( shader )
    {
        glGetShaderiv(shader, type, &rc);
	    glGetShaderiv(shader, GL_INFO_LOG_LENGTH, &infologLength);
        CHECK_GL_ERROR("check shader log");

        if( (infologLength > 0)  && (rc == GL_FALSE) )
	    {
	        GLint charsWritten  = 0;
	        char *infoLog;

	        infoLog = new char[infologLength];
	        glGetShaderInfoLog(shader, infologLength, &charsWritten, infoLog);
            CHECK_GL_ERROR("check shader log");
            qWarning("*** GL shader %s info log:", fileName);

			// Add file name, so that MSVC recognize that as an error message :
			QStringList lines = QString(infoLog).split('\n', QString::SkipEmptyParts);
	        delete infoLog;
			for(int l=0; l<lines.count(); ++l)
				if( lines[l][0] == '(' )
					lines[l] = QString(fileName)+lines[l];

			// And report.
		    qWarning(lines.join("\n").toLatin1().constData());

            return false;
	    }
    }
    if( program )
    {
        glGetProgramiv(program, type, &rc);
	    glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infologLength);
        CHECK_GL_ERROR("check program log");

        if( (infologLength > 0) && (rc == GL_FALSE) )
	    {
	        GLint charsWritten  = 0;
	        char *infoLog;

	        infoLog = new char[infologLength];
	        glGetProgramInfoLog(program, infologLength, &charsWritten, infoLog);
            CHECK_GL_ERROR("check program log");
            qWarning("*** GL program %s info log:", fileName);

			// Add file name, so that MSVC recognize that as an error message :
			QStringList lines = QString(infoLog).split('\n', QString::SkipEmptyParts);
	        delete infoLog;
			for(int l=0; l<lines.count(); ++l)
				if( lines[l][0] == '(' )
					lines[l] = QString(fileName)+lines[l];

			// And report.
		    qWarning(lines.join("\n").toLatin1().constData());

            return false;
	    }
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

static bool
loadNCompile(GLint type,
             GLuint &shader,
             const char *fileName)
{
    // ---- Read shader source -----------------------------------------------
    QFile in(fileName);
    if( !in.open(QIODevice::ReadOnly) )
    {
        qWarning("Unable to find shader source %s",
            fileName);
        return false;
    }
    qint64 len = in.size();
    char *shaderSource = new char[len+1];
    Q_CHECK_PTR(shaderSource);
    if( in.read(shaderSource, len) != len )
    {
        qWarning("Error reading shader file %s",
            fileName);
        in.close();
        delete shaderSource;
        return false;
    }
    in.close();
    shaderSource[len] = '\0';       // Close string, just for debug purpose.

    // ---- Create shader object ---------------------------------------------
    shader = glCreateShader(type);
    if( ! shader )
    {
        qWarning("*** Unable to create shader object for %s",
                 fileName);
        return false;
    }

    // ---- Load and compile source into shader object -----------------------
    const char *lines[] = { shaderSource };
    const GLint lens [] = { len };
    glShaderSource(shader, 1, lines, lens);
    glCompileShader(shader);
    delete[] shaderSource;

    if( ! checkLog(fileName, 0, shader, GL_COMPILE_STATUS) )
    {
        qWarning("*** Compile shader %s failed.",
               fileName);
        return false;
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
    QApplication app(argc, argv);

    GLuint vs=0, gs=0, fs=0;
    GLint inType = GL_TRIANGLES;
    GLint outType = GL_TRIANGLE_STRIP;
    GLint nbOut   = 4;
	QString output = QString::null;

    // ---- Init OpenGL extensions wrangler, and check for needed ones -------
    QGLWidget w;        // Creates a temporary OpenGL context, just to make 
    w.makeCurrent();    // sure that GLEW have a chance to get GL capabilities.
    const char *glversion = (char*)glGetString(GL_VERSION);
    glewInit();         // Setup all variables, for instant checks later on.

    qWarning("OpenGL version %s", glversion);
    if( ! GLEW_VERSION_2_0 )
    {
        QStringList list = QString((char*)glGetString(GL_EXTENSIONS)).split(' ');
        list.sort();
        qWarning("OpenGL extension string = \n\t%s\n",
                list.join("\n\t").toAscii().constData());
        qFatal("*** OpenGL 2.0 is mandatory for plain Shader Objects");
    }

    // ---- Parse options and run compiler...

    for(int i=1; i<argc; ++i)
    {
        QString arg(argv[i]);

		if( arg == "-o" )
			output = argv[++i];
		else if( arg == "-vertex" )
        {
            if( ! GLEW_ARB_vertex_shader )
                qFatal("*** Missing OpenGL extension GLEW_ARB_vertex_shader");

            vs = glCreateShader(GL_VERTEX_SHADER);
            if( ! loadNCompile(GL_VERTEX_SHADER, vs, argv[++i]) )
                return -1;

			if( !output.isEmpty() )
			{
				QFile out(output);
				if( out.open(QIODevice::WriteOnly) )
				{
					QTextStream sout( &out );
					sout << "Shader listing not yet implemented in glslc" << endl;
				}
				out.close();
			}
        }
#ifdef GL_GEOMETRY_SHADER_EXT
        else if( arg == "-geometry" )
        {
            if( ! GLEW_EXT_geometry_shader4 )
                qFatal("*** Missing OpenGL extension GL_EXT_geometry_shader4");

            gs = glCreateShader(GL_GEOMETRY_SHADER_EXT);
            if( ! loadNCompile(GL_GEOMETRY_SHADER_EXT, gs, argv[++i]) )
                return -2;

			if( !output.isEmpty() )
			{
				QFile out(output);
				if( out.open(QIODevice::WriteOnly) )
				{
					QTextStream sout( &out );
					sout << "Shader listing not yet implemented in glslc" << endl;
				}
				out.close();
			}
        }
#endif
        else if( arg == "-fragment" )
        {
            if( ! GLEW_ARB_fragment_shader )
                qFatal("*** Missing OpenGL extension GL_ARB_fragment_shader");

            fs = glCreateShader(GL_FRAGMENT_SHADER);
            if( ! loadNCompile(GL_FRAGMENT_SHADER, fs, argv[++i]) )
                return -3;

			if( !output.isEmpty() )
			{
				QFile out(output);
				if( out.open(QIODevice::WriteOnly) )
				{
					QTextStream sout( &out );
					sout << "Shader listing not yet implemented in glslc" << endl;
				}
				out.close();
			}
        }
        else
            qWarning( "Unknown argument %s", arg.toAscii().constData() );
    }

    if( vs || fs )	// Linking do need a vs when there is a gs.
    {
        GLuint program = glCreateProgram();
        if( vs ) glAttachShader(program, vs);
	    if( fs ) glAttachShader(program, fs);
	    if( vs && gs )
        {
            glAttachShader(program, gs);
            glProgramParameteriEXT(program, GL_GEOMETRY_INPUT_TYPE_EXT,   inType);
            glProgramParameteriEXT(program, GL_GEOMETRY_OUTPUT_TYPE_EXT,  outType);
            glProgramParameteriEXT(program, GL_GEOMETRY_VERTICES_OUT_EXT, nbOut);
        }
        glLinkProgram(program);
        if( !checkLog("LINK", program, 0, GL_LINK_STATUS) )
            return -4;
    }

    return 0;
}
