////////////////////////////////////////////////////////////////////////
/*! \file   main.cpp
 *  \brief  Demo program for Spherical mapping.
 *  \author Jean-Dominique Gascuel
 *  \copyright ARTIS, Grenoble Universities, 2008.
 *
 */
////////////////////////////////////////////////////////////////////////
// $Id: main.cpp,v 1.52 2008/02/20 11:49:46 gascuel Exp $
// File History:
// [jDG 2006/11/29] : Creation.

#include <GL/glew.h>
#include <QApplication>

#include <QMainWindow>
#include <QMenuBar>

#include "BasicGLViewer.h"

////////////////////////////////////////////////////////////////////////

static QMainWindow* mainWindow = 0;
static BasicGLViewer* viewer = 0;
static QAction *noGShaderAction = 0;
static QAction *noFragsAction = 0;
static QAction *wireframeAction = 0;
static QAction *opaqWireframeAction = 0;
static QAction *shadowsAction = 0;
static QAction *twoPassAction = 0;
static QAction *texturesAction = 0;
static QAction *normalMapsAction = 0;
static QAction *environmentMapsAction = 0;
static QAction *dynamicEnvmapsAction = 0;
static QAction *glossyMapsAction = 0;
static QAction *dynamicGlossymapsAction = 0;
static QAction *cullingAction = 0;
static QAction *depthtestAction = 0;
static QAction *blendingAction = 0;

////////////////////////////////////////////////////////////////////////

void
resetInterface()
{
    noGShaderAction->setChecked(false);
    noFragsAction->setChecked(false);
    wireframeAction->setChecked(false);
    opaqWireframeAction->setChecked(false);
    viewer->setLights(1);
    shadowsAction->setChecked(true);
    twoPassAction->setChecked(false);
    texturesAction->setChecked(true);
    normalMapsAction->setChecked(false);
    environmentMapsAction->setChecked(false);

	dynamicEnvmapsAction->setChecked(false);
    dynamicEnvmapsAction->setEnabled(false);
	glossyMapsAction->setChecked(false);
	dynamicGlossymapsAction->setChecked(false);
    dynamicGlossymapsAction->setEnabled(false);
	cullingAction->setChecked(true);
	depthtestAction->setChecked(true);
	blendingAction->setChecked(false);
}

////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
    QApplication app(argc, argv, QApplication::GuiClient);

    // ---- Init OpenGL extensions wrangler, and check for needed ones -------
    const char *glversion = 0, *glrender = 0;
    {
        QGLWidget w;        // Creates a temporary OpenGL context, just to make 
        w.makeCurrent();    // sure that GLEW have a chance to get GL capabilities.
        glversion = (char*)glGetString(GL_VERSION);
        glrender  = (char*)glGetString(GL_RENDERER); 
        glewInit();         // Setup all variables, for instant
    }
    qWarning("OpenGL version %s on %s", glversion, glrender);

    const char *needed[] = {
        "GL_VERSION_2_0",
        "GL_ARB_vertex_shader",         // Needed by GLShader()
        "GL_EXT_gpu_shader4",           // Not defined yet ???
        "GL_EXT_geometry_shader4",
        "GL_ARB_fragment_shader",
        0
    };
    QStringList missing;
    for(const char **e = needed; *e; ++e)
    {
        if( ! glewIsSupported(*e) )
            missing += QString(*e);
        else
            qWarning("OpenGL %s ok.", *e);
    }
    if( !missing.isEmpty() )
        qWarning("*** Missing OpenGL extensions:\n\t%s\n\n"
               "Not running an a G80 GPU (or NVEmulate badly installed)",
               missing.join("\n\t").toLatin1().constData());

    // ---- Open & Run interface ---------------------------------------------
    mainWindow = new QMainWindow;
    Q_CHECK_PTR(mainWindow);
    mainWindow->resize(512,533);	// Add 21px for menu bar	

    viewer = new BasicGLViewer(mainWindow);
    Q_CHECK_PTR(viewer);

    mainWindow->setMenuBar( new QMenuBar(mainWindow) );
    QMenu *file = mainWindow->menuBar()->addMenu("File");
        file->addAction("Load...", viewer, SLOT(loadScene()), Qt::ControlModifier+Qt::Key_O);
        file->addAction("Merge...", viewer, SLOT(mergeScene()));
        file->addSeparator();
        QAction *s = file->addAction("Accumulate statistics"); 
            s->setShortcut(Qt::Key_F); 
            s->setCheckable(true);
            QObject::connect(s, SIGNAL(toggled(bool)), viewer, SLOT(setStatistics(bool)));
        file->addAction("Quit", &app, SLOT(quit()), Qt::Key_Escape);

    // ==== Add a detachable menu with all rendering options =================
    QMenu *modeMenu = mainWindow->menuBar()->addMenu("Render mode");
    modeMenu->setTearOffEnabled(true);

    QActionGroup* rendermodeAction = new QActionGroup(modeMenu);
    QAction *a;
    a = rendermodeAction->addAction("OpenGL 1"      ); a->setShortcut(Qt::Key_1); a->setCheckable(true); modeMenu->addAction(a);
    a = rendermodeAction->addAction("Shaders"       ); a->setShortcut(Qt::Key_2); a->setCheckable(true); modeMenu->addAction(a); a->setChecked(true);
    a = rendermodeAction->addAction("Sphere  / Tris"); a->setShortcut(Qt::Key_3); a->setCheckable(true); modeMenu->addAction(a);
    a = rendermodeAction->addAction("Parabol / Tris"); a->setShortcut(Qt::Key_4); a->setCheckable(true); modeMenu->addAction(a);
    a = rendermodeAction->addAction("Sphere  / Quad"); a->setShortcut(Qt::Key_5); a->setCheckable(true); modeMenu->addAction(a);
    a = rendermodeAction->addAction("Parabol / Quad"); a->setShortcut(Qt::Key_6); a->setCheckable(true); modeMenu->addAction(a);
    a = rendermodeAction->addAction("Lens    / Quad"); a->setShortcut(Qt::Key_7); a->setCheckable(true); modeMenu->addAction(a);
    QObject::connect(rendermodeAction, SIGNAL(triggered(QAction*)), viewer, SLOT(setShader(QAction*)));

    twoPassAction = modeMenu->addAction( "Two passes");
        twoPassAction->setCheckable(true);
        QObject::connect(twoPassAction, SIGNAL(toggled(bool)), viewer, SLOT(setTwoPass(bool)));

    QActionGroup* submodeAction = new QActionGroup(modeMenu);
    modeMenu->addSeparator();
    a = submodeAction->addAction("Coverage" ); a->setShortcut(Qt::SHIFT + Qt::Key_1); a->setCheckable(true); modeMenu->addAction(a);
    a = submodeAction->addAction("Depth"    ); a->setShortcut(Qt::SHIFT + Qt::Key_2); a->setCheckable(true); modeMenu->addAction(a);
    a = submodeAction->addAction("Shadeless"); a->setShortcut(Qt::SHIFT + Qt::Key_4); a->setCheckable(true); modeMenu->addAction(a);
    a = submodeAction->addAction("Color"    ); a->setShortcut(Qt::SHIFT + Qt::Key_5); a->setCheckable(true); modeMenu->addAction(a); a->setChecked(true);
    a = submodeAction->addAction("DEBUG"    ); a->setShortcut(Qt::SHIFT + Qt::Key_6); a->setCheckable(true); modeMenu->addAction(a);
    QObject::connect(submodeAction, SIGNAL(triggered(QAction*)), viewer, SLOT(setSubmode(QAction*)));

    modeMenu->addSeparator();
    noGShaderAction = modeMenu->addAction( "No geometry shader");
        noGShaderAction->setCheckable(true);
        QObject::connect(noGShaderAction, SIGNAL(toggled(bool)), viewer, SLOT(setNoGeometryShader(bool)));
    noFragsAction = modeMenu->addAction( "No fragments");
        noFragsAction->setCheckable(true);
        QObject::connect(noFragsAction, SIGNAL(toggled(bool)), viewer, SLOT(setNoFragments(bool)));

    modeMenu->addSeparator();
    wireframeAction = modeMenu->addAction( "Wireframe");
        wireframeAction->setCheckable(true);
        wireframeAction->setShortcut(Qt::Key_W);
        QObject::connect(wireframeAction, SIGNAL(toggled(bool)), viewer, SLOT(setWireframe(bool)));
    opaqWireframeAction = modeMenu->addAction( "Scribe");
        opaqWireframeAction->setCheckable(true);
		opaqWireframeAction->setShortcut(Qt::Key_W+Qt::ControlModifier);
        QObject::connect(opaqWireframeAction, SIGNAL(toggled(bool)), viewer, SLOT(setScribe(bool)));

    modeMenu->addSeparator();
    QAction *lite = modeMenu->addAction( "Lights");
        lite->setShortcut(Qt::Key_L);
        QObject::connect(lite, SIGNAL(triggered(bool)), viewer, SLOT(setLights()));

	shadowsAction = modeMenu->addAction( "Shadows");
        shadowsAction->setShortcut(Qt::Key_S);
        shadowsAction->setCheckable(true);
        QObject::connect(shadowsAction, SIGNAL(toggled(bool)), viewer, SLOT(setShadows(bool)));
	QAction *dyn = modeMenu->addAction( "    dynamic");
        dyn->setShortcut(Qt::Key_S + Qt::ShiftModifier);
        dyn->setCheckable(true);
        QObject::connect(dyn,  SIGNAL(toggled(bool)), viewer, SLOT(setDynamicShadows(bool)));
        QObject::connect(shadowsAction, SIGNAL(toggled(bool)), dyn,    SLOT(setEnabled(bool)));

	texturesAction = modeMenu->addAction( "Color maps");
        texturesAction->setShortcut(Qt::Key_T);
        texturesAction->setCheckable(true);
        QObject::connect(texturesAction, SIGNAL(toggled(bool)), viewer, SLOT(setTextures(bool)));

	normalMapsAction = modeMenu->addAction( "Normal maps");
        normalMapsAction->setShortcut(Qt::Key_N);
        normalMapsAction->setCheckable(true);
        QObject::connect(normalMapsAction, SIGNAL(toggled(bool)), viewer, SLOT(setBumps(bool)));

	environmentMapsAction = modeMenu->addAction( "Environment maps");
        environmentMapsAction->setShortcut(Qt::Key_E);
        environmentMapsAction->setCheckable(true);
        QObject::connect(environmentMapsAction, SIGNAL(toggled(bool)), viewer, SLOT(setEnvMaps(bool)));

    QActionGroup *envmm = new QActionGroup(modeMenu);
        QObject::connect(envmm, SIGNAL(triggered(QAction*)), viewer, SLOT(setEnvmapMode(QAction*)));
        a = envmm->addAction("    Faces"   ); a->setCheckable(true); modeMenu->addAction(a); a->trigger();
	    a = envmm->addAction("    Parabola"); a->setCheckable(true); modeMenu->addAction(a);
	    a = envmm->addAction("    Sphere"  ); a->setCheckable(true); modeMenu->addAction(a);
	    a = envmm->addAction("    Lambert" ); a->setCheckable(true); modeMenu->addAction(a);

	dynamicEnvmapsAction = modeMenu->addAction( "    Dynamic envmaps");
        dynamicEnvmapsAction->setShortcut(Qt::Key_E + Qt::ControlModifier);
        dynamicEnvmapsAction->setCheckable(true);
        QObject::connect(dynamicEnvmapsAction, SIGNAL(toggled(bool)), viewer, SLOT(setDynamicEnvMaps(bool)));
        QObject::connect(environmentMapsAction, SIGNAL(toggled(bool)), dynamicEnvmapsAction,   SLOT(setEnabled(bool)));

	glossyMapsAction = modeMenu->addAction( "Glossy maps");
        glossyMapsAction->setShortcut(Qt::Key_G);
        glossyMapsAction->setCheckable(true);
        QObject::connect(glossyMapsAction, SIGNAL(toggled(bool)), viewer, SLOT(setGlossyMaps(bool)));

	dynamicGlossymapsAction = modeMenu->addAction( "    dynamic");
        dynamicGlossymapsAction->setShortcut(Qt::Key_G + Qt::ControlModifier);
        dynamicGlossymapsAction->setCheckable(true);
        QObject::connect(dynamicGlossymapsAction, SIGNAL(toggled(bool)), viewer,  SLOT(setDynamicGlossyMaps(bool)));
        QObject::connect(glossyMapsAction,  SIGNAL(toggled(bool)), dynamicGlossymapsAction, SLOT(setEnabled(bool)));

    cullingAction = modeMenu->addAction( "Culling");
        cullingAction->setCheckable(true);
        cullingAction->setShortcut(Qt::Key_C);
        QObject::connect(cullingAction, SIGNAL(toggled(bool)), viewer, SLOT(setCulling(bool)));

    depthtestAction = modeMenu->addAction("Depth test");
        depthtestAction->setCheckable(true);
        depthtestAction->setShortcut(Qt::Key_Z);
        QObject::connect(depthtestAction, SIGNAL(toggled(bool)), viewer, SLOT(setDepthTest(bool)));

    blendingAction = modeMenu->addAction("Blending");
        blendingAction->setCheckable(true);
        blendingAction->setShortcut(Qt::Key_B);
        QObject::connect(blendingAction, SIGNAL(toggled(bool)), viewer, SLOT(setBlending(bool)));

    // ==== Add another menu with optional #defines in shaders ===============
    QMenu *code = mainWindow->menuBar()->addMenu("Code option");
    code->setTearOffEnabled(true);

    QActionGroup *intersect = new QActionGroup(code);
    QObject::connect(intersect, SIGNAL(triggered(QAction*)), viewer, SLOT(setIntersect(QAction*)));
    code->addSeparator();
	a = code->addAction("Ray-Triangle intersect"); a->setEnabled(false);
	a = intersect->addAction("BADOUEL" ); a->setCheckable(true); code->addAction(a); a->trigger();
    a = intersect->addAction("MTH"     ); a->setCheckable(true); code->addAction(a);

    QActionGroup *lens = new QActionGroup(code);
    QObject::connect(lens, SIGNAL(triggered(QAction*)), viewer, SLOT(setLens(QAction*)));
    code->addSeparator();
	a = code->addAction("Lens formula"); a->setEnabled(false);
    a = lens->addAction("1 / LAMBERT"); a->setCheckable(true); code->addAction(a); a->trigger();
	a = lens->addAction("2"      ); a->setCheckable(true); code->addAction(a);

    resetInterface();
	mainWindow->menuBar()->show();
    mainWindow->show();

    //========================================================================
    // Add central viewer once environment is defined...
    mainWindow->setCentralWidget( viewer );
    viewer->show();

    viewer->setStateFileName("TBender.xml");
    viewer->restoreStateFromFile();

    int rc = app.exec();

    viewer->setStateFileName("TBender.xml");
    viewer->saveStateToFile();

    return rc;
}
